#!/usr/bin/perl

$dorun=0;
$BINCODER=$ENV{'BINCODER'};
$decoder="$BINCODER/bin/bincoder";

$usage="USAGE: mert-tst.pl -i f -o f -optimal f [-run] [decoder f] [params]
   -optimal  f : file with optimal file (optimization)
   -i        f : input file to translate
   -o        f : translated output file
   -run        : run test [default $dorun]
   -decoder  f : decoder binary [default $decoder]
   params      : passed to the decoder
";

$#params=-1;
while (@ARGV>0){
    $tok=shift @ARGV;
    if ($tok eq "-optimal" && $#ARGV>=0) {$optimal=shift @ARGV; next;}
    if ($tok eq "-i" && $#ARGV>=0) {$input=shift @ARGV; next;}
    if ($tok eq "-o" && $#ARGV>=0) {$output=shift @ARGV; next;}
    if ($tok eq "-run") {$dorun=1; next;}
    if ($tok eq "-decoder" && $#ARGV>=0) {$decoder=shift @ARGV; next;}
    $params[$#params+1]=$tok;
}

die "ERROR: undefined -optimal parameter\n$usage" unless defined $optimal;
die "ERROR: undefined -i parameter\n$usage" unless defined $input;
die "ERROR: output file: $output already exists!\n" if (-e $output);

open (FILE,"<$optimal") or die "ERROR: unavailable optimal file:$file\n";
$finished=0;
$bestbleu=0;
$bestpars="";
$bestconf="";
while (<FILE>){
    if (/^params \= (.+)$/){
	$pars=$1;
	next;
    }
    if (/^decoder_config \= (.+)$/){
	$conf=$1;
	next;
    }
    if (/^BLEU SCORE for [^\:]+\:(.+)$/){
	$bleu=$1;
	print "BLEU: $bleu CONF: $conf\n";
	if ($bleu>$bestbleu){
	    $bestpars=$pars;
	    $bestconf=$conf;
	    $bestbleu=$bleu;
	}
	next
    }
    if (/^Training finished at/){
	$finished=1;
	print "[FINAL] BLEU: $bestbleu CONF: $bestconf\n\n";
    }
}
close FILE;

die "OPTIMIZATION NOT FINISHED!!\n" unless ($finished);
die "ERROR: undefined -i parameter\n" unless defined $input;
die "ERROR: undefined -o parameter\n" unless defined $output;

$cmd="$decoder $bestpars $bestconf -i $input -o $output @params 2> $output.log";
if ($dorun == 1) {
    print "Runing: $cmd\n";
    system($cmd);
}
else {
    print "to Run: $cmd\n";
}

