#!/usr/bin/perl

print STDERR "   =================================================================
     pruneTuples: prune bilingual units following several criteria 
     jmcrego[at]limsi[dot]fr, February 2010           
   =================================================================
";

$usage="USAGE: pruneTuples -tnb n -max_size n -max_fert n -voc units.voc [-verbose] [-lattices] < units 
   -voc      f : output file with units vocabulary (f.rfreq1.rfreq2 is also created)
   -max_size n : max number of tuple source/target words                 [def unlimitted]
   -max_fert n : max fertility: prune if (ns*fert>nt || nt*fert<ns)      [def unlimitted]
   -lattices   : units extracted after building lattices to avoid source NULLs
";

$lattices=0;
$ns=0;
$nt=0;
$fert=0;
$verbose=0;

while ($#ARGV>=0){
    $tok = shift @ARGV;
    if ($tok =~ /^\-voc$/  && $#ARGV>=0)  {$voc = shift @ARGV;next;}
    if ($tok =~ /^\-max_size$/&& $#ARGV>=0){$ns = shift @ARGV; $nt = $ns; next;}
    if ($tok =~ /^\-max_fert$/&& $#ARGV>=0){$fert = shift @ARGV;next;}
    if ($tok =~ /^\-verbose$/)            {$verbose=1;next;}
    if ($tok =~ /^\-lattices$/)           {$lattices=1;next;}
    die "ERROR: parsing parameter $tok!\n$usage";
}

die "ERROR: no output vocabulary file passed!\n".$usage unless ($voc !~ /^$/);

###########################################################################
# pruning ns,nt ###########################################################
###########################################################################

$maxcost=99;
print STDERR "[BEGIN reading units file (max\_src:$ns,max\_trg:$nt,max\_fert:$fert)]\n";
$nlines=0;
while (<>){
    if (/^EOS/){
	$nlines++;
	if ($nlines % 100000 == 0) {if ($nlines % 1000000 == 0) {print STDERR $nlines;} else {print STDERR ".";}}
	next;
    }
    chomp;
    s/^ +//;
    s/ +$//;
    s/  +/ /g;
    print "READ: $_\n" if ($verbose);
    $unit=$_;
    $unit=~ s/^\(\d+\,\d+\:[^\)]+\) // if ($lattices);
    next if ($unit =~ /^\<s\>$/);
    next if ($unit =~ /^\<\/s\>$/);
    ($u,$v,$rest) = split / \|\|\| /,$unit;
    $N_u{$u}++;
    $N_v{$v}++;
    @uwords=split / /,$u;
    @vwords=split / /,$v;
    next if ($ns>0 && @uwords>$ns); ### pruning by number of src/trg words
    next if ($nt>0 && @vwords>$nt); ### pruning by number of src/trg words
    if ($fert>0){
	$max_side=@uwords; 
	$min_side=@vwords;
	if ($max_side<$min_side){
	    $max_side=$#vwords+1; 
	    $min_side=$#uwords+1;
	}
	next if ($max_side/$min_side > $fert); ### pruning by fertility
    }
    $y = "$u \|\|\| $v";
    $N_y{$y}++;
}
print STDERR "(sents:$nlines, units:".(keys %N_y).")";

print STDERR "\n[OUTPUT voc of units and RelFreq models]\n";
open (FILE,">$voc") or die "ERROR: cannot create file: $voc\n";
open (RELFREQ,">$voc.rfreq1.rfreq2") or die "ERROR: cannot create file: $voc.rfreq1.rfreq2\n";
$nlines=0;
foreach $y (keys %N_y){
    $nlines++;
    if ($nlines % 100000 == 0) {if ($nlines % 1000000 == 0) {print STDERR $nlines;} else {print STDERR ".";}}

    ### RelFreq models
    ($u,$v) = split / \|\|\| /,$y;

    die "ERROR: unit: $y does not exists!\n"     unless (exists $N_y{$y});
    die "ERROR: unit_src: $u does not exists!\n" unless (exists $N_u{$u});
    die "ERROR: unit_trg: $v does not exists!\n" unless (exists $N_v{$v});

    $relfreq_vu=-log ($N_y{$y}/$N_u{$u}); $relfreq_vu=0 if ($relfreq_vu==0);
    die "ERROR: unit: $y with negative cost for vu!\n" unless ($relfreq_vu>=0);

    $relfreq_uv=-log ($N_y{$y}/$N_v{$v}); $relfreq_uv=0 if ($relfreq_uv==0);
    die "ERROR: unit: $y with negative cost for uv!\n" unless ($relfreq_uv>=0);

    print FILE "$y\n";
    print RELFREQ "$relfreq_vu $relfreq_uv\n";
}
print STDERR "(units:$nlines)\n";
