#ifndef _AUXILIAR
#define _AUXILIAR

#include <map>
#include <set>
#include <ext/hash_set>
#include <string>
#include <fstream>
#include <iostream>
#include <sstream>
#include <vector>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <ext/hash_map>
#include <pthread.h>

#include <fst/fstlib.h>
#include <fst/fst.h>
using namespace fst;

using namespace __gnu_cxx;
namespace __gnu_cxx{
  template<> struct hash< std::string >{
    size_t operator()( const std::string& x ) const{return hash< const char* >()( x.c_str() );}
  };
}

using namespace std;

const string str_wrdseparator =  "_|_";
const string str_tplseparator =  "|||";
const string str_separator =    " ||| ";
const string str_wordindexseparator = "@";
const string str_null = "NULL";
const string str_unk = "<unk>";
const string str_eps = "<eps>";
const string str_begsent = "<s>";
const string str_endsent = "</s>";
const string str_endofgraph = "<END>";
const int floatprecision = 6;

class hashfvecstring : public unary_function<vector<string>, size_t>{
 public:
  size_t operator()(vector<string> const &s) const{
    string key;
    for (int i=0;i<s.size();i++) {if (key.size()>0) key.append(" "); key.append(s[i]);}
    return hash< const char* >()( key.c_str() );
  }
};
class hasheqvecstring{
 public:
  bool operator()(vector<string> const &s1, vector<string> const &s2) const{return s1==s2;}
};

template <typename T> string join(vector<T> &v,string sep){
  ostringstream result;
  if (v.size()>0) result << v[0];
  for (int i=1;i<v.size();i++){result << sep << v[i];}
  return result.str();
}

template <typename T> string join(vector<vector<T> > &v,string sep){
  ostringstream result;
  for (int k=0;k<v.size();k++){
    if (k>0) result << " | ";
    if (v[k].size()>0) result << v[k][0];
    for (int i=1;i<v[k].size();i++){result << sep << v[k][i];}
  }
  return result.str();
}

inline vector<string> split(string str,const char *match){
  vector<string> result;
  size_t from=0;
  while (str.find(match,from)!=string::npos){
    result.push_back(str.substr(from,str.find(match,from)-from));
    from=str.find(match,from)+strlen(match);
  }
  result.push_back(str.substr(from));
  return result;
}

class Result{
  string _normalized;
  float _score;
public:
  Result(const string& n, const float s):_normalized(n),_score(s){};
  Result(){};
  Result(const Result& r):_normalized(r._normalized),_score(r._score){};
  ~Result(){};
  string getNormalized(){ return _normalized; }
  float getScore(){ return _score; }
  void setScore(const float score){ _score = score; }
  void setNormalized(const string normalized){ _normalized = normalized; }
};
class fstToStrings{
  vector<string> result;
  void explore(fst::StdFst *f,
               fst::StdArc::StateId currentState,
               Result& r,
               vector<Result>& res,
               fst::SymbolTable* _tableSym){
    if(f->Final(currentState) != fst::StdArc::Weight::Zero()){res.push_back(r);}
    for (fst::ArcIterator<fst::StdFst> aiter(*f,currentState);
         !aiter.Done();
         aiter.Next()){
      string chaine=r.getNormalized();
      float score=r.getScore();
      fst::StdArc arc = aiter.Value();
      string sym=_tableSym->Find ( arc.ilabel );
      float w=arc.weight.Value();
      r.setNormalized(chaine + " " + sym);
      r.setScore(score+w);
      explore(f,arc.nextstate,r,res,_tableSym);
      r.setNormalized(chaine);
      r.setScore(score);
    }
    return;
  }
public:
  fstToStrings(StdVectorFst *fst,fst::SymbolTable *_tableSym){
    vector<Result> res;
    fst::StdArc::StateId initialState = fst->Start();
    Result r("",0);
    explore(fst,initialState,r,res,_tableSym);
    for(int i=0;i<res.size();i++){
      string str=res[i].getNormalized();
      if (str.size()>0) str=str.substr(1); //removes initial 0
      char score[20]; sprintf(score,"%f",res[i].getScore());
      string line=str+str_separator+score;
      result.push_back(line);
    }
    return;
  }
  ~fstToStrings(){result.clear();}
  int size(){return result.size();}
  string n(int i){return result[i];}
};

#endif
