#include "BeamSrch.h"

/****************************************************************************************/
/*** BeamSrch ***************************************************************************/
/****************************************************************************************/

BeamSrch::BeamSrch(Argument &pars, Sentence &sent,int nSent,
		   vector<LangModel*> &BilLM, vector<LangModel*> &SrcLM, vector<LangModel*> &TrgLM,
		   vector<LangModel*> &sBilLM, vector<LangModel*> &sSrcLM, vector<LangModel*> &sTrgLM,
		   vector<XModel*> &sBilXM, vector<XModel*> &sSrcXM, vector<XModel*> &sTrgXM){
  numsent=nSent;
  /*** set initial hypothesis in search space ***/
  initial(pars,sent);
  /*** traverse the search space (staks) ***/
  traverse(pars,sent,BilLM,SrcLM,TrgLM,sBilLM,sSrcLM,sTrgLM,sBilXM,sSrcXM,sTrgXM);
  /*** output ***/
  output(pars);
}

void BeamSrch::initial(Argument &pars,Sentence &sent){
  if (pars.verbose) pars.f_verbose << "<set initial hypothesis>" << endl;
  hyp h;
  h.father=NULL;
  h.tunit=NULL;
  h.Id=sent.hypAdded;
  h.node_in_lattice=0;
  h.Nsrc_words_covered=0;

  bool isFinal=(sent.hset_nodesfinal.find(h.node_in_lattice)!=sent.hset_nodesfinal.end());

  //key (empty vectors)
  vector<bool> bitvector;
  h.bitvector=bitvector;
  vector<vector<string> > vv;
  h.NBilFactors=vv;
  for (int m=0;m<pars.fBilLM.size();m++){
    if (pars.lb.size()<=m) break;
    vector<string> v;
    h.NBilFactors.push_back(v);
    h.wBilLM.push_back(0);
  }
  h.NSrcFactors=vv;
  for (int m=0;m<pars.ls.size();m++){
    vector<string> v;
    h.NSrcFactors.push_back(v);
    h.wSrcLM.push_back(0);
  }
  h.NTrgFactors=vv;
  for (int m=0;m<pars.lt.size();m++){
    vector<string> v;
    h.NTrgFactors.push_back(v);
    h.wTrgLM.push_back(0);
  }

  h.NsBilFactors=vv;
  for (int m=0;m<pars.lsb.size();m++){
    vector<string> v;
    h.NsBilFactors.push_back(v);
    h.wsBilLM.push_back(0);
  }
  h.NsSrcFactors=vv;
  for (int m=0;m<pars.lss.size();m++){
    vector<string> v;
    h.NsSrcFactors.push_back(v);
    h.wsSrcLM.push_back(0);
  }
  h.NsTrgFactors=vv;
  for (int m=0;m<pars.lst.size();m++){
    vector<string> v;
    h.NsTrgFactors.push_back(v);
    h.wsTrgLM.push_back(0);
  }


  h.NxBilFactors=vv;
  for (int m=0;m<pars.lxb.size();m++){
    vector<string> v;
    h.NxBilFactors.push_back(v);
    h.wsBilXM.push_back(0);
  }
  h.NxSrcFactors=vv;
  for (int m=0;m<pars.lss.size();m++){
    vector<string> v;
    h.NxSrcFactors.push_back(v);
    h.wsSrcXM.push_back(0);
  }
  h.NxTrgFactors=vv;
  for (int m=0;m<pars.lst.size();m++){
    vector<string> v;
    h.NxTrgFactors.push_back(v);
    h.wsTrgXM.push_back(0);
  }

  for (int m=0;m<pars.la.size();m++){
    h.wScores.push_back(0);
  }

  h.wlatticeScore=0;
  h.wdistortion=0;
  h.wp0=0;
  h.wp1=0;
  h.wp2=0;

  //h.wlexReordering;
  for (int m=0;m<12;m++) //MSDCFB
    h.wlexReordering.push_back(0);

  /*** weighted global score ***/
  h.score=0;

  /*** insert hyp ***/
  if (isFinal){
    h.numstack=-1;
    finalStack.insert(pars,sent,h);
    initialStack=-1;
  }
  else{
    if (pars.strategy.compare("2J")==0) h.numstack=h.node_in_lattice;
    else if (pars.strategy.compare("J")==0) h.numstack=h.Nsrc_words_covered;
    else {cerr << "error: wrong search strategy!" << endl; exit(1);}

    Stack stack;
    num2stack[h.numstack]=stack;
    num2stack[h.numstack].insert(pars,sent,h);
    initialStack=h.numstack;
  }
}

void BeamSrch::traverse(Argument &pars,Sentence &sent,
			vector<LangModel*> &BilLM, vector<LangModel*> &SrcLM, vector<LangModel*> &TrgLM,
			vector<LangModel*> &sBilLM, vector<LangModel*> &sSrcLM, vector<LangModel*> &sTrgLM,
			vector<XModel*> &sBilXM, vector<XModel*> &sSrcXM, vector<XModel*> &sTrgXM){
  set<int>::iterator curr_node_it;
  int curr_stack;
  first_stack(pars,sent,curr_node_it,curr_stack);
  while (curr_stack!=-1){
    hash_map<int, Stack>::iterator stack_it=num2stack.find(curr_stack);
    if (pars.verbose) pars.f_verbose << "<expand stack:" << curr_stack << " nhyps:" << num2stack[curr_stack].nhyps() << ">" << endl;
    stack_it->second.sbegin();
    while (!stack_it->second.send()){
      sent.hypExpanded++;
      //expand hypotheses (father_it) using tunits (Tunits_it)
      set<hyp,sethypminorcost>::iterator father_it=stack_it->second.get_shyp_it(); 
      if (pars.verbose) pars.f_verbose << "   <expand hyp:" << father_it->Id << " in stack:" << curr_stack << " score:" << father_it->score << ">" << endl;
      hash_set<int>::iterator Tunits_it=sent.node2hset_TunitsPos[father_it->node_in_lattice].begin();
      while (Tunits_it!=sent.node2hset_TunitsPos[father_it->node_in_lattice].end()){
	addHyp(pars,sent,*father_it,sent.Tunits[*Tunits_it],BilLM,SrcLM,TrgLM,sBilLM,sSrcLM,sTrgLM,sBilXM,sSrcXM,sTrgXM);
	Tunits_it++;
      }
      //next stack
      stack_it->second.snext();
    }
    next_stack(pars,sent,curr_node_it,curr_stack);
  }
  if (pars.verbose) pars.f_verbose << "<endSearch stacks:" << num2stack.size() << " hypExpanded:" << sent.hypExpanded << " hypPruned:" << sent.hypPruned << " hypAdded:" << sent.hypAdded << " hypRecombined:" << sent.hypRecombined << ">" << endl;
}

void BeamSrch::addHyp(Argument &pars,Sentence &sent,
		      const hyp &father,Tunit &tunit,
		      vector<LangModel*> &BilLM, vector<LangModel*> &SrcLM, vector<LangModel*> &TrgLM,
		      vector<LangModel*> &sBilLM, vector<LangModel*> &sSrcLM, vector<LangModel*> &sTrgLM,
		      vector<XModel*> &sBilXM, vector<XModel*> &sSrcXM, vector<XModel*> &sTrgXM){
  hyp h;
  h.father=&father;
  h.tunit=&tunit;
  h.Id=sent.hypAdded;
  h.node_in_lattice=h.tunit->Arc.back();
  h.Nsrc_words_covered=father.Nsrc_words_covered+h.tunit->srcTunit.size();

  h.bitvector=father.bitvector;
  for (int i=0;i<h.tunit->srcIndex.size();i++){
    if (h.tunit->srcIndex[i]==-1) continue;
    while (h.bitvector.size()<=h.tunit->srcIndex[i]) h.bitvector.push_back(false);
    h.bitvector[h.tunit->srcIndex[i]]=true;
  }

  /**************************/
  /*** compute next stack ***/
  /**************************/
  bool isFinal=(sent.hset_nodesfinal.find(h.node_in_lattice)!=sent.hset_nodesfinal.end());
  Stack *nextStack;
  if (isFinal){
    if (pars.verbose) pars.f_verbose << "\t\t<new stack:" << h.numstack << " (final)>" << endl;
    nextStack=&finalStack;
    h.numstack=-1;
  }
  else{
    if (pars.strategy.compare("2J")==0) h.numstack=h.node_in_lattice;
    else if (pars.strategy.compare("J")==0) h.numstack=h.Nsrc_words_covered;
    else {cerr << "error: wrong search strategy!" << endl; exit(1);}
    if (num2stack.find(h.numstack)==num2stack.end()){
      if (pars.verbose) pars.f_verbose << "\t\t<new stack:" << h.numstack << ">" << endl;
      Stack stack;
      num2stack[h.numstack]=stack;
    }
    nextStack=&num2stack[h.numstack];
  }

  if (pars.verbose){
    pars.f_verbose << "\t<tunit: " << join(h.tunit->srcTunit," ") << " ||| " << join(h.tunit->trgTunit," ") << ">" << endl;
    pars.f_verbose << "\t\tlatticeNodes: " << join(h.tunit->Arc," ") << endl;
    pars.f_verbose << "\t\tnSrc: " << h.Nsrc_words_covered << endl;
    pars.f_verbose << "\t\tbitvector: " << join(h.bitvector,"") << endl;
  }

  /********************/
  /*** global score ***/
  /********************/
  h.score=father.score;

  /*************/
  /*** bonus ***/
  /*************/
  h.wp0=-1;
  h.score+=h.wp0*pars.lp0;
  if (pars.verbose) pars.f_verbose << "\t\tbonus_tpl: " << h.wp0 << "x" << pars.lp0 << " (accScore:" << h.score << ")" << endl;
  h.wp1=-h.tunit->trgTunit.size();
  h.score+=h.wp1*pars.lp1;
  if (pars.verbose) pars.f_verbose << "\t\tbonus_trg: " << h.wp1 << "x" << pars.lp1 << " (accScore:" << h.score << ")" << endl;
  h.wp2=-h.tunit->srcTunit.size();
  h.score+=h.wp2*pars.lp2;
  if (pars.verbose) pars.f_verbose << "\t\tbonus_src: " << h.wp2 << "x" << pars.lp2 << " (accScore:" << h.score << ")" << endl;

  /******************/
  /*** distortion ***/
  /******************/
  //distortion is calculated word by word!! It computes the 'distortion' suffered by the current word
  //i.e. if the current occupies position p, while to be monotonous it should occupy position k then
  //distortion is abs(k-p)
  //distorition for initial and final tuples is 0 (<s>, </s>)
  h.wdistortion=0;
  if (h.tunit->srcIndex[0]!=0 && h.tunit->srcIndex[0]!=-1){ //tuples <s> and </s> are not considered
    int prev=-1;
    if (father.tunit->srcIndex.size()>0) prev=father.tunit->srcIndex.back();
    for (int i=0;i<h.tunit->srcIndex.size();i++){
      if (h.tunit->srcIndex[i]!=0 && h.tunit->srcIndex[i]!=-1)
	h.wdistortion+=abs(h.tunit->srcIndex[i]-(prev+1));
      prev=h.tunit->srcIndex[i];
    }
  }
  h.score+=h.wdistortion*pars.ld;
  vector<int> vfather;
  if (father.tunit!=NULL) vfather=father.tunit->srcIndex; //case tuple <s> (father has NULLed tunit)
  if (pars.verbose) 
    pars.f_verbose << "\t\tdistortion: prev_indexs(" << join(vfather," ") << ") curr_indexs(" << join(h.tunit->srcIndex," ") << ") " << h.wdistortion << "x" << pars.ld << " (accScore:" << h.score << ")" << endl;
  //  if (nofuture(pars,sent,nextStack,h.score)) return;

  /*********************/
  /*** lexreordering ***/
  /*********************/
  //lexicalized reordering for initial and final tuples is 0 (<s>, </s>)
  bool C=false,F=false,B=false,M=false,S=false,D=false;
  if (h.tunit->srcIndex[0]!=0 && h.tunit->srcIndex[0]!=-1){ //tuples <s> and </s> are not considered
    int min_prev=99999,max_prev=0;
    for (int i=0;i<father.tunit->srcIndex.size();i++){
      if (father.tunit->srcIndex[i]>max_prev) max_prev=father.tunit->srcIndex[i];
      if (father.tunit->srcIndex[i]<min_prev) min_prev=father.tunit->srcIndex[i];
    }
    int min_curr=99999,max_curr=0;
    for (int i=0;i<h.tunit->srcIndex.size();i++){
      if (h.tunit->srcIndex[i]>max_curr)  max_curr=h.tunit->srcIndex[i];
      if (h.tunit->srcIndex[i]<min_curr)  min_curr=h.tunit->srcIndex[i];
    }
    if      (max_prev+1==min_curr){M=true; C=true;}
    else if (max_curr+1==min_prev){S=true; C=true;}
    else if (max_prev+1<min_curr) {F=true; D=true;}
    else if (max_curr+1<min_prev) {B=true; D=true;}
    else                          {S=true; C=true;}
  }
  if (pars.verbose) pars.f_verbose << "\t\t(MSD-CFB):" << M << S << D << "-" << C << F << B << endl;
  for (int i=0;i<12;i++) h.wlexReordering.push_back(0);
  // lexreo(curr|prev)
  h.wlexReordering[0]=0;
  if (M && h.tunit->lexReordering.size()==6){
    h.wlexReordering[0]=h.tunit->lexReordering[0];
    h.score+=h.wlexReordering[0]*pars.lmc;
    if (pars.verbose) pars.f_verbose << "\t\tlexreo(M,curr|prev): " << h.wlexReordering[0] << "x" << pars.lmc << " (accScore:" << h.score << ")" << endl;
  }
  h.wlexReordering[1]=0;
  if (S && h.tunit->lexReordering.size()==6){
    h.wlexReordering[1]=h.tunit->lexReordering[1];
    h.score+=h.wlexReordering[1]*pars.lsc;
    if (pars.verbose) pars.f_verbose << "\t\tlexreo(S,curr|prev): " << h.wlexReordering[1] << "x" << pars.lsc << " (accScore:" << h.score << ")" << endl;
  }
  h.wlexReordering[2]=0;
  if (D && h.tunit->lexReordering.size()==6){
    h.wlexReordering[2]=h.tunit->lexReordering[2];
    h.score+=h.wlexReordering[2]*pars.ldc;
    if (pars.verbose) pars.f_verbose << "\t\tlexreo(D,curr|prev): " << h.wlexReordering[2] << "x" << pars.ldc << " (accScore:" << h.score << ")" << endl;
  }
  h.wlexReordering[3]=0;
  if (C && h.tunit->lexReordering.size()==6){
    h.wlexReordering[3]=h.tunit->lexReordering[3];
    h.score+=h.wlexReordering[3]*pars.lcc;
    if (pars.verbose) pars.f_verbose << "\t\tlexreo(C,curr|prev): " << h.wlexReordering[3] << "x" << pars.lcc << " (accScore:" << h.score << ")" << endl;
  }
  h.wlexReordering[4]=0;
  if (F && h.tunit->lexReordering.size()==6){
    h.wlexReordering[4]=h.tunit->lexReordering[4];
    h.score+=h.wlexReordering[4]*pars.lfc;
    if (pars.verbose) pars.f_verbose << "\t\tlexreo(F,curr|prev): " << h.wlexReordering[4] << "x" << pars.lfc << " (accScore:" << h.score << ")" << endl;
  }
  h.wlexReordering[5]=0;
  if (B && h.tunit->lexReordering.size()==6){
    h.wlexReordering[5]=h.tunit->lexReordering[5];
    h.score+=h.wlexReordering[5]*pars.lbc;
    if (pars.verbose) pars.f_verbose << "\t\tlexreo(B,curr|prev): " << h.wlexReordering[5] << "x" << pars.lbc << " (accScore:" << h.score << ")" << endl;
  }
  // lexreo(prev|curr)
  h.wlexReordering[6]=0;
  if (M && father.tunit!=NULL && h.tunit->lexReordering.size()==6){
    h.wlexReordering[6]=father.tunit->lexReordering[0];
    h.score+=h.wlexReordering[6]*pars.lmp;
    if (pars.verbose) pars.f_verbose << "\t\tlexreo(M,prev|curr): " << h.wlexReordering[6] << "x" << pars.lmp << " (accScore:" << h.score << ")" << endl;
  }
  h.wlexReordering[7]=0;
  if (S && father.tunit!=NULL && h.tunit->lexReordering.size()==6){
    h.wlexReordering[7]=father.tunit->lexReordering[1];
    h.score+=h.wlexReordering[7]*pars.lsp;
    if (pars.verbose) pars.f_verbose << "\t\tlexreo(S,prev|curr): " << h.wlexReordering[7] << "x" << pars.lsp << " (accScore:" << h.score << ")" << endl;
  }
  h.wlexReordering[8]=0;
  if (D && father.tunit!=NULL && h.tunit->lexReordering.size()==6){
    h.wlexReordering[8]=father.tunit->lexReordering[2];
    h.score+=h.wlexReordering[8]*pars.ldp;
    if (pars.verbose) pars.f_verbose << "\t\tlexreo(D,prev|curr): " << h.wlexReordering[8] << "x" << pars.ldp << " (accScore:" << h.score << ")" << endl;
  }
  h.wlexReordering[9]=0;
  if (C && father.tunit!=NULL && h.tunit->lexReordering.size()==6){
    h.wlexReordering[9]=father.tunit->lexReordering[3];
    h.score+=h.wlexReordering[9]*pars.lcp;
    if (pars.verbose) pars.f_verbose << "\t\tlexreo(C,prev|curr): " << h.wlexReordering[9] << "x" << pars.lcp << " (accScore:" << h.score << ")" << endl;
  }
  h.wlexReordering[10]=0;
  if (F && father.tunit!=NULL && h.tunit->lexReordering.size()==6){
    h.wlexReordering[10]=father.tunit->lexReordering[4];
    h.score+=h.wlexReordering[10]*pars.lfp;
    if (pars.verbose) pars.f_verbose << "\t\tlexreo(F,prev|curr): " << h.wlexReordering[10] << "x" << pars.lfp << " (accScore:" << h.score << ")" << endl;
  }
  h.wlexReordering[11]=0;
  if (B && father.tunit!=NULL && h.tunit->lexReordering.size()==6){
    h.wlexReordering[11]=father.tunit->lexReordering[5];
    h.score+=h.wlexReordering[11]*pars.lbp;
    if (pars.verbose) pars.f_verbose << "\t\tlexreo(B,prev|curr): " << h.wlexReordering[11] << "x" << pars.lbp << " (accScore:" << h.score << ")" << endl;
  }
  //  if (nofuture(pars,sent,nextStack,h.score)) return;

  /**************/
  /*** scores ***/
  /**************/
  for (int m=0;m<h.tunit->Scores.size();m++){
    if (pars.la.size()<=m) break;
    h.wScores.push_back(h.tunit->Scores[m]);
    h.score+=h.wScores[m]*pars.la[m];
    if (pars.verbose) pars.f_verbose << "\t\tscores[" << m << "]: " << h.wScores[m] << "x" << pars.la[m] << " (accScore:" << h.score << ")" << endl;
  }  
  //  if (nofuture(pars,sent,nextStack,h.score)) return;

  /********************/
  /*** latticeScore ***/
  /********************/
  h.wlatticeScore=h.tunit->scoreArc;
  h.score+=h.wlatticeScore*pars.lg;
  if (pars.verbose) pars.f_verbose << "\t\tlattice: " << h.wlatticeScore << "x" << pars.lg << " (accScore:" << h.score << ")" << endl;
  //  if (nofuture(pars,sent,nextStack,h.score)) return;

  /***********************/
  /*** Xmodels ***/
  /***********************/
  h.NxBilFactors=father.NxBilFactors;
  for (int m=0;m<sBilXM.size();m++){
    if (pars.lxb.size()<=m) break;
    h.wsBilXM.push_back(0);
    h.NxBilFactors[m].push_back(h.tunit->bilFactors[pars.sBilFactorX[m]]);
    if (h.NxBilFactors[m].size()>sBilXM[m]->order) h.NxBilFactors[m].erase(h.NxBilFactors[m].begin());
    h.wsBilXM[m]=sBilXM[m]->getCost(h.NxBilFactors[m]);
    h.score+=h.wsBilXM[m]*pars.lxb[m];
    if (pars.verbose) pars.f_verbose << "\t\txBilLM[" << m << "] '" << join(h.NxBilFactors[m]," ") << "' " << h.wsBilXM[m] << endl;
    if (pars.verbose) pars.f_verbose << "\t\txBilLM[" << m << "]: " << h.wsBilXM[m] << "x" << pars.lxb[m] << " (accScore:" << h.score << ")" << endl;
  }
  //  if (nofuture(pars,sent,nextStack,h.score)) return;

  h.NxTrgFactors=father.NxTrgFactors;
  for (int m=0;m<sTrgXM.size();m++){
    if (pars.lxt.size()<=m) break;
    h.wsTrgXM.push_back(0);
    vector<string> factors=h.tunit->trgFactors[pars.sTrgFactorX[m]];
    for (int f=0;f<factors.size();f++){
      h.NxTrgFactors[m].push_back(factors[f]);
      if (h.NxTrgFactors[m].size()>sTrgXM[m]->order) h.NxTrgFactors[m].erase(h.NxTrgFactors[m].begin());
      float mycost=sTrgXM[m]->getCost(h.NxTrgFactors[m]);
      h.wsTrgXM[m]+=mycost;
      h.score+=mycost*pars.lxt[m];
      if (pars.verbose) pars.f_verbose << "\t\tsTrgXM[" << m << "] '" << join(h.NxTrgFactors[m]," ") << "' " << mycost << endl;
    }
    if (pars.verbose) pars.f_verbose << "\t\tsTrgXM[" << m << "]: " << h.wsTrgXM[m] << "x" << pars.lxt[m] << " (accScore:" << h.score << ")" << endl;
  }
  //  if (nofuture(pars,sent,nextStack,h.score)) return;

  h.NxSrcFactors=father.NxSrcFactors;
  for (int m=0;m<sSrcXM.size();m++){
    if (pars.lxs.size()<=m) break;
    h.wsSrcXM.push_back(0);
    vector<string> factors=h.tunit->srcFactors[pars.sSrcFactorX[m]];
    for (int f=0;f<factors.size();f++){
      h.NxSrcFactors[m].push_back(factors[f]);
      if (h.NxSrcFactors[m].size()>sSrcXM[m]->order) h.NxSrcFactors[m].erase(h.NxSrcFactors[m].begin());
      float mycost=sSrcXM[m]->getCost(h.NxSrcFactors[m]);
      h.wsSrcXM[m]+=mycost;
      h.score+=mycost*pars.lxs[m];
      if (pars.verbose) pars.f_verbose << "\t\tsSrcXM[" << m << "] '" << join(h.NxSrcFactors[m]," ") << "' " << mycost << endl;
    }
    if (pars.verbose) pars.f_verbose << "\t\tsSrcXM[" << m << "]: " << h.wsSrcXM[m] << "x" << pars.lxs[m] << " (accScore:" << h.score << ")" << endl;
  }
  //  if (nofuture(pars,sent,nextStack,h.score)) return;

  /***********************/
  /*** language models ***/
  /***********************/
  h.NsBilFactors=father.NsBilFactors;
  for (int m=0;m<sBilLM.size();m++){
    if (pars.lsb.size()<=m) break;
    h.wsBilLM.push_back(0);
    h.NsBilFactors[m].push_back(h.tunit->bilFactors[pars.sBilFactor[m]]);
    if (h.NsBilFactors[m].size()>sBilLM[m]->order) h.NsBilFactors[m].erase(h.NsBilFactors[m].begin());
    h.wsBilLM[m]=sBilLM[m]->getCost(h.NsBilFactors[m]);
    h.score+=h.wsBilLM[m]*pars.lsb[m];
    if (pars.verbose) pars.f_verbose << "\t\tsBilLM[" << m << "] '" << join(h.NsBilFactors[m]," ") << "' " << h.wsBilLM[m] << endl;
    if (pars.verbose) pars.f_verbose << "\t\tsBilLM[" << m << "]: " << h.wsBilLM[m] << "x" << pars.lsb[m] << " (accScore:" << h.score << ")" << endl;
  }
  if (nofuture(pars,sent,nextStack,h.score)) return;

  h.NsTrgFactors=father.NsTrgFactors;
  for (int m=0;m<sTrgLM.size();m++){
    if (pars.lst.size()<=m) break;
    h.wsTrgLM.push_back(0);
    vector<string> factors=h.tunit->trgFactors[pars.sTrgFactor[m]];
    for (int f=0;f<factors.size();f++){
      h.NsTrgFactors[m].push_back(factors[f]);
      if (h.NsTrgFactors[m].size()>sTrgLM[m]->order) h.NsTrgFactors[m].erase(h.NsTrgFactors[m].begin());
      float mycost=sTrgLM[m]->getCost(h.NsTrgFactors[m]);
      h.wsTrgLM[m]+=mycost;
      h.score+=mycost*pars.lst[m];
      if (pars.verbose) pars.f_verbose << "\t\tsTrgLM[" << m << "] '" << join(h.NsTrgFactors[m]," ") << "' " << mycost << endl;
    }
    if (pars.verbose) pars.f_verbose << "\t\tsTrgLM[" << m << "]: " << h.wsTrgLM[m] << "x" << pars.lst[m] << " (accScore:" << h.score << ")" << endl;
  }
  if (nofuture(pars,sent,nextStack,h.score)) return;

  h.NsSrcFactors=father.NsSrcFactors;
  for (int m=0;m<sSrcLM.size();m++){
    if (pars.lss.size()<=m) break;
    h.wsSrcLM.push_back(0);
    vector<string> factors=h.tunit->srcFactors[pars.sSrcFactor[m]];
    for (int f=0;f<factors.size();f++){
      h.NsSrcFactors[m].push_back(factors[f]);
      if (h.NsSrcFactors[m].size()>sSrcLM[m]->order) h.NsSrcFactors[m].erase(h.NsSrcFactors[m].begin());
      float mycost=sSrcLM[m]->getCost(h.NsSrcFactors[m]);
      h.wsSrcLM[m]+=mycost;
      h.score+=mycost*pars.lss[m];
      if (pars.verbose) pars.f_verbose << "\t\tsSrcLM[" << m << "] '" << join(h.NsSrcFactors[m]," ") << "' " << mycost << endl;
    }
    if (pars.verbose) pars.f_verbose << "\t\tsSrcLM[" << m << "]: " << h.wsSrcLM[m] << "x" << pars.lss[m] << " (accScore:" << h.score << ")" << endl;
  }
  if (nofuture(pars,sent,nextStack,h.score)) return;

  h.NBilFactors=father.NBilFactors;
  for (int m=0;m<BilLM.size();m++){
    if (pars.lb.size()<=m) break;
    h.wBilLM.push_back(0);
    h.NBilFactors[m].push_back(h.tunit->bilFactors[pars.BilFactor[m]]);
    if (h.NBilFactors[m].size()>BilLM[m]->order) h.NBilFactors[m].erase(h.NBilFactors[m].begin());
    h.wBilLM[m]=BilLM[m]->getCost(h.NBilFactors[m]);
    h.score+=h.wBilLM[m]*pars.lb[m];
    if (pars.verbose) pars.f_verbose << "\t\tBilLM[" << m << "] '" << join(h.NBilFactors[m]," ") << "' " << h.wBilLM[m] << endl;
    if (pars.verbose) pars.f_verbose << "\t\tBilLM[" << m << "]: " << h.wBilLM[m] << "x" << pars.lb[m] << " (accScore:" << h.score << ")" << endl;
  }
  if (nofuture(pars,sent,nextStack,h.score)) return;

  h.NTrgFactors=father.NTrgFactors;
  for (int m=0;m<TrgLM.size();m++){
    if (pars.lt.size()<=m) break;
    h.wTrgLM.push_back(0);
    vector<string> factors=h.tunit->trgFactors[pars.TrgFactor[m]];
    for (int f=0;f<factors.size();f++){
      h.NTrgFactors[m].push_back(factors[f]);
      if (h.NTrgFactors[m].size()>TrgLM[m]->order) h.NTrgFactors[m].erase(h.NTrgFactors[m].begin());
      float mycost=TrgLM[m]->getCost(h.NTrgFactors[m]);
      h.wTrgLM[m]+=mycost;
      h.score+=mycost*pars.lt[m];
      if (pars.verbose) pars.f_verbose << "\t\tTrgLM[" << m << "] '" << join(h.NTrgFactors[m]," ") << "' " << mycost << endl;
    }
    if (pars.verbose) pars.f_verbose << "\t\tTrgLM[" << m << "]: " << h.wTrgLM[m] << "x" << pars.lt[m] << " (accScore:" << h.score << ")" << endl;
  }
  if (nofuture(pars,sent,nextStack,h.score)) return;

  h.NSrcFactors=father.NSrcFactors;
  for (int m=0;m<SrcLM.size();m++){
    if (pars.ls.size()<=m) break;
    h.wSrcLM.push_back(0);
    vector<string> factors=h.tunit->srcFactors[pars.SrcFactor[m]];
    for (int f=0;f<factors.size();f++){
      h.NSrcFactors[m].push_back(factors[f]);
      if (h.NSrcFactors[m].size()>SrcLM[m]->order) h.NSrcFactors[m].erase(h.NSrcFactors[m].begin());
      float mycost=SrcLM[m]->getCost(h.NSrcFactors[m]);
      h.wSrcLM[m]+=mycost;
      h.score+=mycost*pars.ls[m];
      if (pars.verbose) pars.f_verbose << "\t\tSrcLM[" << m << "] '" << join(h.NSrcFactors[m],"' ") << " " << mycost << endl;
    }
    if (pars.verbose) pars.f_verbose << "\t\tSrcLM[" << m << "]: " << h.wSrcLM[m] << "x" << pars.ls[m] << " (accScore:" << h.score << ")" << endl;
  }
  if (nofuture(pars,sent,nextStack,h.score)) return;

  /**********************/
  /*** add hypothesis ***/
  /**********************/
  nextStack->insert(pars,sent,h);
  return;
}

void BeamSrch::output(Argument &pars){
  if (pars.verbose) pars.f_verbose << "<output>" << endl;
  vector<string> output_tunits;
  vector<string> output_target;
  vector<string> output_source;
  if (finalStack.nhyps()!=0){
    finalStack.sbegin();
    set<hyp,sethypminorcost>::iterator hyp_it=finalStack.get_shyp_it();
    const hyp *h=&(*hyp_it);
    if (h!=NULL){
      while (h->father!=NULL){
	vector<string> srcTunit=h->tunit->srcTunit;
	vector<string> trgTunit=h->tunit->trgTunit;
	string unit=join(srcTunit,str_wrdseparator);
	unit.append(str_tplseparator);
	if (trgTunit.size()==0) unit.append("NULL"); //or <unk>
	else unit.append(join(trgTunit,str_wrdseparator));
	//update sequence
	output_tunits.insert(output_tunits.begin(),unit);
	output_source.insert(output_source.begin(),join(srcTunit," "));
	if (trgTunit.size()>0) output_target.insert(output_target.begin(),join(trgTunit," "));
	h=h->father;
      }
    }
  }
  else{
    cerr << "warning: finalStack with 0 hypotheses!\n";
  }
  if (pars.verbose){
    pars.f_verbose << "TUNITS: " << join(output_tunits," ") << endl;
    pars.f_verbose << "SOURCE: " << join(output_source," ") << endl;
    pars.f_verbose << "TARGET: " << join(output_target," ") << endl;
  }
  if (output_target.size()<2 || output_tunits.size()<2) {cerr << "error: malformed output sequence!\n"; exit(1);}

  //I delete initial and last words/tuples, referred to <s> </s> ...
  if (output_target[0].compare(str_begsent)==0 && output_target.back().compare(str_endsent)==0){
    output_target.erase(output_target.begin());
    output_tunits.erase(output_tunits.begin());
    output_target.erase(--output_target.end());
    output_tunits.erase(--output_tunits.end());
  }
  outWords=join(output_target," ");
  if (pars.units) outUnits=join(output_tunits," ");
  if (pars.ograph || pars.nbest){
    outputGraph(pars);
    outputNbest(pars);
  }
  if (pars.nthreads==1){
    pars.f_out << outWords << endl;
    if (pars.units) pars.f_units << outUnits << endl;
    if (pars.ograph){
      for (int k=0;k<outGraph.size();k++) pars.f_graph << outGraph[k] << endl;
      pars.f_graph << "<END>" << endl;
    }
    if (pars.nbest){
      for (int k=0;k<outNbest.size();k++) pars.f_nbest << outNbest[k] << endl;
    }
  }
  return;
}

void BeamSrch::outputGraph(Argument &pars){
  if (pars.verbose) pars.f_verbose << "<outputGraph>" << endl;
  hash_map<int, Stack>::iterator num2stack_it=num2stack.begin();
  while (num2stack_it!=num2stack.end()){
    num2stack_it->second.sbegin();
    while (!num2stack_it->second.send()){
      set<hyp,sethypminorcost>::iterator hyp_it=num2stack_it->second.get_shyp_it(); 
      addArc(pars,&(*hyp_it));
      /*** recombined hypotheses linked to hyp_it ***/
      hash_map<hyp,vector<hyp>,hashfhyp,hasheqhyp>::iterator hrec_it=num2stack_it->second.hrec.find(*hyp_it);
      if (hrec_it!=num2stack_it->second.hrec.end()){
	for (int k=0;k<hrec_it->second.size();k++) addArc(pars,&(hrec_it->second[k]));
      }
      /*** next hyp ***/
      num2stack_it->second.snext();
    }
    /*** next stack ***/
    num2stack_it++;
  }
  finalStack.sbegin();
  while (!finalStack.send()){
    set<hyp,sethypminorcost>::iterator hyp_it=finalStack.get_shyp_it(); 
    addArc(pars,&(*hyp_it));
    /*** recombined hypotheses linked to hyp_it ***/
    hash_map<hyp,vector<hyp>,hashfhyp,hasheqhyp>::iterator hrec_it=finalStack.hrec.find(*hyp_it);
    if (hrec_it!=finalStack.hrec.end()){
      for (int k=0;k<hrec_it->second.size();k++) addArc(pars,&(hrec_it->second[k]));
    }
    /*** next hyp ***/
    finalStack.snext();
  }
  return;
}

void BeamSrch::addArc(Argument &pars, const hyp *h){
  if (h->father==NULL){
    return;
  }
  if (h->numstack==-1){ //isfinal
    ostringstream data;
    data << h->Id;
    outGraph.push_back(data.str());
  }
  ostringstream line;
  hyp myhyp=*h;
  line << myhyp.father->Id << "\t" << myhyp.Id << "\t" << myhyp.info(pars);
  outGraph.push_back(line.str());
  //I use the hypotheses recombined linked to myhyp.father to build the arc
  //(rec,myhyp) 
  hash_map<int, Stack>::iterator num2stack_it=num2stack.find(myhyp.father->numstack);
  if (num2stack_it==num2stack.end()) return;
  hash_map<hyp,vector<hyp>,hashfhyp,hasheqhyp>::iterator hrec_it=num2stack_it->second.hrec.find(*myhyp.father);
  if (hrec_it!=num2stack_it->second.hrec.end()){
    for (int k=0;k<hrec_it->second.size();k++){
      ostringstream line;
      hyp rechyp=hrec_it->second[k];
      line << rechyp.Id << "\t" << myhyp.Id << "\t" << myhyp.info(pars);
      outGraph.push_back(line.str());
    }
  }
}


bool BeamSrch::nofuture(Argument &pars,Sentence &sent,Stack *nextStack,float currscore){
  if (pars.stacksize>0 && nextStack->nhyps()>=pars.stacksize && nextStack->worstScore()<currscore){
    if (pars.verbose) pars.f_verbose << "\t[pruned] " << currscore << " > " << nextStack->worstScore() << endl;
    sent.hypPruned++;
    return true;
  }
  return false;
}

void BeamSrch::first_stack(Argument &pars, Sentence &sent, set<int>::iterator &node_it, int &curr_stack){
  if (pars.strategy.compare("2J")==0){
    node_it=sent.set_nodestopsort.begin();
    curr_stack=*node_it; 
  }
  else curr_stack=0;
}

void BeamSrch::next_stack(Argument &pars, Sentence &sent, set<int>::iterator &node_it, int &curr_stack){
  if (pars.strategy.compare("2J")==0){
    node_it++;
    if (node_it==sent.set_nodestopsort.end()) curr_stack=-1;
    else curr_stack=*node_it; 
    //ATTENTION: if we are not in the final node but there is not next stack I create the stack !!!
    // as there is not path arriving here (the stack will be empty but I will continue my search)
    //jmcrego    if (num2stack.find(curr_stack)==num2stack.end()) curr_stack=-1;
    if (num2stack.find(curr_stack)==num2stack.end()){ //jmcrego
      if (pars.verbose) pars.f_verbose << "\t\t<new stack:" << curr_stack << ">" << endl;//jmcrego
      Stack stack;//jmcrego
      num2stack[curr_stack]=stack;//jmcrego
    }//jmcrego
  }
  else{
    curr_stack++;
    //jmcrego    if (num2stack.find(curr_stack)==num2stack.end()) curr_stack=-1;
    if (num2stack.find(curr_stack)==num2stack.end()){ //jmcrego
      if (pars.verbose) pars.f_verbose << "\t\t<new stack:" << curr_stack << ">" << endl;//jmcrego
      Stack stack;//jmcrego
      num2stack[curr_stack]=stack;//jmcrego
    }//jmcrego
  }
}

void BeamSrch::outputNbest(Argument &pars){
  if (pars.verbose) pars.f_verbose << "<outputNbest>" << endl;
  //------------------
  // parse graph
  //------------------
  hash_map<int, hash_map<int, pair<string,float> > > node2node2wordcost;
  hash_set<int> Finals;
  int Initial=-1;
  int lastNode=-1;
  for (int k=0;k<outGraph.size();k++){
    stringstream ssline(outGraph[k]);
    vector<string> tokens;
    string str;
    while (ssline >> str) tokens.push_back(str);
    if (tokens.size()==4){
      int nodefrom=atoi(tokens[0].c_str());
      if (lastNode<nodefrom) lastNode=nodefrom;
      if (Initial==-1) Initial=nodefrom;
      int nodeto=atoi(tokens[1].c_str());
      if (lastNode<nodeto) lastNode=nodeto;
      string word=tokens[2];
      float cost=atof(tokens[3].c_str());
      if (node2node2wordcost.find(nodefrom)!=node2node2wordcost.end() && node2node2wordcost[nodefrom].find(nodeto)!=node2node2wordcost[nodefrom].end()){
	cerr << "error: repeated ouput graph arc (" << nodefrom << "," << nodeto << ")" << endl;
	exit(1);
      }
      node2node2wordcost[nodefrom][nodeto]=make_pair(word,cost);
    }
    else if (tokens.size()==1){
      Finals.insert(atoi(tokens[0].c_str()));
    }
    else {
      cerr << "error: unparsed output graph line: " << outGraph[k] << endl;
      exit(1);
    }
  }
  //------------------
  // build fst
  //------------------
  fst::SymbolTable* isyms=new fst::SymbolTable("inputsym");
  isyms->AddSymbol(str_eps, 0);
  int next_isym=1;
  fst::SymbolTable* osyms=new fst::SymbolTable("outputsym");
  osyms->AddSymbol(str_eps, 0);
  int next_osym=1;
  StdVectorFst fst;
  for (int k=0;k<=lastNode;k++) fst.AddState();
  hash_map<int, hash_map<int, pair<string,float> > >::iterator it_nodefrom=node2node2wordcost.begin();
  while (it_nodefrom!=node2node2wordcost.end()){
    int nodefrom=(int64)it_nodefrom->first;
    hash_map<int, pair<string,float> >::iterator it_nodeto=it_nodefrom->second.begin();
    while (it_nodeto!=it_nodefrom->second.end()){
      //--------------------
      //extract graph arc info
      //--------------------
      int nodeto=(int64)it_nodeto->first;
      string arc=it_nodeto->second.first;
      float cost=it_nodeto->second.second;
      size_t pos1=arc.find(str_tplseparator.c_str());
      if (pos1==string::npos) {cerr << "error: unparsed arc in graph: " << arc << endl; exit(1);}
      size_t pos2=arc.rfind(str_tplseparator.c_str());
      if (pos1==pos2) {cerr << "error: unparsed arc in graph: " << arc << endl; exit(1);}
      //      cout << "arc: " << arc << endl;
      //      cout << "pos1: " << pos1 << endl;
      //      cout << "pos2: " << pos2 << endl;
      string Target=arc.substr(pos1+3,pos2-(pos1+3));
      //     cout << "Target: " << Target << endl; //delete
      vector<string> TARGET=split(Target,str_wrdseparator.c_str());
      if (TARGET.size()==1 && TARGET[0].compare(str_null)==0) TARGET[0]=str_eps;
      //--------------------
      //update fst i/o symbols
      //--------------------
      if (isyms->Find(arc) == SymbolTable::kNoSymbol) isyms->AddSymbol(arc,next_isym++);
      for (int k=0;k<TARGET.size();k++) if (osyms->Find(TARGET[k]) == SymbolTable::kNoSymbol) osyms->AddSymbol(TARGET[k],next_osym++);
      if (TARGET.size()==1) fst.AddArc(nodefrom, StdArc(isyms->Find(arc),osyms->Find(TARGET[0]),cost,nodeto));
      else{
	int64 nodeprev;
	int64 nodenext=fst.AddState();
	fst.AddArc(nodefrom, StdArc(isyms->Find(arc),osyms->Find(TARGET[0]),cost,nodenext));
	for (int k=1;k<TARGET.size()-1;k++){
	  nodeprev=nodenext;
	  nodenext=fst.AddState();
	  fst.AddArc(nodeprev, StdArc(isyms->Find(str_eps),osyms->Find(TARGET[k]),StdArc::Weight::One(),nodenext));
	}
	fst.AddArc(nodenext, StdArc(isyms->Find(str_eps),osyms->Find(TARGET.back()),StdArc::Weight::One(),nodeto));
      }
      if (nodefrom==Initial) fst.SetStart(nodefrom);  
      if (Finals.find(nodeto)!=Finals.end()) fst.SetFinal(nodeto,StdArc::Weight::One());  
      //----------	
      it_nodeto++;
    }
    it_nodefrom++;
  }
  Connect(&fst);
  //-------------------
  //output nbest
  //-------------------
  RmEpsilon(&fst);
  StdVectorFst fstnbest;
  //I produce 5 times the nbest I need since I cannot use the -unique property of shortestpath on a transductor
  //Later I discard repeated strings using the hash Output
  ShortestPath(fst, &fstnbest, pars.nbest*5);
  Project(&fstnbest,PROJECT_INPUT);
  RmEpsilon(&fstnbest);
  fstToStrings nbest(&fstnbest,isyms);
  int noutput=0;
  hash_set<string> Output;
  for (int i=0; i<nbest.size()&&noutput<pars.nbest; i++){
    vector<string> words;
    vector<string> names;
    vector<float> costs;
    vector<string> arcstotalscore=split(nbest.n(i),str_separator.c_str());
    vector<string> tokens=split(arcstotalscore[0]," ");
    float totalscore=-atof(arcstotalscore[1].c_str());
    for (int k=0;k<tokens.size();k++){
      size_t pos1=tokens[k].find(str_tplseparator.c_str());
      if (pos1==string::npos) {cerr << "error: unparsed token in nbest: " << tokens[k] << endl; exit(1);}
      size_t pos2=tokens[k].rfind(str_tplseparator.c_str());
      if (pos1==pos2) {cerr << "error: unparsed token in nbest: " << tokens[k] << endl; exit(1);}
      string Src=tokens[k].substr(0,pos1);
      string Trg=tokens[k].substr(pos1+3,pos2-(pos1+3));
      string Costs=tokens[k].substr(pos2+3);
      // target words
      vector<string> WORDS=split(Trg,str_wrdseparator.c_str());
      for (int j=0;j<WORDS.size();j++) 
	if (WORDS[j].compare(str_null)!=0 && WORDS[j].compare(str_begsent)!=0 && WORDS[j].compare(str_endsent)!=0) words.push_back(WORDS[j]);
      // scores
      vector<string> namecost=split(Costs,",");
      if (costs.size()==0){ //get names
	for (int j=0;j<namecost.size();j++){
	  vector<string> NAMECOST=split(namecost[j],"=");
	  names.push_back(NAMECOST[0]);
	  costs.push_back(-atof(NAMECOST[1].c_str()));
	}
      }
      else{ //get scores
	for (int j=0;j<namecost.size();j++){
	  vector<string> NAMECOST=split(namecost[j],"=");
	  string name=NAMECOST[0];
	  if (name.compare(names[j])!=0){
	    cerr << "error: unexpected name of score: " << name << " k:" << k << " nbest: " << nbest.n(i) << endl;
	    exit(1);
	  }
	  costs[j]+=-atof(NAMECOST[1].c_str());
	}
      }
    }
    //check if repeated hypothesis
    if (Output.find(join(words," "))!=Output.end()) continue;
    else Output.insert(join(words," "));

    if (costs.size()!=names.size()){
      cerr << "error: diff num of scores/names!" << endl << "scores: " << join(costs," ") << endl << "names: " << join(names," ") << endl ;
      exit(1);
    }
    stringstream snbest;
    if (pars.zmert){
      snbest << numsent-1 << str_separator << join(words," ") << str_separator << join(costs," ") << str_separator << totalscore;
    }
    else{
      vector<string> Costs;
      for (int k=0;k<costs.size();k++){
	ostringstream data;
	data.precision(floatprecision);
	data << names[k] <<  ": " << costs[k];
	Costs.push_back(data.str());
      }
      snbest << numsent-1 << str_separator << join(words," ") << str_separator << join(Costs," ") << str_separator << totalscore;
    }
    outNbest.push_back(snbest.str());
    noutput++;
  }
  return;
}

/****************************************************************************************/
/*** Stack ******************************************************************************/
/****************************************************************************************/

void Stack::insert(Argument &pars,Sentence &sent,hyp &h){
  sent.hypAdded++;
  hash_map<hyp, set<hyp,sethypminorcost>::iterator, hashfhyp, hasheqhyp>::iterator hhyp_it=hhyp.find(h);
  if (hhyp_it!=hhyp.end()){//hyp already exists (recombination)
    sent.hypRecombined++;

    set<hyp,sethypminorcost>::iterator shyp_it=hhyp_it->second;
    if (h.score < (*shyp_it).score){ //new hyp has a lower (better) score -> replace
      if (pars.verbose) pars.f_verbose << "\t<recombination:replace in stack:" << h.numstack << ">" << endl;
      //save recombination
      hyp prev_h=*shyp_it;
      hash_map<hyp,vector<hyp>,hashfhyp,hasheqhyp>::iterator hrec_it=hrec.find(prev_h);
      if (hrec_it==hrec.end()){
	vector<hyp> vhyp;
	hrec[prev_h]=vhyp;
      }
      hrec[prev_h].push_back(prev_h);
      //delete previous
      shyp.erase(shyp_it);
      hhyp.erase(hhyp_it);
      std::pair< set<hyp,sethypminorcost>::iterator, bool > mypair;
      mypair=shyp.insert(h);
      hhyp[h]=mypair.first;
    }
    else{ //oldw hyp has a lower score -> do nothing
      if (pars.verbose) pars.f_verbose << "\t<recombination:keepOld in stack:" << h.numstack << ">" << endl;
      //save recombination
      hash_map<hyp,vector<hyp>,hashfhyp,hasheqhyp>::iterator hrec_it=hrec.find(*shyp_it);
      if (hrec_it==hrec.end()){
	vector<hyp> vhyp;
	hrec[h]=vhyp;
      }
      hrec[h].push_back(h);
    }
  }
  else{
    std::pair< set<hyp,sethypminorcost>::iterator, bool > mypair;
    mypair=shyp.insert(h);
    if (mypair.second==false) {cerr<< "error: inserting hyp!"; exit(1);}
    hhyp[h]=mypair.first;
    if (pars.verbose) pars.f_verbose << "\t<inserted hyp:" << h.Id << " in stack:" << h.numstack << " nHyps:" << shyp.size() << ">" << endl;
    //delete last if more then stacksize elements
    if (pars.stacksize>0 && shyp.size()>pars.stacksize){
      hhyp.erase(*(shyp.rbegin()));
      shyp.erase(--shyp.end());
      if (pars.verbose) pars.f_verbose << "\t<deleted in stack:" << h.numstack << " nHyps:" << shyp.size() << ">" << endl;
    }
  }
  return;
}

void Stack::sbegin(){
  shyp_it=shyp.begin();
  return;
}

bool Stack::send(){
  return shyp_it==shyp.end();
}

void Stack::snext(){
  shyp_it++;
  return;
}

set<hyp,sethypminorcost>::iterator Stack::get_shyp_it(){
  return shyp_it;
}

/****************************************************************************************/
/*** hyp ********************************************************************************/
/****************************************************************************************/

string hyp::info(Argument &pars){
  ostringstream data;
  data.precision(floatprecision);
  float myscore=0;
  data << join(tunit->srcTunit,str_wrdseparator) << str_tplseparator;
  if (tunit->trgTunit.size()==0) data << "NULL" << str_tplseparator; //or <unk>
  else data << join(tunit->trgTunit,str_wrdseparator) << str_tplseparator;
  bool firstscoreoutput=false;
  for (int m=0;m<wBilLM.size();m++){
    if (pars.lb.size()<=m) break;
    if (pars.lb[m]!=0) {
      if (firstscoreoutput) data << ","; 
      firstscoreoutput=true;
      data << "b" << m << "=" << wBilLM[m]; 
      myscore+=wBilLM[m]*pars.lb[m];
    }
  }
  for (int m=0;m<wTrgLM.size();m++){
    if (pars.lt.size()<=m) break;
    if (pars.lt[m]!=0) {
      if (firstscoreoutput) data << ","; 
      firstscoreoutput=true;
      data << "t" << m << "=" << wTrgLM[m]; 
      myscore+=wTrgLM[m]*pars.lt[m];
    }
  }
  for (int m=0;m<wSrcLM.size();m++){
    if (pars.ls.size()<=m) break;
    if (pars.ls[m]!=0) {
      if (firstscoreoutput) data << ","; 
      firstscoreoutput=true; 
      data << "s" << m << "=" << wSrcLM[m]; 
      myscore+=wSrcLM[m]*pars.ls[m];
    }
  }
  for (int m=0;m<wsBilLM.size();m++){
    if (pars.lsb.size()<=m) break;
    if (pars.lsb[m]!=0) {
      if (firstscoreoutput) data << ","; 
      firstscoreoutput=true; 
      data << "sb" << m << "=" << wsBilLM[m]; 
      myscore+=wsBilLM[m]*pars.lsb[m];
    }
  }
  for (int m=0;m<wsTrgLM.size();m++){
    if (pars.lst.size()<=m) break;
    if (pars.lst[m]!=0) {
      if (firstscoreoutput) data << ","; 
      firstscoreoutput=true; 
      data << "st" << m << "=" << wsTrgLM[m]; 
      myscore+=wsTrgLM[m]*pars.lst[m];
    }
  }
  for (int m=0;m<wsSrcLM.size();m++){
    if (pars.lss.size()<=m) break;
    if (pars.lss[m]!=0) {
      if (firstscoreoutput) data << ","; 
      firstscoreoutput=true; 
      data << "ss" << m << "=" << wsSrcLM[m]; 
      myscore+=wsSrcLM[m]*pars.lss[m];
    }
  }
  for (int m=0;m<wsBilXM.size();m++){
    if (pars.lxb.size()<=m) break;
    if (pars.lxb[m]!=0) {
      if (firstscoreoutput) data << ","; 
      firstscoreoutput=true; 
      data << "xb" << m << "=" << wsBilXM[m]; 
      myscore+=wsBilXM[m]*pars.lxb[m];
    }
  }
  for (int m=0;m<wsTrgXM.size();m++){
    if (pars.lxt.size()<=m) break;
    if (pars.lxt[m]!=0) {
      if (firstscoreoutput) data << ","; 
      firstscoreoutput=true; 
      data << "xt" << m << "=" << wsTrgXM[m]; 
      myscore+=wsTrgXM[m]*pars.lxt[m];
    }
  }
  for (int m=0;m<wsSrcXM.size();m++){
    if (pars.lxs.size()<=m) break;
    if (pars.lxs[m]!=0) {
      if (firstscoreoutput) data << ","; 
      firstscoreoutput=true; 
      data << "xs" << m << "=" << wsSrcXM[m]; 
      myscore+=wsSrcXM[m]*pars.lxs[m];
    }
  }
  for (int m=0;m<wScores.size();m++){
    if (pars.la.size()<=m) break;
    if (pars.la[m]!=0) {
      if (firstscoreoutput) data << ","; 
      firstscoreoutput=true;
      data << "a" << m << "=" << wScores[m]; 
      myscore+=wScores[m]*pars.la[m];
    }
  }  
  if (pars.lmc!=0) {data << ",mc=" << wlexReordering[0];  myscore+=wlexReordering[0]*pars.lmc;}
  if (pars.lsc!=0) {data << ",sc=" << wlexReordering[1];  myscore+=wlexReordering[1]*pars.lsc;}
  if (pars.ldc!=0) {data << ",dc=" << wlexReordering[2];  myscore+=wlexReordering[2]*pars.ldc;}

  if (pars.lcc!=0) {data << ",cc=" << wlexReordering[3];  myscore+=wlexReordering[3]*pars.lcc;}
  if (pars.lfc!=0) {data << ",fc=" << wlexReordering[4];  myscore+=wlexReordering[4]*pars.lfc;}
  if (pars.lbc!=0) {data << ",bc=" << wlexReordering[5];  myscore+=wlexReordering[5]*pars.lbc;}

  if (pars.lmp!=0) {data << ",mp=" << wlexReordering[6];  myscore+=wlexReordering[6]*pars.lmp;}
  if (pars.lsp!=0) {data << ",sp=" << wlexReordering[7];  myscore+=wlexReordering[7]*pars.lsp;}
  if (pars.ldp!=0) {data << ",dp=" << wlexReordering[8];  myscore+=wlexReordering[8]*pars.ldp;}

  if (pars.lcp!=0) {data << ",cp=" << wlexReordering[9];  myscore+=wlexReordering[9]*pars.lcp;}
  if (pars.lfp!=0) {data << ",fp=" << wlexReordering[10]; myscore+=wlexReordering[10]*pars.lfp;}
  if (pars.lbp!=0) {data << ",bp=" << wlexReordering[11]; myscore+=wlexReordering[11]*pars.lbp;}

  if (pars.lg!=0)  {data << ",g=" << wlatticeScore; myscore+=wlatticeScore*pars.lg;}
  if (pars.ld!=0)  {data << ",d=" << wdistortion; myscore+=wdistortion*pars.ld;}
  if (pars.lp0!=0) {data << ",p0=" << wp0; myscore+=wp0*pars.lp0;}
  if (pars.lp1!=0) {data << ",p1=" << wp1; myscore+=wp1*pars.lp1;}
  if (pars.lp2!=0) {data << ",p2=" << wp2; myscore+=wp2*pars.lp2;}

  data << "\t" << myscore;
  return data.str();
}
