#ifndef _BEAMSRCH
#define _BEAMSRCH

#include "Auxiliar.h"
#include "Argument.h"
#include "TestFile.h"
#include "LangMods.h"

/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

class hyp{
 public:
  const hyp *father;
  int Id;
  Tunit *tunit;
  int numstack;           //stack where the hyp is stored (-1 if isfinal)
  int node_in_lattice;    //node reached in input lattice by this hyp
  int Nsrc_words_covered; //so far
  /*** hypothesis key ***/
  vector<bool> bitvector; //source words already covered
  vector<vector<string> > NBilFactors; //each entry refers to a BilLM used (built from N tokens of factor i)
  vector<vector<string> > NSrcFactors; //each entry refers to a SrcLM used (built from N tokens of factor i)
  vector<vector<string> > NTrgFactors; //each entry refers to a TrgLM used (built from N tokens of factor i)
  vector<vector<string> > NsBilFactors; //each entry refers to a sBilLM used (built from N tokens of factor i)
  vector<vector<string> > NsSrcFactors; //each entry refers to a sSrcLM used (built from N tokens of factor i)
  vector<vector<string> > NsTrgFactors; //each entry refers to a sTrgLM used (built from N tokens of factor i)
  vector<vector<string> > NxBilFactors; //each entry refers to a sBilXM used (built from N tokens of factor i)
  vector<vector<string> > NxSrcFactors; //each entry refers to a sSrcXM used (built from N tokens of factor i)
  vector<vector<string> > NxTrgFactors; //each entry refers to a sTrgXM used (built from N tokens of factor i)
  /*** unweighted model scores ***/
  vector<float> wBilLM;
  vector<float> wTrgLM;
  vector<float> wSrcLM;
  vector<float> wsBilLM;
  vector<float> wsTrgLM;
  vector<float> wsSrcLM;
  vector<float> wsBilXM;
  vector<float> wsTrgXM;
  vector<float> wsSrcXM;
  vector<float> wScores;
  vector<float> wlexReordering;
  float wlatticeScore;
  float wdistortion;
  float wp0;
  float wp1;
  float wp2;
  /*** weighted global score ***/
  float score;
  string info(Argument &pars);
};

class sethypminorcost{
 public:
  bool operator()(hyp const h1, hyp const h2) const{return h1.score <= h2.score;}
};

class hashfhyp : public unary_function<hyp, size_t>{
 public:
  size_t operator()(hyp const &h) const{
    /*
    ostringstream result;
    //    cerr << "<Start hashfyp>" << endl;
    for (int j=0;j<h.bitvector.size();j++) result << " " << h.bitvector[j];
    for (int j=0;j<h.NBilFactors.size();j++) for (int i=1;i<h.NBilFactors[j].size();i++) result << " " << h.NBilFactors[j][i];
    for (int j=0;j<h.NSrcFactors.size();j++) for (int i=1;i<h.NSrcFactors[j].size();i++) result << " " << h.NSrcFactors[j][i];
    for (int j=0;j<h.NTrgFactors.size();j++) for (int i=1;i<h.NTrgFactors[j].size();i++) result << " " << h.NTrgFactors[j][i];
    for (int j=0;j<h.NsBilFactors.size();j++) for (int i=1;i<h.NsBilFactors[j].size();i++) result << " " << h.NsBilFactors[j][i];
    for (int j=0;j<h.NsSrcFactors.size();j++) for (int i=1;i<h.NsSrcFactors[j].size();i++) result << " " << h.NsSrcFactors[j][i];
    for (int j=0;j<h.NsTrgFactors.size();j++) for (int i=1;i<h.NsTrgFactors[j].size();i++) result << " " << h.NsTrgFactors[j][i];
    //    cerr << "<key>" << result.str() << "</key>" << endl;
    return hash< const char* >()( result.str().c_str() );
    */

    unsigned int result=0;
    //    cerr << "<Start hashfhyp>" << endl;
    for (int j=0;j<h.bitvector.size();j++) result+= (unsigned int)pow((double)h.bitvector[j],double((j%3)+1));
    for (int j=0;j<h.NBilFactors.size();j++) 
      for (int i=1;i<h.NBilFactors[j].size();i++) 
	for (int k=0;k<h.NBilFactors[j][i].size();k++) result+= (unsigned int)pow((double)h.NBilFactors[j][i][k],double((k%3)+1));
    for (int j=0;j<h.NSrcFactors.size();j++) 
      for (int i=1;i<h.NSrcFactors[j].size();i++) 
	for (int k=0;k<h.NSrcFactors[j][i].size();k++) result+= (unsigned int)pow((double)h.NSrcFactors[j][i][k],double((k%3)+1));
    for (int j=0;j<h.NTrgFactors.size();j++) 
      for (int i=1;i<h.NTrgFactors[j].size();i++) 
	for (int k=0;k<h.NTrgFactors[j][i].size();k++) result+= (unsigned int)pow((double)h.NTrgFactors[j][i][k],double((k%3)+1));

    for (int j=0;j<h.NsBilFactors.size();j++) 
      for (int i=1;i<h.NsBilFactors[j].size();i++)
	for (int k=0;k<h.NsBilFactors[j][i].size();k++) result+= (unsigned int)pow((double)h.NsBilFactors[j][i][k],double((k%3)+1));
    for (int j=0;j<h.NsSrcFactors.size();j++) 
      for (int i=1;i<h.NsSrcFactors[j].size();i++)
	for (int k=0;k<h.NsSrcFactors[j][i].size();k++) result+= (unsigned int)pow((double)h.NsSrcFactors[j][i][k],double((k%3)+1));
    for (int j=0;j<h.NsTrgFactors.size();j++) 
      for (int i=1;i<h.NsTrgFactors[j].size();i++)
	for (int k=0;k<h.NsTrgFactors[j][i].size();k++) result+= (unsigned int)pow((double)h.NsTrgFactors[j][i][k],double((k%3)+1));

    for (int j=0;j<h.NxBilFactors.size();j++) 
      for (int i=1;i<h.NxBilFactors[j].size();i++)
	for (int k=0;k<h.NxBilFactors[j][i].size();k++) result+= (unsigned int)pow((double)h.NxBilFactors[j][i][k],double((k%3)+1));
    for (int j=0;j<h.NxSrcFactors.size();j++) 
      for (int i=1;i<h.NxSrcFactors[j].size();i++)
	for (int k=0;k<h.NxSrcFactors[j][i].size();k++) result+= (unsigned int)pow((double)h.NxSrcFactors[j][i][k],double((k%3)+1));
    for (int j=0;j<h.NxTrgFactors.size();j++) 
      for (int i=1;i<h.NxTrgFactors[j].size();i++)
	for (int k=0;k<h.NxTrgFactors[j][i].size();k++) result+= (unsigned int)pow((double)h.NxTrgFactors[j][i][k],double((k%3)+1));

    return result%100000;
  }
};

class hasheqhyp{
 public:
  bool operator()(hyp const &h1, hyp const &h2) const{
    if (h1.bitvector!=h2.bitvector) return false;

    for (int j=0;j<h1.NBilFactors.size();j++){
      if (h1.NBilFactors[j].size()!=h2.NBilFactors[j].size()) return false;
      for (int i=1;i<h1.NBilFactors[j].size();i++) if (h1.NBilFactors[j][i].compare(h2.NBilFactors[j][i])!=0) return false;
    }
    for (int j=0;j<h1.NSrcFactors.size();j++){
      if (h1.NSrcFactors[j].size()!=h2.NSrcFactors[j].size()) return false;
      for (int i=1;i<h1.NSrcFactors[j].size();i++) if (h1.NSrcFactors[j][i].compare(h2.NSrcFactors[j][i])!=0) return false;
    }
    for (int j=0;j<h1.NTrgFactors.size();j++){
      if (h1.NTrgFactors[j].size()!=h2.NTrgFactors[j].size()) return false;
      for (int i=1;i<h1.NTrgFactors[j].size();i++) if (h1.NTrgFactors[j][i].compare(h2.NTrgFactors[j][i])!=0) return false;
    }

    for (int j=0;j<h1.NsBilFactors.size();j++){
      if (h1.NsBilFactors[j].size()!=h2.NsBilFactors[j].size()) return false;
      for (int i=1;i<h1.NsBilFactors[j].size();i++) if (h1.NsBilFactors[j][i].compare(h2.NsBilFactors[j][i])!=0) return false;
    }
    for (int j=0;j<h1.NsSrcFactors.size();j++){
      if (h1.NsSrcFactors[j].size()!=h2.NsSrcFactors[j].size()) return false;
      for (int i=1;i<h1.NsSrcFactors[j].size();i++) if (h1.NsSrcFactors[j][i].compare(h2.NsSrcFactors[j][i])!=0) return false;
    }
    for (int j=0;j<h1.NsTrgFactors.size();j++){
      if (h1.NsTrgFactors[j].size()!=h2.NsTrgFactors[j].size()) return false;
      for (int i=1;i<h1.NsTrgFactors[j].size();i++) if (h1.NsTrgFactors[j][i].compare(h2.NsTrgFactors[j][i])!=0) return false;
    }

    for (int j=0;j<h1.NxBilFactors.size();j++){
      if (h1.NxBilFactors[j].size()!=h2.NxBilFactors[j].size()) return false;
      for (int i=1;i<h1.NxBilFactors[j].size();i++) if (h1.NxBilFactors[j][i].compare(h2.NxBilFactors[j][i])!=0) return false;
    }
    for (int j=0;j<h1.NxSrcFactors.size();j++){
      if (h1.NxSrcFactors[j].size()!=h2.NxSrcFactors[j].size()) return false;
      for (int i=1;i<h1.NxSrcFactors[j].size();i++) if (h1.NxSrcFactors[j][i].compare(h2.NxSrcFactors[j][i])!=0) return false;
    }
    for (int j=0;j<h1.NxTrgFactors.size();j++){
      if (h1.NxTrgFactors[j].size()!=h2.NxTrgFactors[j].size()) return false;
      for (int i=1;i<h1.NxTrgFactors[j].size();i++) if (h1.NxTrgFactors[j][i].compare(h2.NxTrgFactors[j][i])!=0) return false;
    }
    return true;
  }
};

class Stack{
  hash_map<hyp, set<hyp,sethypminorcost>::iterator, hashfhyp, hasheqhyp> hhyp;
  set<hyp,sethypminorcost> shyp;
  set<hyp,sethypminorcost>::iterator shyp_it;
 public:
  hash_map<hyp,vector<hyp>,hashfhyp,hasheqhyp> hrec;
  void insert(Argument&,Sentence&,hyp&);
  void sbegin();
  void snext();
  bool send();
  set<hyp,sethypminorcost>::iterator get_shyp_it();
  int nhyps(){return shyp.size();}
  float worstScore(){return (--shyp.end())->score;}
};

class globalHyp{
 public:
  vector<string> words;
  vector<float> scores;
  float score;
};

class setglobalhypminorcost{
 public:
  bool operator()(globalHyp const h1, globalHyp const h2) const{return h1.score <= h2.score;}
};

/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

class BeamSrch{
  int numsent;
  hash_map<int, Stack> num2stack;
  Stack finalStack;
  int initialStack;
  string outWords,outUnits;
  vector<string> outGraph;
  vector<string> outNbest;
 public:
  BeamSrch(Argument&,Sentence&,int,vector<LangModel*>&,vector<LangModel*>&,vector<LangModel*>&,vector<LangModel*>&,vector<LangModel*>&,vector<LangModel*>&,vector<XModel*>&,vector<XModel*>&,vector<XModel*>&);
  void initial(Argument&,Sentence&);
  void traverse(Argument&,Sentence&,vector<LangModel*>&,vector<LangModel*>&,vector<LangModel*>&,vector<LangModel*>&,vector<LangModel*>&,vector<LangModel*>&,vector<XModel*>&,vector<XModel*>&,vector<XModel*>&);
  void outputGraph(Argument&);
  void outputNbest(Argument&);
  void addHyp(Argument&,Sentence&,const hyp&,Tunit&,vector<LangModel*>&,vector<LangModel*>&,vector<LangModel*>&,vector<LangModel*>&,vector<LangModel*>&,vector<LangModel*>&,vector<XModel*>&,vector<XModel*>&,vector<XModel*>&);
  void output(Argument&);
  void addArc(Argument&,const hyp*);
  void first_stack(Argument&,Sentence&,set<int>::iterator&,int&);
  void next_stack(Argument&,Sentence&,set<int>::iterator&,int&);
  bool nofuture(Argument&,Sentence&,Stack*,float);
  string getWords(){return outWords;}
  string getUnits(){return outUnits;}
  vector<string> getGraph(){return outGraph;}
  vector<string> getNbest(){return outNbest;}
};

#endif
