#include "LangMods.h"

XModel::XModel(string f){
  order=0;
  string line;
  ifstream infile;
  infile.open(f.data());
  if (infile.fail()){
    cerr << "error: opening file " << f << endl;
    exit(1);
  }
  while (getline(infile,line)){
    vector<string> tokens=split(line," ");
    float cost=atof(tokens[0].c_str());
    tokens.erase(tokens.begin());
    if (tokens.size()>order) order=tokens.size();
    model[tokens]=cost;
  }
  infile.close();
}

float XModel::getCost(vector<string> &seq){
  hash_map<vector<string>,float,hashfvecstring,hasheqvecstring>::iterator it_model=model.find(seq);
  if (it_model==model.end()) return 0;
  else return it_model->second;
}

LangModel::LangModel(string f, float incr){
  incr_unk=incr;
  if (lm::ngram::RecognizeBinary(f.c_str(), model_type)) {
    isarpa=false;
    switch(model_type) {
    case lm::ngram::HASH_PROBING:
      hash_probing=new lm::ngram::ProbingModel(f.c_str());
      order=hash_probing->Order();
      index_unk=hash_probing->GetVocabulary().Index(str_unk);
      break;
    case lm::ngram::TRIE_SORTED:
      trie_sorted=new lm::ngram::TrieModel(f.c_str());
      order=trie_sorted->Order();
      index_unk=trie_sorted->GetVocabulary().Index(str_unk);
      break;
    case lm::ngram::HASH_SORTED:
    default:
      cerr << "error: Unrecognized kenlm model type " << model_type << std::endl;
      abort();
    }
  }
  else{
    isarpa=true;
    arpa=new lm::ngram::Model(f.c_str());
    order=arpa->Order();
    index_unk=arpa->GetVocabulary().Index(str_unk);
  }
}

float LangModel::getCost(vector<string> &seq){
  if (seq.size()==1 && seq[0].compare(str_begsent)==0) return 0; //I want p(<s>) be 0!!

  float score;
  lm::ngram::State out,state;
  if (isarpa){
    state=arpa->NullContextState();
    for (int i=0;i<seq.size()-1;i++){arpa->Score(state, arpa->GetVocabulary().Index(seq[i]) , out);state=out;}
    score=-arpa->Score(state, arpa->GetVocabulary().Index(seq.back()) , out);
    if (arpa->GetVocabulary().Index(seq.back())==index_unk) score+=incr_unk;
    return score;
  }
  if (model_type == lm::ngram::HASH_PROBING){
    state=hash_probing->NullContextState();
    for (int i=0;i<seq.size()-1;i++){hash_probing->Score(state, hash_probing->GetVocabulary().Index(seq[i]) , out);state=out;}
    score=-hash_probing->Score(state, hash_probing->GetVocabulary().Index(seq.back()) , out);
    if (hash_probing->GetVocabulary().Index(seq.back())==index_unk) score+=incr_unk;
    return score;
  }
  //lm::ngram::TRIE_SORTED
  state=trie_sorted->NullContextState();
  for (int i=0;i<seq.size()-1;i++){trie_sorted->Score(state, trie_sorted->GetVocabulary().Index(seq[i]) , out);state=out;}
  score=-trie_sorted->Score(state, trie_sorted->GetVocabulary().Index(seq.back()) , out);
  if (trie_sorted->GetVocabulary().Index(seq.back())==index_unk) score+=incr_unk;
  return score;
}

