#include "TestFile.h"

void TestFile::read(Argument &pars){
  if (pars.firstSent==0) pars.firstSent=1;
  int nSent=0;
  cerr << "Reading TestFile" << endl;
  string line;
  std::ifstream infile (pars.fin.c_str(), std::ios_base::in);
  if (!infile.is_open()) {cerr << "error: opening TestFile " << pars.fin << endl; exit(1);}
  //first line is the file description
  getline(infile, line, '\n');
  description=split(line,str_separator.c_str());
  while (getline(infile, line, '\n')){
    if (line.find("<SENTENCE:")==0){ //starts new sentence
      vector<string> data; //vector that will contain all entries of the new sentence
      Data.push_back(data);
      continue;
    }
    if (pars.lastSent==0 || (Data.size()>=pars.firstSent && Data.size()<=pars.lastSent)) Data[Data.size()-1].push_back(line);
  }
  if (pars.lastSent==0) pars.lastSent=Data.size();
}

Tunit::Tunit(Argument &pars, vector<string> &entry, vector<string> &description){
  weightedScore=0;
  string srcwordwithoutindex;
  bool isoov=false;
  //in case of NULL in trgTunit, I clear trgTunit (there is no word) as well as trgFactors
  //in case of <unk> it depends on pars.dropoov:
  //case dropoov=true: trgTunit is cleared as well as trgFactors (same behaviour as NULL)
  //case dropoov=false: trgTunit contains srcwordwithoutindex, srcFactors and trgFactors contain also srcwordwithoutindex
  for (int i=0;i<entry.size();i++){
    /*********************************************************/
    /*** srcTunit ********************************************/
    /*********************************************************/
    if (description[i].compare("srcTunit")==0){
      srcTunit=split(entry[i],string(" ").c_str());
      for (int k=0;k<srcTunit.size();k++){
	size_t pos = srcTunit[k].find(str_wordindexseparator);
	if (pos == string::npos){
	  if (srcTunit[k].compare(str_begsent)==0) srcIndex.push_back(0);
	  else if (srcTunit[k].compare(str_endsent)==0) srcIndex.push_back(-1);
	  else {cerr << "error: missing " << str_wordindexseparator << " in srcTunit entry: " << srcTunit[k] << endl; exit(1);}
	}
	else{
	  string str=srcTunit[k].substr(pos+1);
	  srcIndex.push_back(atoi(str.c_str()));
	}
      }
      if (pars.verbose) pars.f_verbose << "\tsrcIndex:" << join(srcIndex," ") << " (" << srcIndex.size() << ")" << endl;
      if (pars.verbose) pars.f_verbose << "\tsrcTunit:" << join(srcTunit," ") << " (" << srcTunit.size() << ")" << endl;
    }
    /*********************************************************/
    /*** trgTunit ********************************************/
    /*********************************************************/
    else if (description[i].compare("trgTunit")==0){
      trgTunit=split(entry[i],string(" ").c_str());
      if (trgTunit.size()==1 && trgTunit[0].compare(str_null)==0) trgTunit.clear();
      if (trgTunit.size()==1 && trgTunit[0].compare(str_unk)==0){
	isoov=true;
	size_t pos = srcTunit[0].find(str_wordindexseparator);
	if (pos == string::npos) {cerr << "error: missing " << str_wordindexseparator << " in srcTunit entry: " << srcTunit[0] << endl; exit(1);}
	srcwordwithoutindex=srcTunit[0].substr(0,pos);
	if (pars.dropoov) trgTunit.clear();
	else trgTunit[0]=srcwordwithoutindex;
      }
      if (pars.verbose) pars.f_verbose << "\ttrgTunit:" << join(trgTunit," ") << " (" << trgTunit.size() << ")" << endl;
    }
    /*********************************************************/
    /*** Scores **********************************************/
    /*********************************************************/
    else if (description[i].compare("Scores")==0){
      vector<string> scores=split(entry[i],string(" ").c_str());
      if (pars.la.size()>scores.size()) {cerr<< "error: incompatible number of additional model weights!" << endl; exit(1);}
      for (int k=0;k<scores.size()&&k<pars.la.size();k++){
	float score=atof(scores[k].c_str());
	Scores.push_back(score);
	weightedScore+=score*pars.la[k];
      }
      if (pars.verbose) pars.f_verbose << "\tScores:" << join(Scores," ") << " (" << Scores.size() << ") weightedScore:" << weightedScore << endl;
    }
    /*********************************************************/
    /*** lexReordering (MSDCFB) 6 scores! ********************/
    /*********************************************************/
    else if (description[i].compare("lexReordering")==0){
      vector<string> lexreordering=split(entry[i],string(" ").c_str());
      if (lexreordering.size()!=6) {cerr<< "error: incompatible number of lexreordering weights!" << endl; exit(1);}
      for (int k=0;k<lexreordering.size();k++){
	float score=atof(lexreordering[k].c_str());
	lexReordering.push_back(score);
      }
      if (pars.verbose) pars.f_verbose << "\tlexReordering:" << join(lexReordering," ") << " (" << lexreordering.size() << ")" << endl;
    }
    /*********************************************************/
    /*** bilFactors ******************************************/
    /*********************************************************/
    else if (description[i].compare("bilFactors")==0){
      bilFactors=split(entry[i],string(" ").c_str());
      for (int k=0;k<pars.BilFactor.size();k++) if (pars.BilFactor[k]>=bilFactors.size()) {cerr<< "error: wrong number of BilFactors!" << endl; exit(1);}
      for (int k=0;k<pars.sBilFactor.size();k++)if (pars.sBilFactor[k]>=bilFactors.size()){cerr<< "error: wrong number of sBilFactors!" << endl;exit(1);}
      if (pars.verbose) pars.f_verbose << "\tbilFactors:" << join(bilFactors," ") << " (" << bilFactors.size() << ")" << endl;
    }
    /*********************************************************/
    /*** srcFactors ******************************************/
    /*********************************************************/
    else if (description[i].compare("srcFactors")==0){
      vector<string> factors=split(entry[i],string(" ").c_str());
      if (factors.size() % srcTunit.size() != 0) {cerr << "error: bad number of srcFactors tokens: " << factors.size() << endl; exit(1);}
      if (isoov) {for (int k=0; k<factors.size(); k++) factors[k]=srcwordwithoutindex;}
      int nfactors=factors.size()/srcTunit.size();
      for (int k=0;k<pars.SrcFactor.size();k++) if (pars.SrcFactor[k]>=nfactors) {cerr<< "error: wrong number of SrcFactors!" << endl; exit(1);}
      for (int k=0;k<pars.sSrcFactor.size();k++)if (pars.sSrcFactor[k]>=nfactors){cerr<< "error: wrong number of sSrcFactors!" << endl;exit(1);}
      int token=0;
      for (int k=0;k<nfactors;k++){
	vector<string> myfactor;
	for (int aux=0;aux<srcTunit.size();aux++){ 
	  myfactor.push_back(factors[token]);
	  token++;
	}
	srcFactors.push_back(myfactor);
      }
      if (pars.verbose) pars.f_verbose << "\tsrcFactors:" << join(srcFactors," ") << " (" << srcFactors.size() << ")" << endl;
    }
    /*********************************************************/
    /*** trgFactors ******************************************/
    /*********************************************************/
    else if (description[i].compare("trgFactors")==0){ 
      vector<string> factors=split(entry[i],string(" ").c_str());
      if (trgTunit.size()>0 && factors.size()%trgTunit.size() != 0) 
	{cerr << "error: bad number of trgFactors tokens: " << factors.size() << endl; exit(1);}

      int nfactors;
      if (trgTunit.size()>0) nfactors=factors.size()/trgTunit.size();
      else  nfactors=factors.size()/(1+trgTunit.size());
      for (int k=0;k<pars.TrgFactor.size();k++) if (pars.TrgFactor[k]>=nfactors) {cerr<< "error: wrong number of TrgFactors:" << pars.TrgFactor[k] << "!" << endl; exit(1);}
      for (int k=0;k<pars.sTrgFactor.size();k++)if (pars.sTrgFactor[k]>=nfactors){cerr<< "error: wrong number of sTrgFactors!" << endl;exit(1);}

      if (trgTunit.size()==0){//case NULL or <unk> + dropoov
	for (int k=0;k<factors.size();k++){//k empty factors
	  vector<string> myfactor;
	  trgFactors.push_back(myfactor);
	}
      }
      else if (isoov) {
	for (int k=0; k<factors.size(); k++){//k factors with one word (srcwordwithoutindex)
	  vector<string> myfactor;
	  myfactor.push_back(srcwordwithoutindex);
	  trgFactors.push_back(myfactor);
	}
      }
      else{//standard tuple with one or more words in target side
	int token=0;
	for (int k=0;k<nfactors;k++){
	  vector<string> myfactor;
	  for (int aux=0;aux<trgTunit.size();aux++){ 
	    myfactor.push_back(factors[token]);
	    token++;
	  }
	  trgFactors.push_back(myfactor);
	}
      }
      if (pars.verbose) pars.f_verbose << "\ttrgFactors:" << join(trgFactors," ") << " (" << trgFactors.size() << ")" << endl;
    }
    /*********************************************************/
    /*** default *********************************************/
    /*********************************************************/
    else {cerr << "error: unparsed description token: " << description[i] << endl; exit(1);}
  }
}

Sentence::Sentence(Argument &pars, TestFile &testFile, int nSent){
  if (pars.verbose) pars.f_verbose << "<Sentence:" << nSent << ">" << endl;
  if (nSent>testFile.Data.size() || nSent<=0) {cerr << "error: num of sentence:" << nSent << " out of bounds!" << endl; exit(1);}
  if (pars.verbose) pars.f_verbose << "<Parse sentence data>" << endl;
  for (int iData=0;iData<testFile.Data[nSent-1].size();iData++){
    string line=testFile.Data[nSent-1][iData];
    size_t pos=line.find(str_separator);
    if (pos != string::npos) addTunit(pars,line,testFile.description);
    else addLattice(line);
  }
  //*****************************
  //compute sentence structure
  //*****************************
  //hash_map<int, hash_map<string, vector<pair<int,float> > > > node2word2vec_nodecost;
  //hash_map<vector<string>, set<Tunit,settunitminorcost>, hashfvecstring, hasheqvecstring> srcTunit2set_Tunit;
  //hash_set<vector<string>, hashfvecstring, hasheqvecstring > hset_srcTunits_prefixs;
  //vector<Tunit> Tunits;
  //hash_map<int, set<vector<Tunit>::iterator> > node2hset_TunitsPos;
  set<int>::iterator it_node=set_nodestopsort.begin();
  if (pars.verbose) pars.f_verbose << "<sentence structure>" << endl;
  while (it_node!=set_nodestopsort.end()){
    //compute arcs leaving node *it_nodes
    if (pars.verbose){
      string isfinal="";
      if (hset_nodesfinal.find(*it_node)!=hset_nodesfinal.end()) isfinal=" (Final)";
      pars.f_verbose << "[" << *it_node << "]" << isfinal << endl;
    }
    vector<string> prefix;
    compute_structure(pars,*it_node,prefix,*it_node);
    it_node++;
  }
  hypExpanded=0;
  hypAdded=0;
  hypRecombined=0;
  hypPruned=0;
}

void Sentence::addTunit(Argument &pars, string line, vector<string> &description){
  if (pars.verbose) pars.f_verbose << "<Read: " << line << ">" << endl;
  vector<string> entry=split(line,str_separator.c_str());
  if (entry.size()!=description.size()){
    cerr << "error: diff num of tokens between entry and file description!" << endl << "entry: " << line << endl;
    exit(1);
  }
  //add tunit
  Tunit tunit(pars,entry,description);
  if (srcTunit2set_Tunit.find(tunit.srcTunit)==srcTunit2set_Tunit.end()){
    set<Tunit,settunitminorcost> settunits;
    srcTunit2set_Tunit[tunit.srcTunit]=settunits; //empty set
  }
  srcTunit2set_Tunit[tunit.srcTunit].insert(tunit);
  //keep N-best translation choices
  while (pars.tnb>0 && srcTunit2set_Tunit[tunit.srcTunit].size()>pars.tnb) srcTunit2set_Tunit[tunit.srcTunit].erase(--srcTunit2set_Tunit[tunit.srcTunit].end());
  //update hset_srcTunits_prefixs
  vector<string> prefix;
  for (int j=0;j<tunit.srcTunit.size();j++){
    prefix.push_back(tunit.srcTunit[j]);
    hset_srcTunits_prefixs.insert(prefix);
  }	 
}

void Sentence::addLattice(string line){
  vector<string> entry=split(line,string("\t").c_str());
  if (entry.size()!=1 && entry.size()!=3 && entry.size()!=4){
    cerr << "error: unparsed input lattice entry: " << line << endl;
    exit(1);
  }
  if (entry.size()==1){
    int node=atoi(entry[0].c_str());
    hset_nodesfinal.insert(node);
  }
  else{
    int node1=atoi(entry[0].c_str());
    int node2=atoi(entry[1].c_str());
    string wrd=entry[2];
    //test if already exist (node1,node2)=wrd
    if (node2word2vec_nodecost.find(node1) != node2word2vec_nodecost.end()){
      if (node2word2vec_nodecost.find(node1)->second.find(wrd) != node2word2vec_nodecost.find(node1)->second.end()){
	vector<pair<int,float> > vec_tuple=node2word2vec_nodecost.find(node1)->second.find(wrd)->second;
	for (int i=0;i<vec_tuple.size();i++){
	  if (vec_tuple[i].first==node2) {cerr << "error: repeated (node1,node2,wrd) in lattice: " << line << endl; exit(1);}
	}
      }
    }
    float cst=0;
    if (entry.size()==4) cst=atof(entry[3].c_str());
    pair<int,float> tuple=make_pair(node2,cst);
    set_nodestopsort.insert(node1);
    set_nodestopsort.insert(node2);
    if (node2word2vec_nodecost.find(node1)==node2word2vec_nodecost.end()){
      hash_map<string,vector<pair<int,float> > > hash_data;
      node2word2vec_nodecost[node1]=hash_data;
    }
    node2word2vec_nodecost[node1][wrd].push_back(tuple);
  }
}

void Sentence::compute_structure(Argument &pars,int from,vector<string> &prefix,int curr){
  hash_map<int, hash_map<string, vector<pair<int,float> > > >::iterator it_curr=node2word2vec_nodecost.find(curr);
  if (it_curr==node2word2vec_nodecost.end()) return; //no arc leaving node curr
  hash_map<string,vector<pair<int,float> > > hash_curr=it_curr->second;
  hash_map<string,vector<pair<int,float> > >::iterator it_to=hash_curr.begin();
  while (it_to!=hash_curr.end()){
    string wrd=it_to->first;
    vector<string> myprefix=prefix;
    myprefix.push_back(wrd);
    //update sentence data structure
    vector<pair<int,float> > vec_nodecost=it_to->second;
    for (int ipair=0;ipair<vec_nodecost.size();ipair++){
      int node=vec_nodecost[ipair].first;
      if ( srcTunit2set_Tunit.find(myprefix) != srcTunit2set_Tunit.end() ){
	if (pars.verbose) pars.f_verbose << "\t(" << join(myprefix," ") << ")" << endl;
	vector<vector<int> > pathNodes=getPathNodes(from,myprefix);
	for (int i=0;i<pathNodes.size();i++){
	  pathNodes[i].insert(pathNodes[i].begin(),from);
	  //all units with srcTunit equal to myprefix
	  set<Tunit,settunitminorcost> set_Tunit=srcTunit2set_Tunit.find(myprefix)->second;
	  set<Tunit,settunitminorcost>::iterator it_Tunit=set_Tunit.begin();
	  while (it_Tunit!=set_Tunit.end()){
	    Tunit tunit=*it_Tunit;
	    tunit.Arc=pathNodes[i];
	    tunit.scoreArc=getPathScore(pathNodes[i],myprefix);
	    Tunits.push_back(tunit);
	    if (node2hset_TunitsPos.find(from)==node2hset_TunitsPos.end()){
	      hash_set<int> myhset;
	      node2hset_TunitsPos[from]=myhset;
	    }
	    node2hset_TunitsPos[from].insert(Tunits.size()-1);
	    if (pars.verbose) pars.f_verbose << "\t\t(" << join(tunit.trgTunit," ") << "):" << tunit.weightedScore << " [" << join(tunit.Arc," ") << "]" << endl;
	    it_Tunit++;
	  }
	}
      }
      //recursive call
      if (hset_srcTunits_prefixs.find(myprefix) != hset_srcTunits_prefixs.end()) compute_structure(pars,from,myprefix,node);
    }
    it_to++;
  }
}

vector<vector<int> > Sentence::getPathNodes(int from,vector<string> &words){
  vector<vector<int> > nodes;
  if (words.size()==0) return nodes;
  hash_map<int, hash_map<string, vector<pair<int,float> > > >::iterator it_from=node2word2vec_nodecost.find(from);
  if (it_from==node2word2vec_nodecost.end()) return nodes;
  hash_map<string, vector<pair<int,float> > >::iterator it_word=(it_from->second).find(words[0]);
  if (it_word==(it_from->second).end()) return nodes;
  vector<string> mywords=words;
  mywords.erase(mywords.begin());
  vector<pair<int,float> > vec_nodecost=it_word->second;
  for (int i=0;i<vec_nodecost.size();i++){
    int to=vec_nodecost[i].first;
    vector<vector<int> > nextnodes=getPathNodes(to,mywords);
    if (nextnodes.size()==0){
      vector<int> mynodes;
      nextnodes.push_back(mynodes);
    }
    for (int k=0;k<nextnodes.size();k++) nextnodes[k].insert(nextnodes[k].begin(),to);
    for (int k=0;k<nextnodes.size();k++) nodes.push_back(nextnodes[k]);
  }
  return nodes;
}

float Sentence::getPathScore(vector<int> &nodes,vector<string> &words){
  float score=0;
  if (nodes.size()==0) return score;
  if (nodes.size()!=words.size()+1) {cerr << "error: diff num of tokens between nodes and words!" << endl << "nodes:" << join(nodes," ") << " words:" << join(words," ") << endl; exit(1);}
  for (int curr=0;curr<words.size();curr++){
    int from=nodes[curr];
    int to=nodes[curr+1];
    string word=words[curr];
    hash_map<int, hash_map<string, vector<pair<int,float> > > >::iterator it_from=node2word2vec_nodecost.find(from);
    if (it_from==node2word2vec_nodecost.end()) return score;
    hash_map<string, vector<pair<int,float> > >::iterator it_word=(it_from->second).find(word);
    if (it_word==(it_from->second).end()) return score;
    vector<pair<int,float> > vec_nodecost=it_word->second;
    for (int i=0;i<vec_nodecost.size();i++){
      int myto=vec_nodecost[i].first;
      if (myto!=to) continue;
      score+=vec_nodecost[i].second;
    }
  }
  return score;
}


