#!/usr/bin/perl

$usage="
USAGE: align2lexicon -src f -trg f -ali f -lex f
   src file : training source file
   trg file : training target file
   ali file : alignment file
   lex file : lexicon probabilities will be stored in lexicon.f2n and lexicon.n2f

";

while ($#ARGV>=0){
    $tok = shift @ARGV;
    if ($tok =~ /^\-src$/  && $#ARGV>=0) {$src = shift @ARGV;next;}
    if ($tok =~ /^\-trg$/  && $#ARGV>=0) {$trg = shift @ARGV;next;}
    if ($tok =~ /^\-ali$/  && $#ARGV>=0) {$ali = shift @ARGV;next;}
    if ($tok =~ /^\-lex$/  && $#ARGV>=0) {$lex = shift @ARGV;next;}
    die "ERROR: parsing parameter $tok!\n$usage";
}

die $usage unless ($src!~/^$/ && $trg!~/^$/ && $ali!~/^$/ && $lex!~/^$/);

print STDERR "   ================================================
     alig2lex: computes lexicon probabilities  
               from a word-to-word aligned corpus
   ================================================
";

my (%WORD_TRANSLATION,%TOTAL_FOREIGN,%TOTAL_ENGLISH);

open(F,"$src") or die "Can't read $src\n";
open(E,"$trg") or die "Can't read $trg\n";
open(A,"$ali") or die "Can't read $ali\n";
$alignment_id=0;

while(my $a = <A>) {
    $alignment_id++;
    if (($alignment_id % 10000) == 0) { 
	if (($alignment_id % 100000) == 0) { print STDERR "$alignment_id"; }
	else { print STDERR "."; }
    }
    chomp($a);
    my $e = <E>;  chomp($e);
    my @ENGLISH = split(/ /,$e);
    my $f = <F>; chomp($f);
    my @FOREIGN = split(/ /,$f);
    
    my (%FOREIGN_ALIGNED,%ENGLISH_ALIGNED);
    foreach (split(/ /,$a)) {
	my ($fi,$ei) = split(/\-/);
	if ($fi >= scalar(@FOREIGN) || $ei >= scalar(@ENGLISH)) {
	    print STDERR "alignment point ($fi,$ei) out of range (0-$#FOREIGN,0-$#ENGLISH) in line $alignment_id, ignoring\n";
#jmcrego	    die;
	}
	else {
	    # local counts
	    $FOREIGN_ALIGNED{$fi}++;
	    $ENGLISH_ALIGNED{$ei}++;
	    
	    # global counts
	    $WORD_TRANSLATION{$FOREIGN[$fi]}{$ENGLISH[$ei]}++;
	    $TOTAL_FOREIGN{$FOREIGN[$fi]}++;
	    $TOTAL_ENGLISH{$ENGLISH[$ei]}++;
	}
    }
    
    # unaligned words
    for(my $ei=0;$ei<scalar(@ENGLISH);$ei++) {
	next if defined($ENGLISH_ALIGNED{$ei});
	$WORD_TRANSLATION{"NULL"}{$ENGLISH[$ei]}++;
	$TOTAL_ENGLISH{$ENGLISH[$ei]}++;
	$TOTAL_FOREIGN{"NULL"}++;
    }
    for(my $fi=0;$fi<scalar(@FOREIGN);$fi++) {
	next if defined($FOREIGN_ALIGNED{$fi});
	$WORD_TRANSLATION{$FOREIGN[$fi]}{"NULL"}++;
	$TOTAL_FOREIGN{$FOREIGN[$fi]}++;
	$TOTAL_ENGLISH{"NULL"}++;
    }
}

print STDERR "\n";
close(A);
close(F);
close(E);

open(F2E,">$lex.f2n") or die "Can't write $lex.f2n";
open(E2F,">$lex.n2f") or die "Can't write $lex.n2f";
foreach my $f (keys %WORD_TRANSLATION) {
    foreach my $e (keys %{$WORD_TRANSLATION{$f}}) {
	printf F2E "%s %s %.7f\n",$e,$f,$WORD_TRANSLATION{$f}{$e}/$TOTAL_FOREIGN{$f};
	printf E2F "%s %s %.7f\n",$f,$e,$WORD_TRANSLATION{$f}{$e}/$TOTAL_ENGLISH{$e};
    }
}
close(E2F);
close(F2E);
print STDERR "Saved: $lex.f2n and $lex.n2f\n";
