#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>

#if __GNUC__>2
#include <ext/hash_map>
using __gnu_cxx::hash_map;
#else
#include <hash_map>
#endif

#define MAXWORDS 9999
#define MODUL 1000000

using namespace std;

#ifndef HASHUINT
#define HASHUINT
class hashfuint : public unary_function<unsigned int, size_t>{
public:
  size_t operator()(unsigned int const &v) const{
    return v%MODUL;
  }
};
                                                                                   
class hashequint{
public:
  bool operator()(unsigned int const &v1, unsigned int const &v2) const{
    return v1==v2;
  }
};
#endif

/*****************************************************************/
/*****************************************************************/
/*****************************************************************/

void read_params(int, char **);
void usage();
void unfold(string &,string &,string &,string &);
void output();
void join_trgs_of_src_if_more_than_one(unsigned int);
void update_struct(unsigned int, unsigned int);

bool useStags;
bool keep;
vector<vector<bool> > alimatrix;
hash_map<unsigned int,bool,hashfuint,hashequint> Htrg_deleted;
vector<string> src_words;
vector<string> trg_words;
vector<string> src_tags;
unsigned int nline;
ifstream infali;
ifstream infswrd;
ifstream inftwrd;
ifstream infstag;
string fali;
string fswrd;
string ftwrd;
string fstag;
string myerror;

/*****************************************************************/
/* MAIN **********************************************************/
/*****************************************************************/

int main(int argc, char **argv){
  cerr << "   ==========================================================" << endl;
  cerr << "     align2unfold: to extract tuples from a given alignment  " << endl;
  cerr << "                   after reordering the source-side words    " << endl;
  cerr << "     jmcrego[at]limsi[dot]fr, April 2008                     " << endl;
  cerr << "   ==========================================================" << endl;

  read_params(argc,argv);

  myerror="Error";
  if (keep) myerror="Warning";

  infali.open(fali.data());
  if(!infali) {cerr << "ALIGN2UNFOLD: ERROR opening file=" << fali << endl; exit(1);}

  infswrd.open(fswrd.data());
  if(!infswrd) {cerr << "ALIGN2UNFOLD: ERROR opening file=" << fswrd << endl; exit(1);}

  inftwrd.open(ftwrd.data());
  if(!inftwrd) {cerr << "ALIGN2UNFOLD: ERROR opening file=" << ftwrd << endl; exit(1);}

  if (useStags){
    infstag.open(fstag.data());
    if(!infstag) {cerr << "ALIGN2UNFOLD: ERROR opening file=" << fstag << endl; exit(1);}
  }

  string aline,sline,tline,tsline;
  nline=0;
  getline(infali, aline);
  getline(infswrd, sline);
  getline(inftwrd, tline);
  if (useStags) getline(infstag, tsline);
  while (!infali.eof() && !infswrd.eof() && !inftwrd.eof()){
    if (useStags && infstag.eof()) break;
    
    nline++;
    unfold(aline,sline,tline,tsline);
    if (nline % 10000 == 0){
      if (nline % 100000 == 0) cerr << nline;
      else cerr << ".";
    }
    getline(infali,aline);
    getline(infswrd, sline);
    getline(inftwrd, tline);
    if (useStags) getline(infstag, tsline);
  }

  infali.close();
  infswrd.close();
  inftwrd.close();
  if (useStags) infstag.close();
}

/*****************************************************************/
/*****************************************************************/
/*****************************************************************/

void read_params(int argc, char **argv){
  fali="";
  fswrd="";
  ftwrd="";
  fstag="";
  useStags=false;
  keep=false;

  string list("");
  for (int i=1;i<argc;i++){list.append(argv[i]).append(" ");}
  istringstream inStr(list);
  while (inStr){
    string word;
    inStr >> word;
    if (!inStr) continue;

    if (word.compare("-keep")==0){ keep=true; continue; }
    if (word.compare("-ali")==0){ inStr >> word; fali=word; continue; }
    if (word.compare("-swrd")==0){ inStr >> word; fswrd=word; continue; }
    if (word.compare("-twrd")==0){ inStr >> word; ftwrd=word; continue; }
    if (word.compare("-stag")==0){ inStr >> word; useStags=true; fstag=word; continue; }

    cerr << "ALIGN2UNFOLD: Error unparsed parameter " << word << endl;
    usage();
  }

  if (fali.size()==0 || fswrd.size()==0 || ftwrd.size()==0){
    cerr << "ALIGN2UNFOLD: Error missing parameters!" << endl;
    usage();
  }
}

/*****************************************************************/
/*****************************************************************/
/*****************************************************************/

void unfold(string &aline,string &sline,string &tline,string &tsline){
  string numbers = "0123456789";
  src_words.clear();
  trg_words.clear();
  src_tags.clear();
  alimatrix.clear();
  Htrg_deleted.clear();

  istringstream inStrSrc(sline);
  while (inStrSrc){
    string word;
    inStrSrc >> word;                                                                                
    if (!inStrSrc) continue;
    src_words.push_back(word);
  }

  istringstream inStrTrg(tline);
  while (inStrTrg){
    string word;
    inStrTrg >> word;                                                                                
    if (!inStrTrg) continue;
    trg_words.push_back(word);
  }

  if (useStags){
    istringstream inStrtSrc(tsline);
    while (inStrtSrc){
      string word;
      inStrtSrc >> word;                                                                                
      if (!inStrtSrc) continue;
      src_tags.push_back(word);
    }
    if (src_tags.size()!=src_words.size()){
      cerr << "ALIGN2UNFOLD: " << myerror << " different size of src_words and src_tags! line=" << nline << endl;
      cerr << "src_wrds:" << sline << endl;
      cerr << "src_tags:" << tsline << endl;
      if (!keep) exit(1);
      while (src_tags.size()>src_words.size()){src_tags.pop_back();}
      while (src_tags.size()<src_words.size()){src_words.pop_back();}
    }
  }

  for (int t=0;t<trg_words.size();t++){
    vector<bool> vsrc;
    for (int s=0;s<src_words.size();s++) vsrc.push_back(false);
    alimatrix.push_back(vsrc);
  }

  istringstream inStrLnk(aline);
  while (inStrLnk){
    bool goodlink=true;
    string word;
    inStrLnk >> word;                                                                                
    if (!inStrLnk) continue;
    unsigned int pos=word.find_first_not_of(numbers);
    unsigned int src=atoi(word.substr(0,pos).data());
    unsigned int trg=atoi(word.substr(pos+1).data());
    if (src>=src_words.size()){
      cerr << "ALIGN2UNFOLD: " << myerror << " link: " << src << "-" << trg << " incompatible with max_src_words=" << src_words.size()-1 << " line=" << nline << endl;
      cerr << "src:" << sline << endl;
      cerr << "trg:" << tline << endl;
      cerr << "ali:" << aline << endl;
      if (!keep) exit(1);
      goodlink=false;
    }
    if (trg>=trg_words.size()){
      cerr << "ALIGN2UNFOLD: " << myerror << " link: " << src << "-" << trg << " incompatible with max_trg_words=" << trg_words.size()-1 << " line=" << nline << endl;
      cerr << "src:" << sline << endl;
      cerr << "trg:" << tline << endl;
      cerr << "ali:" << aline << endl;
      if (!keep) exit(1);
      goodlink=false;
    }
    if (goodlink) alimatrix[trg][src]=true;
  }

  for (int s=0;s<src_words.size();s++) join_trgs_of_src_if_more_than_one(s);
  output();
}

/*****************************************************************/
/*****************************************************************/
/*****************************************************************/

void join_trgs_of_src_if_more_than_one(unsigned int s){
  unsigned int mintrg=trg_words.size()-1;
  unsigned int maxtrg=0;
  int n=0;
  for (int t=0;t<alimatrix.size();t++){
    if (alimatrix[t][s]){
      n++;
      if (t<mintrg) mintrg=t;
      if (t>maxtrg) maxtrg=t;
    }
  }
  if (n<2) return;
  //all links of trg_words in [mintrg,maxtrg] are collapsed in alimatrix[mintrg]
  //all links of trg_words in [mintrg+1,maxtrg] are marked false
  //trg_words from mintrg+1 to maxtrg are marked deleted
  for (int t=mintrg+1;t<=maxtrg;t++){
    for (int s=0;s<alimatrix[t].size();s++){
      if (alimatrix[t][s]) alimatrix[mintrg][s]=true;
      alimatrix[t][s]=false;
    }
    if (Htrg_deleted.find(t)==Htrg_deleted.end()){
      trg_words[mintrg].append(" ").append(trg_words[t]);
    }
    Htrg_deleted[t]=true;
  }
}

/*****************************************************************/
/*****************************************************************/
/*****************************************************************/

void output(){
  //^src1 src1 ||| trg1 trg2 ||| stag1 stag2 ||| jsrc1 jsrc2$

  //firs I collect the list of src words linked to NULL
  vector<unsigned int> src_NULLed;
  unsigned int next_src_NULLed;
  for (int s=0;s<src_words.size();s++){
    bool nulled=true;
    for (int t=0;t<trg_words.size()&&nulled;t++){
      if (alimatrix[t][s]) nulled=false;
    }
    if (nulled) {
      src_NULLed.push_back(s);
    }
  }
  src_NULLed.push_back(MAXWORDS); //I always have one element in this list

  //now I output tuples following the target side
  next_src_NULLed=src_NULLed[0]; src_NULLed.erase(src_NULLed.begin());
  vector<unsigned int> src_indexs;
  for (int t=0;t<alimatrix.size();t++){
    if (Htrg_deleted.find(t)!=Htrg_deleted.end()) continue;
    vector<unsigned int> swrds;
    string seqswrds;
    string seqstags;
    string seqindex;
    for (int s=0;s<alimatrix[t].size();s++){
      if (alimatrix[t][s]){
	swrds.push_back(s);
	char aux[10]; sprintf(aux,"%d",s);
	if (seqswrds.size()==0){
	  if (useStags) seqstags.append(src_tags[s]);
	  seqswrds.append(src_words[s]);
	  seqindex.append(aux);
	}
	else{
	  if (useStags) seqstags.append(" ").append(src_tags[s]);
	  seqswrds.append(" ").append(src_words[s]);
	  seqindex.append(" ").append(aux);
	}
      }
    }
    //before the next unit, I check if there is a srcNULLed word missed
    int smax;
    if (swrds.size()>0) smax=swrds.back();
    else smax=-1; //trg_NULLed word
    while (smax>=0 && next_src_NULLed<smax){ //next_src_NULLed ||| NULL ||| next_tag_NULLed ||| j_next_src_NULLed
      if (useStags) cout << src_words[next_src_NULLed] << " ||| NULL ||| " << src_tags[next_src_NULLed] << " ||| " << next_src_NULLed << endl;
      else  cout << src_words[next_src_NULLed] << " ||| NULL ||| " << next_src_NULLed << endl;
      next_src_NULLed=src_NULLed[0]; src_NULLed.erase(src_NULLed.begin());
    }
    //now I output the unit
    if (seqswrds.size()==0){
      seqswrds="NULL";
      if (useStags) seqstags="NULL";
      seqindex="-1";
    }
    if (useStags) cout << seqswrds << " ||| " << trg_words[t] << " ||| " << seqstags << " ||| " << seqindex << endl;
    else  cout << seqswrds << " ||| " << trg_words[t] << " ||| " << seqindex << endl;
  }
  //if there are additional src_NULLed words
  while (next_src_NULLed!=MAXWORDS){
    if (useStags) cout << src_words[next_src_NULLed] << " ||| NULL ||| " << src_tags[next_src_NULLed] << " ||| " << next_src_NULLed << endl;
    else  cout << src_words[next_src_NULLed] << " ||| NULL ||| " << next_src_NULLed << endl;
    next_src_NULLed=src_NULLed[0]; src_NULLed.erase(src_NULLed.begin());
  }
  cout << "EOS" << endl;
}

/*****************************************************************/
/*****************************************************************/
/*****************************************************************/

void usage(){
    cerr << "USAGE: align2unfold -ali f -swrd f -twrd f [-stag f] [-keep]" << endl;
    cerr << "   -ali  f : alignment file" << endl;
    cerr << "   -swrd f : training src file" << endl;
    cerr << "   -twrd f : training trg file" << endl;
    cerr << "   -stag f : training src tags file" << endl;
    cerr << "   -keep   : continue if erroneous link" << endl;
    exit(1);
}
