#include "Auxiliar.h"
#include "Argument.h"
#include "LangMods.h"
#include "TestFile.h"
#include "BeamSrch.h"

static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
void *decode(void *);
vector<LangModel*> BilLM;
vector<LangModel*> TrgLM;
vector<LangModel*> SrcLM;
TestFile testFile;
Argument pars;
int currSent;
hash_map<int,string> outWords;
hash_map<int,string> outUnits;
hash_map<int,vector<string> > outGraph;
hash_map<int,vector<string> > outNbest;
int hypExpanded,hypPruned,hypAdded,hypRecombined;

/**************************************************************************/
/********** MAIN **********************************************************/
/**************************************************************************/
int main(int argc, char **argv){
  char hostname[1024];
  hostname[1023] = '\0';
  gethostname(hostname, 1023);
  cerr << "   ==========================================" << endl;
  cerr << "     bincoder: BIlingual N-gram smt deCODER  " << endl;
  cerr << "     jmcrego[at]limsi[dot]fr (May 2011)      " << endl;
  cerr << "   ==========================================" << endl;
  time_t tini;
  tini=time(NULL);
  time( &tini );
  tm *current_time = localtime( &tini );
  string t=asctime(current_time);
  
  /***************************************************************/
  /********** READING PARAMETERS *********************************/
  /***************************************************************/
  pars.read(argc,argv);
  cerr << "host    : " << hostname << endl;

  /***************************************************************/
  /********** READING GLOBAL MODELS ******************************/
  /***************************************************************/
  cerr << "Start loading global data @ " << t.substr(0,t.size()-1) << endl;
  for (int i=0;i<pars.fBilLM.size();i++){
    if (pars.fBilLM.size()==0) continue;
    cerr << "Reading BilLM" << i << endl;
    LangModel *lm=new LangModel(pars.fBilLM[i],pars.BilUnk[i]);
    BilLM.push_back(lm);
  }
  for (int i=0;i<pars.fTrgLM.size();i++){
    if (pars.fTrgLM.size()==0) continue;
    cerr << "Reading TrgLM" << i << endl;
    LangModel *lm=new LangModel(pars.fTrgLM[i],pars.TrgUnk[i]);
    TrgLM.push_back(lm);
  }
  for (int i=0;i<pars.fSrcLM.size();i++){
    if (pars.fSrcLM.size()==0) continue;
    cerr << "Reading SrcLM" << i << endl;
    LangModel *lm=new LangModel(pars.fSrcLM[i],pars.SrcUnk[i]);
    SrcLM.push_back(lm);
  }

  /***************************************************************/
  /********** READING TEST ***************************************/
  /***************************************************************/
  testFile.read(pars);
  currSent=pars.firstSent;
  cerr << "[" << difftime(time(NULL),tini) << "''] Done (" << testFile.Data.size() << " sents)" << endl;
  hypExpanded=0;
  hypPruned=0;
  hypAdded=0;
  hypRecombined=0;

  /***************************************************************/
  /********** DECODING *******************************************/
  /***************************************************************/
  time_t tinisent;
  tinisent=time(NULL);
  time( &tinisent );
  current_time = localtime( &tinisent );
  t=asctime(current_time);
  cerr << "Start decoding @ " << t.substr(0,t.size()-1) << endl;
  pthread_attr_t attr;
  pthread_attr_init(&attr);
  pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
  pthread_t tid[pars.nthreads];
  for (int k = 0; k < pars.nthreads; k++){
    if (pthread_create(&tid[k], &attr, decode, NULL) != 0){
      cerr << "failed to create child thread" << endl;
      exit(1);
    }
  }
  for (int k = 0; k < pars.nthreads; k++){
    if (pthread_join(tid[k], NULL) != 0){
      cerr << "failed to join child thread" << endl;
      exit(1);
    }
  }
  if (pars.nthreads>1){//hypotheses has not been output as are not produced in order
    hash_map<int,string>::iterator it;
    hash_map<int,vector<string> >::iterator it2;
    for (int nSent=pars.firstSent;nSent<=pars.lastSent;nSent++){
      it=outWords.find(nSent);
      if (it==outWords.end()) {cerr << "error: missing sentence " << nSent << endl; exit(1);}
      pars.f_out << it->second << endl;
      if (pars.units){
	it=outUnits.find(nSent);
	if (it==outUnits.end()) {cerr << "error: missing sentence " << nSent << endl; exit(1);}
	pars.f_units << it->second << endl;
      }
      if (pars.ograph){
	it2=outGraph.find(nSent);
	if (it2==outGraph.end()) {cerr << "error: missing graph sentence " << nSent << endl; exit(1);}
	pars.f_graph << join(it2->second,"\n") << endl << "<END>" << endl;
      }
      if (pars.nbest){
	it2=outNbest.find(nSent);
	if (it2==outNbest.end()) {cerr << "error: missing nbest sentence " << nSent << endl; exit(1);}
	pars.f_nbest << join(it2->second,"\n") << endl;
      }
    }
  }
  cerr << endl << "[" << difftime(time(NULL),tinisent) << "''] Done (" << (pars.lastSent-pars.firstSent)+1 << " sents) <hypExpanded:" << hypExpanded << " hypPruned:" << hypPruned << " hypAdded:" << hypAdded << " hypRecombined:" << hypRecombined << ">" << endl;
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

void *decode(void *ud){
  while (true) {
    pthread_mutex_lock(&lock);
    int nSent=currSent++; /*************** decode sentence nSent ***/
    pthread_mutex_unlock(&lock);
    if (nSent>pars.lastSent) break;
    /***************************************************************/
    /********** READING SENTENCE-BASED MODELS **********************/
    /***************************************************************/
    vector<LangModel*> sBilLM;
    for (int i=0;i<pars.fsBilLM.size();i++){
      if (pars.fsBilLM.size()==0) continue;
      ostringstream file;
      file << pars.fsBilLM[i] << "." << nSent;
      LangModel *lm=new LangModel(file.str(),pars.sBilUnk[i]);
      sBilLM.push_back(lm);
    }    
    vector<LangModel*> sTrgLM;
    for (int i=0;i<pars.fsTrgLM.size();i++){
      if (pars.fsTrgLM.size()==0) continue;
      ostringstream file;
      file << pars.fsTrgLM[i] << "." << nSent;
      LangModel *lm=new LangModel(file.str(),pars.sTrgUnk[i]);
      sTrgLM.push_back(lm);
    }    
    vector<LangModel*> sSrcLM;
    for (int i=0;i<pars.fsSrcLM.size();i++){
      if (pars.fsSrcLM.size()==0) continue;
      ostringstream file;
      file << pars.fsSrcLM[i] << "." << nSent;
      LangModel *lm=new LangModel(file.str(),pars.sSrcUnk[i]);
      sSrcLM.push_back(lm);
    }    
    vector<XModel*> sBilXM;
    for (int i=0;i<pars.fsBilXM.size();i++){
      if (pars.fsBilXM.size()==0) continue;
      ostringstream file;
      file << pars.fsBilXM[i] << "." << nSent;
      XModel *xm=new XModel(file.str());
      sBilXM.push_back(xm);
    }    
    vector<XModel*> sTrgXM;
    for (int i=0;i<pars.fsTrgXM.size();i++){
      if (pars.fsTrgXM.size()==0) continue;
      ostringstream file;
      file << pars.fsTrgXM[i] << "." << nSent;
      XModel *xm=new XModel(file.str());
      sTrgXM.push_back(xm);
    }    
    vector<XModel*> sSrcXM;
    for (int i=0;i<pars.fsSrcXM.size();i++){
      if (pars.fsSrcXM.size()==0) continue;
      ostringstream file;
      file << pars.fsSrcXM[i] << "." << nSent;
      XModel *xm=new XModel(file.str());
      sSrcXM.push_back(xm);
    }    
    /***************************************************************/
    /********** PARSING SENTENCE nSent *****************************/
    /***************************************************************/
    Sentence Sent(pars,testFile,nSent);
    BeamSrch search(pars,Sent,nSent,BilLM,SrcLM,TrgLM,sBilLM,sSrcLM,sTrgLM,sBilXM,sSrcXM,sTrgXM);
    pthread_mutex_lock(&lock);
    outWords[nSent]=search.getWords();
    outUnits[nSent]=search.getUnits();
    outGraph[nSent]=search.getGraph();
    outNbest[nSent]=search.getNbest();
    hypExpanded+=Sent.hypExpanded;
    hypPruned+=Sent.hypPruned;
    hypAdded+=Sent.hypAdded;
    hypRecombined+=Sent.hypRecombined;
    pthread_mutex_unlock(&lock);
    cerr << ".";
    if (nSent%10==0) cerr << nSent;
  }
}


