#include <stdlib.h>
#include <iostream> 
#include <fstream> 
#include <vector> 
#include <math.h>
#include <string.h>
#include <ext/hash_map>
#include <fst/fstlib.h>

const string str_tplseparator = "|||";
const string str_ruleseparator = "///";
const string str_separator = " ||| ";

const string str_begsent = "<s>";
const string str_endsent = "</s>";
const string str_eps = "<eps>";
const string str_endofgraph = "<END>";
const string str_wordindexseparator = "@";

using namespace fst;
using namespace std;
using namespace __gnu_cxx;
namespace __gnu_cxx{
  template<> struct hash< std::string >{
    size_t operator()( const std::string& x ) const{return hash< const char* >()( x.c_str() );}
  };
}

void run(int,vector<string> &);
void buildNgrams(hash_map<int,hash_map<int,string > > &,hash_map<string,hash_map<string,bool> > &,hash_map<int,vector<vector<string> > > &,int);
bool postfix(hash_map<string,hash_map<string,bool> > &,vector<string> &);
void usage();
vector<string> split(string,const char*);
string join(vector<string> &,string);
string ftunits, fscores, flexrm;
vector<string> fbfactor;
vector<string> fsfactor;
vector<string> ftfactor;
int maxs;
bool verbose, help;
time_t tini;
vector<string> TUNITS;
hash_map<string,vector<int> > stunit2tunitsid;//stunit2rrulesid{"la maison"}=<33,34,35,...> ints are indexes of TUNITS
hash_map<string,bool> postStunit;//postStunit{"la"}=true (stunit postfixes)
vector<vector<string> > BFACTOR;
vector<vector<string> > SFACTOR;
vector<vector<string> > TFACTOR;
vector<string> SCORES;
vector<string> LEXRM;

string Nscores;
string Nlexrm;
string unkbfactors;
string begbfactors;
string endbfactors;
string unksfactors;
string begsfactors;
string endsfactors;
string unktfactors;
string begtfactors;
string endtfactors;

int main(int argc, char** argv){
  cerr << "   ================================================" << endl;
  cerr << "     binfiltr: BIlingual N-gram smt models FILTeR  " << endl;
  cerr << "     by jmcrego@limsi.fr, January 2011             " << endl;
  cerr << "   ================================================" << endl;
  if (argc<=1){
    usage();
    exit(1);
  }
  verbose=false;
  help=false;
  maxs=5;
  /***************************************/
  /********* parsing parameters **********/
  /***************************************/
  int narg=1;
  while (narg<argc){
    string word=argv[narg++];
    if (word.compare("-tunits")==0 && narg<argc) {ftunits=argv[narg++]; continue;}
    if (word.compare("-bilfactor")==0&&narg<argc){fbfactor.push_back(argv[narg++]); continue;}
    if (word.compare("-srcfactor")==0&&narg<argc){fsfactor.push_back(argv[narg++]); continue;}
    if (word.compare("-trgfactor")==0&&narg<argc){ftfactor.push_back(argv[narg++]); continue;}
    if (word.compare("-scores")==0 && narg<argc) {fscores=argv[narg++]; continue;}
    if (word.compare("-lexrm")==0 && narg<argc)  {flexrm=argv[narg++]; continue;}
    if (word.compare("-maxs")==0 && narg<argc)   {maxs=atoi(argv[narg++]); continue;}
    if (word.compare("-verbose")==0)             {verbose=true; continue;}
    if (word.compare("-help")==0)                {help=true; continue;}
    cerr << "ERROR: unparsed parameter: " << word << endl;
    usage(); 
    exit(1);
  }

  /***************************************/
  /********* testing parameters **********/
  /***************************************/

  if (help){
      usage();
      exit(1);
  }

  if (ftunits.size()==0){
    cerr << "error: missing -tunits parameter" << endl;
    usage();
    exit(1);
  }

  if (verbose){
    cerr << "Par tunits  : " << ftunits << endl;
    cerr << "Par scores  : " << fscores << endl;
    cerr << "Par lexrm   : " << flexrm << endl;
    for (int i=0;i<fbfactor.size();i++) cerr << "Par bilfactor" << i << " : " << fbfactor[i] << endl;
    for (int i=0;i<fsfactor.size();i++) cerr << "Par srcfactor" << i << " : " << fsfactor[i] << endl;
    for (int i=0;i<ftfactor.size();i++) cerr << "Par trgfactor" << i << " : " << ftfactor[i] << endl;
    cerr << "Par maxs    : " << maxs << endl;
    cerr << "Par verbose : " << verbose << endl;
  }

  char hostname[1024];
  hostname[1023] = '\0';
  gethostname(hostname, 1023);
  tini=time(NULL);
  time( &tini );
  tm *current_time = localtime( &tini );
  string t=asctime(current_time);
  cerr << "Start reading global data @ " << t.substr(0,t.size()-1) << " @ " << hostname << endl;

  /**************************************/
  /********* reading ftunits ************/
  /**************************************/
  ifstream infile;
  string line;

  cerr << "\treading " << ftunits << endl;
  infile.open(ftunits.data(),ifstream::in);
  if (infile.fail()){
    cerr << "error: opening file " << ftunits << endl;
    exit(1);
  }

  int totaltunits=0;
  while (getline(infile,line)) {
    vector<string> array=split(line,str_separator.c_str());
    if (array.size()==2){
      string stunit=array[0];
      vector<string> STUNIT=split (stunit," ");
      if (STUNIT.size()>maxs) continue;
      string poststunit;
      for (int i=STUNIT.size()-1;i>=0;i--){
	if (i==STUNIT.size()-1) poststunit=STUNIT[i];
	else{
	  string s=STUNIT[i];
	  s.append(" ");
	  s.append(poststunit);
	  poststunit=s;
	}
	postStunit[poststunit]=true;
      }
      string trg=array[1];
      if (stunit2tunitsid.find(stunit)!=stunit2tunitsid.end()) stunit2tunitsid[stunit].push_back(TUNITS.size());
      else{
	vector<int> aux;
	aux.push_back(TUNITS.size());
	stunit2tunitsid[stunit]=aux;
      }
      TUNITS.push_back(line);
    }
    else {cerr << "error: tunit '" << line << "' with other than 2 fields" << endl; exit(1);}
    totaltunits++;
  }
  infile.close();
  if (verbose) cerr << "[" << totaltunits << " tunits]" << endl;

  /**************************************/
  /********* reading scores *************/
  /**************************************/

  if (fscores.size()>0){
    cerr << "\treading " << fscores << endl;
    infile.open(fscores.data(),ifstream::in);
    if (infile.fail()){
      cerr << "error: opening file " << fscores << endl;
      exit(1);
    }
    while (getline(infile,line)) {
      SCORES.push_back(line);
    }
    if (SCORES.size()!=TUNITS.size()){
      cerr << "error:  diff num of entries between scores file: " << fscores << " and tunits voc file: " << ftunits << endl;
      exit(1);
    }
    infile.close();
    if (verbose) cerr << "[" << SCORES.size() << " score entries]" << endl;

    int nscores=split(SCORES[0]," ").size();
    vector<string> vunkscores(nscores,"0");
    Nscores=join(vunkscores," ");
  }

  /**************************************/
  /********* reading lexrm **************/
  /**************************************/

  if (flexrm.size()>0){
    cerr << "\treading " << flexrm << endl;
    infile.open(flexrm.data(),ifstream::in);
    if (infile.fail()){
      cerr << "error: opening file " << flexrm << endl;
      exit(1);
    }
    while (getline(infile,line)) {
      LEXRM.push_back(line);
    }
    if (LEXRM.size()!=TUNITS.size()){
      cerr << "error:  diff num of entries between lexrm file: " << flexrm << " and tunits voc file: " << ftunits << endl;
      exit(1);
    }
    infile.close();
    if (verbose) cerr << "[" << LEXRM.size() << " score entries]" << endl;

    int nlexrm=split(LEXRM[0]," ").size();
    vector<string> vunklexrm(nlexrm,"0");
    Nlexrm=join(vunklexrm," ");
  }

  /**************************************/
  /********* reading fbfactor/s *********/
  /**************************************/

  for (int f=0;f<fbfactor.size();f++){
    cerr << "\treading " << fbfactor[f] << endl;
    infile.open(fbfactor[f].data(),ifstream::in);
    if (infile.fail()){
      cerr << "error: opening file " << fbfactor[f] << endl;
      exit(1);
    }
    vector<string> bfactor;
    while (getline(infile,line)) {
      bfactor.push_back(line);
    }
    if (bfactor.size()!=TUNITS.size()){
      cerr << "error:  diff num of entries between bilfactor file: " << fbfactor[f] << " and tunits voc file: " << ftunits << endl;
      exit(1);
    }
    BFACTOR.push_back(bfactor);
    infile.close();
    if (verbose) cerr << "[" << bfactor.size() << " factored units]" << endl;
  }
  if (fbfactor.size()>0){
    vector<string> vunkfactors(BFACTOR.size(),"<unk>");
    unkbfactors=join(vunkfactors," ");
    vector<string> vbegfactors(BFACTOR.size(),"<s>");
    begbfactors=join(vbegfactors," ");
    vector<string> vendfactors(BFACTOR.size(),"</s>");
    endbfactors=join(vendfactors," ");
  }

  /**************************************/
  /********* reading fsfactor/s *********/
  /**************************************/

  for (int f=0;f<fsfactor.size();f++){
    cerr << "\treading " << fsfactor[f] << endl;
    infile.open(fsfactor[f].data(),ifstream::in);
    if (infile.fail()){
      cerr << "error: opening file " << fsfactor[f] << endl;
      exit(1);
    }
    vector<string> sfactor;
    while (getline(infile,line)) {
      sfactor.push_back(line);
    }
    if (sfactor.size()!=TUNITS.size()){
      cerr << "error:  diff num of entries between srcfactor file: " << fsfactor[f] << " and tunits voc file: " << ftunits << endl;
      exit(1);
    }
    SFACTOR.push_back(sfactor);
    infile.close();
    if (verbose) cerr << "[" << sfactor.size() << " src factored units]" << endl;
  }
  if (fsfactor.size()>0){
    vector<string> vunkfactors(SFACTOR.size(),"<unk>");
    unksfactors=join(vunkfactors," ");
    vector<string> vbegfactors(SFACTOR.size(),"<s>");
    begsfactors=join(vbegfactors," ");
    vector<string> vendfactors(SFACTOR.size(),"</s>");
    endsfactors=join(vendfactors," ");
  }

  /**************************************/
  /********* reading ftfactor/s *********/
  /**************************************/

  for (int f=0;f<ftfactor.size();f++){
    cerr << "\treading " << ftfactor[f] << endl;
    infile.open(ftfactor[f].data(),ifstream::in);
    if (infile.fail()){
      cerr << "error: opening file " << ftfactor[f] << endl;
      exit(1);
    }
    vector<string> tfactor;
    while (getline(infile,line)) {
      tfactor.push_back(line);
    }
    if (tfactor.size()!=TUNITS.size()){
      cerr << "error:  diff num of entries between trgfactor file: " << ftfactor[f] << " and tunits voc file: " << ftunits << endl;
      exit(1);
    }
    TFACTOR.push_back(tfactor);
    infile.close();
    if (verbose) cerr << "[" << tfactor.size() << " trg factored units]" << endl;
  }
  if (ftfactor.size()>0){
    vector<string> vunkfactors(TFACTOR.size(),"<unk>");
    unktfactors=join(vunkfactors," ");
    vector<string> vbegfactors(TFACTOR.size(),"<s>");
    begtfactors=join(vbegfactors," ");
    vector<string> vendfactors(TFACTOR.size(),"</s>");
    endtfactors=join(vendfactors," ");
  }

  /**************************************/
  /**************************************/
  /**************************************/

  cerr << "[" << difftime(time(NULL),tini) << "''] Done " << endl;

  tini=time(NULL);
  time( &tini );
  current_time = localtime( &tini );
  t=asctime(current_time);
  cerr << "Start filtering Lattices @ " << t.substr(0,t.size()-1) << endl;

  //writing fields description
  cout << "srcTunit" << str_separator << "trgTunit";
  if (fscores.size()>0)  cout << str_separator << "Scores";
  if (flexrm.size()>0)   cout << str_separator << "lexReordering";
  if (fbfactor.size()>0) cout << str_separator << "bilFactors";
  if (fsfactor.size()>0) cout << str_separator << "srcFactors";
  if (ftfactor.size()>0) cout << str_separator << "trgFactors";
  cout << endl;

  vector<string> vgraph;
  string str;
  int nsent=0;
  while(cin) {
    getline(cin, str);
    if (str.compare(str_endofgraph)==0){
      run(nsent,vgraph);
      if (!verbose) if ((nsent+1)%10==0) {if ((nsent+1)%100==0) {cerr << (nsent+1);} else {cerr << ".";}}
      vgraph.clear();
      nsent++;
    }
    else vgraph.push_back(str);
  }
  cerr << endl << "[" << difftime(time(NULL),tini) << "''] Done " << endl;
  return 0;
}

/******************************************************************************************/
/******************************************************************************************/
/******************************************************************************************/

void run(int nsent, vector<string> &vgraph){
  hash_map<string,hash_map<string,bool> > Nwrd2Nwrdid;//Nwrd2Nwrdid{"la maison"}{la@1 maison@2}=true
  hash_map<string,bool> Twrdid;                       //exists translation for 'wrdid'? Twrdid{la@1}=false
  
  //next are data structure for the lattice
  hash_map<int,hash_map<int,string> > arcs;
  int endNode;
  int firstNode=-1;
  
  cout << "<SENTENCE:" << nsent+1 << ">" << endl;
  /***********************/
  /*** compute n-grams ***/
  /***********************/
  // read lattice into arcs (initialize wordsid a Twrdid[wrdid]=false)
  for (int i=0;i<vgraph.size();i++){
    cout << vgraph[i] << endl;
    vector<string> tokens=split(vgraph[i],"\t");
    if (tokens.size()==1){
      endNode=atoi(tokens[0].c_str());
      hash_map<int,string> myhash;
      arcs[endNode]=myhash;
    }
    else if (tokens.size()==4){
      int fromNode=atoi(tokens[0].c_str());
      int toNode=atoi(tokens[1].c_str());
      string wrdid=tokens[2];
      if (arcs.find(fromNode)!=arcs.end() && arcs[fromNode].find(toNode)!=arcs[fromNode].end()){
	cerr << "error: graph with repeated from/to arc!" << endl;
	cerr << vgraph[i] << endl;
	exit(1);
      }
      arcs[fromNode][toNode]=wrdid;
      if (wrdid.compare(str_begsent)!=0 && wrdid.compare(str_endsent)!=0) Twrdid[wrdid]=false;
      if (firstNode==-1) firstNode=fromNode;
    }
    else {
      cerr << "error: unparsed graph line: " << vgraph[i] << endl;
      exit(1);
    }
  }
  // traverse graph to build node2Nwrdid
  hash_map<int,vector<vector<string> > > node2Nwrdid;
  buildNgrams(arcs,Nwrd2Nwrdid,node2Nwrdid,firstNode);
  if (verbose) cerr << " Read [" << Nwrd2Nwrdid.size() << " n-grams]" << endl;
  
  /**************************************/
  /********* building tunits ************/
  /**************************************/   
  int totaltunits=0;
  hash_map<string,hash_map<string,bool > >::iterator it_Nwrd2Nwrdid=Nwrd2Nwrdid.begin(); 
  while (it_Nwrd2Nwrdid!=Nwrd2Nwrdid.end()){
    string stunit=it_Nwrd2Nwrdid->first;
    hash_map<string,vector<int> >::iterator it_stunit2tunitsid=stunit2tunitsid.find(stunit);
    if (it_stunit2tunitsid!=stunit2tunitsid.end()){// exist stunit
      vector<int> vtunitsid=it_stunit2tunitsid->second;
      for (int i=0;i<vtunitsid.size();i++){
	int unitposition=vtunitsid[i];
	string tunit=TUNITS[unitposition];
	vector<string> array=split(tunit,str_separator.c_str());
	string trg=array[1];
	string scores;
	if (fscores.size()>0){
	  scores=SCORES.at(unitposition);
	}
	string lexrm;
	if (flexrm.size()>0){
	  lexrm=LEXRM.at(unitposition);
	}
	string bfactors;	
	if (fbfactor.size()>0){
	  bfactors=BFACTOR[0].at(unitposition);
	  for (int f=1;f<BFACTOR.size();f++) bfactors.append(" ").append(BFACTOR[f].at(unitposition));
	}
	string sfactors;	
	if (fsfactor.size()>0){
	  sfactors=SFACTOR[0].at(unitposition);
	  for (int f=1;f<SFACTOR.size();f++) sfactors.append(" ").append(SFACTOR[f].at(unitposition));
	}
	string tfactors;	
	if (ftfactor.size()>0){
	  tfactors=TFACTOR[0].at(unitposition);
	  for (int f=1;f<TFACTOR.size();f++) tfactors.append(" ").append(TFACTOR[f].at(unitposition));
	}
	hash_map<string,hash_map<string,bool > >::iterator it=Nwrd2Nwrdid.find(stunit);
	hash_map<string,bool >::iterator iter=it->second.begin();
	while (iter!=it->second.end()){ // multiple instances of stunit: {la maison}
	  string src=iter->first;
	  cout << src << str_separator << trg;
	  if (fscores.size()>0) cout << str_separator << scores;
	  if (flexrm.size()>0) cout << str_separator << lexrm;
	  if (fbfactor.size()>0) cout << str_separator << bfactors;
	  if (fsfactor.size()>0) cout << str_separator << sfactors;
	  if (ftfactor.size()>0) cout << str_separator << tfactors;
	  cout << endl;
	  totaltunits++;
	  if (Twrdid.find(src)!=Twrdid.end()){
	    Twrdid[src]=true;//word has translation (used for OOVs)
	    //	    cout << "word: " << src << " has translation!" << endl;
	  }
	  iter++;
	}
      }
    }
    it_Nwrd2Nwrdid++;
  }
  /*********************/
  /*** unknown words ***/
  /*********************/
  int totaloov=0;
  hash_map<string,bool >::iterator it=Twrdid.begin(); 
  while (it!=Twrdid.end()){
    string wrdid=it->first;
    if (it->second==false){ //no translation for word wrdid
      cout << wrdid << str_separator << "<unk>";
      if (fscores.size()>0) cout << str_separator << Nscores;
      if (flexrm.size()>0) cout << str_separator << Nlexrm;
      if (fbfactor.size()>0) cout << str_separator << unkbfactors;
      if (fsfactor.size()>0) cout << str_separator << unksfactors;
      if (ftfactor.size()>0) cout << str_separator << unktfactors;
      cout << endl;
      totaltunits++;
      totaloov++;
    }
    it++;
  }
  /*********************/
  /*** <s> and </s> ****/
  /*********************/
  cout << str_begsent << str_separator << str_begsent;
  if (fscores.size()>0) cout << str_separator << Nscores;
  if (flexrm.size()>0) cout << str_separator << Nlexrm;
  if (fbfactor.size()>0) cout << str_separator << begbfactors;
  if (fsfactor.size()>0) cout << str_separator << begsfactors;
  if (ftfactor.size()>0) cout << str_separator << begtfactors;
  cout << endl;
  cout << str_endsent << str_separator << str_endsent;
  if (fscores.size()>0) cout << str_separator << Nscores;
  if (flexrm.size()>0) cout << str_separator << Nlexrm;
  if (fbfactor.size()>0) cout << str_separator << endbfactors;
  if (fsfactor.size()>0) cout << str_separator << endsfactors;
  if (ftfactor.size()>0) cout << str_separator << endtfactors;
  cout << endl;

  if (verbose) cerr << " Units [" << totaltunits << " tunits " << totaloov << " OOVs]" << endl;
}
											  
/********************************************************************************************/
/********************************************************************************************/
/********************************************************************************************/

void buildNgrams(hash_map<int,hash_map<int,string> > &arcs, hash_map<string,hash_map<string,bool> > &Nwrd2Nwrdid, hash_map<int,vector<vector<string> > > &node2Nwrdid, int currNode){
    //no need to traverse again, node2Nwrdid[currNode] already built!
    if (node2Nwrdid.find(currNode)!=node2Nwrdid.end()) return;

    vector<vector<string> > currNgrams;
    hash_map<int,hash_map<int,string> >::iterator it_arcs=arcs.find(currNode);
    if (it_arcs==arcs.end()) {cerr << "error: invalid currNode:" << currNode << endl; exit(1);}
    hash_map<int,string>::iterator it_currNode=it_arcs->second.begin();
    while(it_currNode!=it_arcs->second.end()){
      int nextNode=it_currNode->first;
      string wrdid=it_currNode->second;
      //update currNgrams
      vector<string> vwrdid;
      vwrdid.push_back(wrdid);
      if (postfix(Nwrd2Nwrdid,vwrdid)) currNgrams.push_back(vwrdid);
      //compute ngrams for nextNode
      buildNgrams(arcs,Nwrd2Nwrdid,node2Nwrdid,nextNode);
      vector<vector<string> > nextNgrams=node2Nwrdid[nextNode];
      //concat wrdid with ngrams for nextNode
      if (nextNgrams.size()>0){
	for (int i=0;i<nextNgrams.size();i++){
	  nextNgrams[i].insert(nextNgrams[i].begin(),wrdid);
	  if (nextNgrams[i].size()>maxs) nextNgrams[i].pop_back();
	}
      }
      //update currNgrams
      for (int i=0;i<nextNgrams.size();i++){
	if (postfix(Nwrd2Nwrdid,nextNgrams[i])) currNgrams.push_back(nextNgrams[i]);
      }      
      it_currNode++;
    }
    node2Nwrdid[currNode]=currNgrams;
}

bool postfix(hash_map<string,hash_map<string,bool> > &Nwrd2Nwrdid, vector<string> &WRDID){
  vector<string> WRD;
  for (int k=0;k<WRDID.size();k++){
    size_t pos=WRDID[k].find(str_wordindexseparator);
    if (WRDID[k].compare(str_begsent)==0 || WRDID[k].compare(str_endsent)==0) pos=WRDID[k].size();
    if (pos==string::npos) {cerr << "error: invalid wrdid: " << WRDID[k] << endl; exit(1);}
    WRD.push_back(WRDID[k].substr(0,pos));
  }
  string wrd=join(WRD," ");
  if (postStunit.find(wrd)==postStunit.end()) return false;
  string wrdid=join(WRDID," ");
  Nwrd2Nwrdid[wrd][wrdid]=true;
  return true;
}
											  
/********************************************************************************************/
/********************************************************************************************/
/********************************************************************************************/

void usage(){
  cerr << "usage: binfiltr -tunits s [-maxs i] [-verbose] [-help] < flattice" << endl;
  cerr << "       flattice   s   : file with input lattices"                  << endl;
  cerr << "       -tunits    s   : translation units"                         << endl;
  cerr << "       -scores    s   : unit scores"                               << endl;
  cerr << "       -lexrm     s   : lexicalized RM scores"                     << endl;
  cerr << "       -bilfactor s   : bilingual factored units"                  << endl;
  cerr << "       -srcfactor s   : src factored units"                        << endl;
  cerr << "       -trgfactor s   : trg factored units"                        << endl;
  cerr << "       -maxs      i   : max size translation units [default 5]"    << endl;
  cerr << "       -verbose       : verbose output             [default false]"<< endl;
  cerr << "       -help          : this help"                                 << endl;
  cerr << endl;
}

vector<string> split(string str,const char *match){
  vector<string> result;
  size_t from=0;
  while (str.find(match,from) != string::npos){
    result.push_back(str.substr(from,str.find(match,from)-from));
    from = str.find(match,from) + strlen(match);
  }
  result.push_back(str.substr(from));
  return result;
}

string join(vector<string> &v,string sep){
  string result=v[0];
  for (int i=1;i<v.size();i++){
    result.append(sep).append(v[i]);
  }
  return result;
}
