#include <stdlib.h>
#include <iostream> 
#include <fstream> 
#include <vector> 
#include <math.h>
#include <string.h>
#include <ext/hash_map>
#include <fst/fstlib.h>

const string str_tplseparator = "|||";
const string str_ruleseparator = "///";
const string str_separator = " ||| ";

const string str_begsent = "<s>";
const string str_endsent = "</s>";
const string str_eps = "<eps>";

using namespace fst;
using namespace std;
using namespace __gnu_cxx;
namespace __gnu_cxx{
  template<> struct hash< std::string >{
    size_t operator()( const std::string& x ) const{return hash< const char* >()( x.c_str() );}
  };
}

void run(int);
void outputLattice(vector<string>,vector<vector<string> > &,vector<vector<string> > &,vector<float> &, int);
void usage();
vector<string> split(string,const char*);
string join(vector<string> &,string);
string frrules, ffrules, finwrd, fintag;
int maxr;
int maxrulesize;
float maxc;
bool verbose, silent, help, rec;
time_t tini;

//to store sentences (word and pos-tagged):
vector<string> sentences_wrd;
vector<string> sentences_tag;
//to store rewrite rules:
hash_map<string,vector<int> > srrule2rrulesid;//srrule2rrulesid{"DET NOM"}=<33,34,35,...> ints are indexes of TRRULES,CRRULES
vector<string> TRRULES;
vector<float> CRRULES;
//to store fuzzy rules:
hash_map<string,vector<int> > sfrule2frulesid;//sfrule2frulesid{"casa blanca"}=<33,34,35,...> ints are indexes of TFRULES,CFRULES
vector<string> TFRULES;
vector<int> CFRULES;

int main(int argc, char** argv){
  cerr << "   ============================================" << endl;
  cerr << "     binrules: BIlingual N-gram rewrite RULES  " << endl;
  cerr << "     by jmcrego@limsi.fr, January 2011          " << endl;
  cerr << "   ============================================" << endl;
  if (argc<=1){
    usage();
    exit(1);
  }
  verbose=false;
  silent=false;
  help=false;
  maxr=9;
  maxrulesize=0;
  maxc=0;
  rec=false;
  /***************************************/
  /********* parsing parameters **********/
  /***************************************/
  int narg=1;
  while (narg<argc){
    string word=argv[narg++];
    if (word.compare("-wrd")==0 && narg<argc)    {finwrd=argv[narg++]; continue;}
    if (word.compare("-tag")==0 && narg<argc)    {fintag=argv[narg++]; continue;}
    if (word.compare("-rrules")==0 && narg<argc) {frrules=argv[narg++]; continue;}
    if (word.compare("-frules")==0 && narg<argc) {ffrules=argv[narg++]; continue;}
    if (word.compare("-maxr")==0 && narg<argc)   {maxr=atoi(argv[narg++]); continue;}
    if (word.compare("-maxc")==0 && narg<argc)   {maxc=atof(argv[narg++]); continue;}
    if (word.compare("-rec")==0)                 {rec=true; continue;}
    if (word.compare("-verbose")==0)             {verbose=true; continue;}
    if (word.compare("-silent")==0)              {silent=true; continue;}
    if (word.compare("-help")==0)                {help=true; continue;}
    cerr << "error: unparsed parameter: " << word << endl;
    usage(); 
    exit(1);
  }

  /***************************************/
  /********* testing parameters **********/
  /***************************************/

  if (help){
      usage();
      exit(1);
  }

  if (finwrd.size()==0){
    cerr << "error: missing -wrd parameter" << endl;
    usage();
    exit(1);
  }

  if (fintag.size()==0){
    cerr << "error: missing -tag parameter" << endl;
    usage();
    exit(1);
  }

  if (frrules.size()==0 && ffrules.size()==0){
    cerr << "error: missing -rrules OR -frules parameter" << endl;
    usage();
    exit(1);
  }

  if (verbose){
    cerr << "Par wrd     : " << finwrd << endl;
    cerr << "Par tag     : " << fintag << endl;
    cerr << "Par rrules  : " << frrules << endl;
    cerr << "Par frules  : " << ffrules << endl;
    cerr << "Par maxr    : " << maxr << endl;
    cerr << "Par maxc    : " << maxc << endl;
    cerr << "Par rec     : " << rec << endl;
    cerr << "Par verbose : " << verbose << endl;
  }

  char hostname[1024];
  hostname[1023] = '\0';
  gethostname(hostname, 1023);
  tini=time(NULL);
  time( &tini );
  tm *current_time = localtime( &tini );
  string t=asctime(current_time);
  cerr << "Start reading global data @ " << t.substr(0,t.size()-1) << " @ " << hostname << endl;

  /**************************************/
  /********* reading words **************/
  /**************************************/

  string line;
  cerr << "\treading " << finwrd << endl;
  ifstream infile;
  infile.open(finwrd.data());
  if (infile.fail()){
    cerr << "error: opening file " << finwrd << endl;
    exit(1);
  }
  while (getline(infile,line)) sentences_wrd.push_back(line);
  infile.close();

  /**************************************/
  /********* reading tags ***************/
  /**************************************/

  cerr << "\treading " << fintag << endl;
  infile.open(fintag.data());
  if (infile.fail()){
    cerr << "error: opening file " << fintag << endl;
    exit(1);
  }
  while (getline(infile,line)) sentences_tag.push_back(line);
  infile.close();

  if (sentences_wrd.size()!=sentences_tag.size()){
    cerr << "error: diff num of sentences in test (wrd/pos) input files" << endl;
    exit(1);
  }
  //  else cerr << "\tRead " << sentences_wrd.size() << " sentences" << endl;

  /**************************************/
  /********* reading frrules ************/
  /**************************************/

  if (frrules.size()>0){
    cerr << "\treading " << frrules << endl;
    infile.open(frrules.data(),ifstream::in);
    if (infile.fail()){
      cerr << "error: opening file " << frrules << endl;
      exit(1);
    }
    string ruleseparator;
    ruleseparator.append(" ").append(str_ruleseparator).append(" ");
    int totalrrules=0;
    while (getline(infile,line)) {
      vector<string> array=split(line,ruleseparator.c_str());
      if (array.size()>=3){
	string srule=array[0];
	vector<string> SRULE=split(srule," ");
	if (SRULE.size()>maxr) continue;
	if (SRULE.size()>maxrulesize) maxrulesize=SRULE.size();
	string trule=array[1];
	string cost=array[2];
	float cst=atof(cost.c_str());
	if (maxc>0 && cst>maxc) continue;
	if (srrule2rrulesid.find(srule)!=srrule2rrulesid.end()) srrule2rrulesid[srule].push_back(TRRULES.size());
	else{
	  vector<int> aux;
	  aux.push_back(TRRULES.size());
	  srrule2rrulesid[srule]=aux;
	}
	TRRULES.push_back(trule);
	CRRULES.push_back(cst);
      }
      else {cerr << "error: rrule '" << line << "' with less than 3 fields" << endl; exit(1);}
      totalrrules++;
    }
    infile.close();
    //  cerr << "\tRead " << totalrrules << " rrules" << endl;
  }

  /**************************************/
  /********* reading ffrules ************/
  /**************************************/

  if (ffrules.size()>0){
    cerr << "\treading " << ffrules << endl;
    infile.open(ffrules.data(),ifstream::in);
    if (infile.fail()){
      cerr << "error: opening file " << ffrules << endl;
      exit(1);
    }
    
    string ruleseparator;
    ruleseparator.append(" ").append(str_ruleseparator).append(" ");
    int totalfrules=0;
    while (getline(infile,line)) {
      vector<string> array=split(line,ruleseparator.c_str());
      if (array.size()>=3){
	string srule=array[0];
	vector<string> SRULE=split(srule," ");
	//	if (SRULE.size()>maxr) continue;
	if (SRULE.size()>maxrulesize) maxrulesize=SRULE.size();
	string trule=array[1];
	string num=array[2];
	int numsent=atoi(num.c_str());
	if (sfrule2frulesid.find(srule)!=sfrule2frulesid.end()) sfrule2frulesid[srule].push_back(TFRULES.size());
	else{
	  vector<int> aux;
	  aux.push_back(TFRULES.size());
	  sfrule2frulesid[srule]=aux;
	}
	TFRULES.push_back(trule);
	CFRULES.push_back(numsent);
      }
      else {cerr << "error: frule '" << line << "' with less than 3 fields" << endl; exit(1);}
      totalfrules++;
    }
    infile.close();
    //  cerr << "\tRead " << totalfrules << " frules" << endl;
  }
  cerr << "[" << difftime(time(NULL),tini) << "''] Done " << endl;

  /**************************************/
  /**************************************/
  /**************************************/

  tini=time(NULL);
  time( &tini );
  current_time = localtime( &tini );
  t=asctime(current_time);
  if (!silent) cerr << "Start computing lattices " << sentences_wrd.size() << " Sents @ " << t.substr(0,t.size()-1) << endl;
  for (int nsent=0;nsent<sentences_wrd.size();nsent++){
    run(nsent);
    if (!verbose && !silent) if ((nsent+1)%10==0) {if ((nsent+1)%100==0) {cerr << (nsent+1);} else {cerr << ".";}}
  }
  if (!silent) cerr << endl << "[" << difftime(time(NULL),tini) << "''] Done " << endl;
  return 0;
}

void run(int nsent){
  if (verbose) cerr << "Sent " << (nsent+1) << endl;
  vector<string> WRD;  //@WRD="la maison blanche"
  vector<string> WRDID;//@WRDID="la@1 maison@2 blanche@3"
  vector<string> TAG;  //@TAG="DET NC ADJ"
  hash_map<string,string >                 Nwrdid2Nwrd;//Nwrdid2Nwrd{la@1 maison@2}="la maison"
  hash_map<string,hash_map<string,bool > > Ntag2Nwrdid;//Ntag2Nwrdid{"DET NC"}{la@1 maison@2}=true
  hash_map<string,hash_map<string,bool > > Nwrd2Nwrdid;//Nwrd2Nwrdid{"la maison"}{la@1 maison@2}=true
  
  /**************************************/
  /********* building linear ************/
  /**************************************/

  WRD=split(sentences_wrd[nsent]," ");
  TAG=split(sentences_tag[nsent]," ");

  if (WRD.size()>0 && WRD[0].size()==0) {WRD.clear();TAG.clear();} //empty line

  if (WRD.size() != TAG.size()){
    cerr << endl << "error: diff num of tokens sent: " << nsent << "!" << endl;
    cerr << "WRD: " << sentences_wrd[nsent] << endl;
    cerr << "TAG: " << sentences_tag[nsent] << endl;
    exit(1);
  }
  
  for (int i=0;i<WRD.size();i++){
    WRDID.push_back(WRD[i]);
    char sid[5]; sprintf(sid,"%d",i+1);
    WRDID[i].append("@").append(sid);
  }

  /*********************/
  /*** linear ngrams ***/
  /*********************/
  for (int from=0;from<WRD.size();from++){
    string seqp=TAG[from];
    string seq=WRD[from];
    string seqid=WRDID[from];
    //Ntag2Nwrdid[seqp][seqid]=true;//not really needed
    Nwrdid2Nwrd[seqid]=seq;
    for (int to=from+1;to<WRD.size()&&to-from<maxrulesize;to++){
      seqp+=" "+TAG[to];
      seq+=" "+WRD[to];
      seqid+=" "+WRDID[to];
      Ntag2Nwrdid[seqp][seqid]=true;
      Nwrd2Nwrdid[seq][seqid]=true; //used for frules
      Nwrdid2Nwrd[seqid]=seq;
    }
  }
  if (verbose) cerr << " Words [" << WRDID.size() << " words " << Nwrdid2Nwrd.size() << " n-grams]" << endl;

  /**************************************************/
  /********* building lexicalized rrules ************/
  /**************************************************/
  
  //these are the rules that match the input sentence
  vector<vector<string> > usedINSTANCEID,usedRINSTANCEID;
  vector<float> usedCOST;

  int totalrrules=0;
  hash_map<string,hash_map<string,bool > >::iterator it_Ntag2Nwrdid=Ntag2Nwrdid.begin(); 
  while (it_Ntag2Nwrdid!=Ntag2Nwrdid.end()){
    string srule=it_Ntag2Nwrdid->first;
    hash_map<string,vector<int> >::iterator it_srrule2rrulesid=srrule2rrulesid.find(srule);
    if (it_srrule2rrulesid!=srrule2rrulesid.end()){// exist srule in global frrules
      vector<int> vtrrulesid=it_srrule2rrulesid->second;
      for (int i=0;i<vtrrulesid.size();i++){
	string trule=TRRULES[vtrrulesid[i]];
	float cost=CRRULES[vtrrulesid[i]];
	hash_map<string,bool >::iterator iter=it_Ntag2Nwrdid->second.begin();
	while (iter!=it_Ntag2Nwrdid->second.end()){
	  string instanceid=iter->first;//'la@1 maison@2' 
	  vector<string> INSTANCEID=split(instanceid," ");
	  vector<string> TRULE=split(trule," ");
	  int prevind=atoi(TRULE[0].data());
	  string rinstanceid=INSTANCEID[prevind];
	  bool monotonic=true;
	  for (int i=1;i<TRULE.size();i++){
	    int currind=atoi(TRULE[i].data());
	    rinstanceid.append(" ").append(INSTANCEID[currind]);
	    if (currind!=prevind+1) monotonic=false;
	    prevind=currind;
	  }
	  //keep a record of the lexicalized rule used
	  if (!monotonic){
	    vector<string> RINSTANCEID=split(rinstanceid," ");
	    usedINSTANCEID.push_back(INSTANCEID);
	    usedRINSTANCEID.push_back(RINSTANCEID);
	    usedCOST.push_back(cost);
	    totalrrules++;
	  }
	  iter++;
	}
      }
    }
    it_Ntag2Nwrdid++;
  }
  if (verbose) cerr << " Rules [" << totalrrules << " rrules]" << endl;


  /*******************************************/
  /********* building fuzzy rules ************/
  /*******************************************/
  
  int totalfrules=0;
  hash_map<string,hash_map<string,bool > >::iterator it_Nwrd2Nwrdid=Nwrd2Nwrdid.begin(); 
  while (it_Nwrd2Nwrdid!=Nwrd2Nwrdid.end()){
    string srule=it_Nwrd2Nwrdid->first;
    hash_map<string,vector<int> >::iterator it_sfrule2frulesid=sfrule2frulesid.find(srule);
    if (it_sfrule2frulesid!=sfrule2frulesid.end()){// exist srule in global ffrules
      vector<int> vtfrulesid=it_sfrule2frulesid->second;
      for (int i=0;i<vtfrulesid.size();i++){
	int numsent=CFRULES[vtfrulesid[i]];
	if (numsent!=nsent+1) continue;
	string trule=TFRULES[vtfrulesid[i]];
	hash_map<string,bool >::iterator iter=it_Nwrd2Nwrdid->second.begin();
	while (iter!=it_Nwrd2Nwrdid->second.end()){
	  string instanceid=iter->first;//'la@1 maison@2' 
	  vector<string> INSTANCEID=split(instanceid," ");
	  vector<string> TRULE=split(trule," ");

	  string fuzzystring="";
	  size_t pos=TRULE[0].find("<fuzzy");
	  if (pos!=string::npos){fuzzystring=TRULE[0].substr(pos);TRULE[0]=TRULE[0].substr(0,pos);}
	  int prevind=atoi(TRULE[0].data());
	  string rinstanceid;
	  pos=INSTANCEID[prevind].rfind("@");
	  if (pos!=string::npos){
	    string index=INSTANCEID[prevind].substr(pos);
	    string word=INSTANCEID[prevind].substr(0,pos);
	    rinstanceid.append(word).append(fuzzystring).append(index);
	  }
	  else {cerr << "error: unparsed token: " << rinstanceid << endl;exit(1);}
	  bool monotonic=true;
	  for (int i=1;i<TRULE.size();i++){
	    string fuzzystring="";
	    size_t pos=TRULE[i].find("<fuzzy");
	    if (pos!=string::npos){fuzzystring=TRULE[i].substr(pos);TRULE[i]=TRULE[i].substr(0,pos);}
	    int currind=atoi(TRULE[i].data());
	    pos=INSTANCEID[currind].rfind("@");
	    if (pos!=string::npos){
	      string index=INSTANCEID[currind].substr(pos);
	      string word=INSTANCEID[currind].substr(0,pos);
	      rinstanceid.append(" ").append(word).append(fuzzystring).append(index);
	    }
	    else {cerr << "error: unparsed token: " << rinstanceid << endl;exit(1);}
	    if (currind!=prevind+1) monotonic=false;
	    prevind=currind;
	  }
	  if (verbose) cout << "FuzzyRULE: " << instanceid << " >>> " << rinstanceid << endl; 
	  //keep a record of the lexicalized rule used
	  vector<string> RINSTANCEID=split(rinstanceid," ");
	  usedINSTANCEID.push_back(INSTANCEID);
	  usedRINSTANCEID.push_back(RINSTANCEID);
	  usedCOST.push_back((float)0);
	  totalfrules++;
	  iter++;
	}
      }
    }
    it_Nwrd2Nwrdid++;
  }
  if (verbose) cerr << " Rules [" << totalfrules << " frules]" << endl;

  outputLattice(WRDID,usedINSTANCEID,usedRINSTANCEID,usedCOST,nsent);
}

/********************************************************************************************/
/********************************************************************************************/
/********************************************************************************************/

void outputLattice(vector<string> WRDID,vector<vector<string> > &usedINSTANCEID,vector<vector<string> > &usedRINSTANCEID,vector<float> &usedCOST,int nsent){
  hash_map<string,int> Hsymbol;//Hsymbol{la@1}=1
  Hsymbol[str_eps]=Hsymbol.size();
  WRDID.push_back(str_endsent);
  WRDID.insert(WRDID.begin(),str_begsent);
  
  /*****************************/
  /******* linear fst **********/
  /*****************************/
  fst::StdVectorFst myfsa;
  StdArc::StateId prev=myfsa.AddState();
  myfsa.SetStart(prev);
  for (int i=0;i<WRDID.size();i++){
    string label=WRDID[i];
    int64 labelId=(int64)Hsymbol.size();
    Hsymbol[label]=labelId;
    StdArc::StateId curr=myfsa.AddState();
    myfsa.AddArc(prev,StdArc(labelId,labelId,fst::StdArc::Weight::One(),curr));
    prev=curr;
  }
  myfsa.SetFinal(prev,fst::StdArc::Weight::One());

  if (verbose) cout << "LINEAR: " << join(WRDID," ") << endl;
  /********** each rule is composed (largers first) ************/
  for (int mysize=2;mysize<=maxrulesize;mysize++){
    int currsize=mysize;
    if (rec) currsize=maxrulesize+2-mysize;
    for (int i=0;i<usedINSTANCEID.size();i++){
      vector<string> SRC=usedINSTANCEID[i];
      if (SRC.size()!=currsize) continue;
      vector<string> TRG=usedRINSTANCEID[i];
      float cost=usedCOST[i];
      //      cerr << "Rule(" << nsent+1 << "): " << join(SRC," ") << " ||| " << join(TRG," ") << endl;
      /****************************/
      /******* rule fst ***********/
      /****************************/
      StdVectorFst myfst;
      StdArc::StateId initial=myfst.AddState();
      myfst.SetStart(initial);
      myfst.SetFinal(initial,fst::StdArc::Weight::One());
      /******* monotonic path *********/
      for (int j=0;j<WRDID.size();j++){
	string label=WRDID[j];
	int64 labelId=(int64)Hsymbol.find(label)->second;
	myfst.AddArc(initial,StdArc(labelId,labelId,fst::StdArc::Weight::One(),initial));
      }
      /******* rule *********/
      StdArc::StateId prev=initial;
      for (int k=0;k<SRC.size()-1;k++){
	string ilabel=SRC[k];  
	int64 ilabelId=(int64)Hsymbol.find(ilabel)->second;
	string olabel=TRG[k];  
	int64 olabelId;
	//add target symbol if it does not exist
	if (Hsymbol.find(olabel)==Hsymbol.end()){
	  olabelId=(int64)Hsymbol.size();
	  Hsymbol[olabel]=olabelId;
	  WRDID.push_back(olabel);
	}
	else olabelId=(int64)Hsymbol.find(olabel)->second;
	if (verbose) cout << "\tSRC: " << ilabel << "(" << ilabelId << ") \tTRG: " << olabel << "(" << olabelId << ")" << endl;
	StdArc::StateId curr=myfst.AddState();
	float mycost=cost;
	if (k>0) mycost=fst::StdArc::Weight::One().Value();
	myfst.AddArc(prev,StdArc(ilabelId,olabelId,mycost,curr));
	prev=curr;
      }
      string ilabel=SRC.back();  
      int64 ilabelId=(int64)Hsymbol.find(ilabel)->second;
      string olabel=TRG.back();  
      int64 olabelId;
      //add target symbol if it does not exist
      if (Hsymbol.find(olabel)==Hsymbol.end()){
	olabelId=(int64)Hsymbol.size();
	Hsymbol[olabel]=olabelId;
	WRDID.push_back(olabel);
      }
      else olabelId=(int64)Hsymbol.find(olabel)->second;
      if (verbose) cout << "\tSRC: " << ilabel << "(" << ilabelId << ") \tTRG: " << olabel << "(" << olabelId << ")" << endl;
      myfst.AddArc(prev,StdArc(ilabelId,olabelId,fst::StdArc::Weight::One(),initial));
      ArcSort(&myfst,StdILabelCompare());
      /*****************************/
      /******* composition *********/
      /*****************************/
      StdVectorFst mycomposed;
      Compose(myfsa, myfst, &mycomposed);
      myfsa=mycomposed;
    }
  }
  Project(&myfsa,PROJECT_OUTPUT); 
  StdVectorFst mydet;
  Determinize(myfsa,&mydet);
  Minimize(&mydet);
  TopSort(&mydet);
  /*****************************/
  /******* print fst ***********/
  /*****************************/
  for (StateIterator<StdVectorFst> siter(mydet); !siter.Done(); siter.Next()){
    StdArc::StateId sid = siter.Value();
    if (mydet.Final(sid) != StdArc::Weight::Zero()) cout << sid << endl;
    for (ArcIterator<StdVectorFst> aiter(mydet,sid); !aiter.Done(); aiter.Next()){
      int64 labelId=aiter.Value().ilabel;
      string label;
      if (labelId==0 || labelId>WRDID.size()) {cerr << "error: incorrect labelId '" << labelId << "' in input lattice!" << endl; exit(1);}
      else label=WRDID[labelId-1];
      cout << sid << "\t" << aiter.Value().nextstate << "\t" << label << "\t" << aiter.Value().weight.Value() << endl;
    } 
  }
  cout << "<END>" << endl;
}

/********************************************************************************************/
/********************************************************************************************/
/********************************************************************************************/

void usage(){
  cerr << "usage: binrules -wrd s -tag s -rrules s [-maxr i] [-maxc f] [-rec] [-verbose] [-help]" << endl;
  cerr << "  Input files:"                                                         << endl;
  cerr << "       -wrd       s   : file to translate (word forms)"                 << endl;
  cerr << "       -tag       s   : file to translate (POS tags)"                   << endl;
  cerr << "       -rrules    s   : reordering rules"                               << endl;
  cerr << "  Rules/Units filtering:"                                               << endl;
  cerr << "       -maxc      f   : max cost reordering rules       [default 0:all]"    << endl;
  cerr << "       -maxr      i   : max size reordering rules       [default 9]"    << endl;
  cerr << "  Other:"                                                               << endl;
  cerr << "       -rec           : apply recursive reorderings     [default false]"<< endl;
  cerr << "       -verbose       : verbose output                  [default false]"<< endl;
  cerr << "       -silent        : silent output                   [default false]"<< endl;
  cerr << "       -help          : this help"                                      << endl << endl;
}

vector<string> split(string str,const char *match){
  vector<string> result;
  size_t from=0;
  while (str.find(match,from) != string::npos){
    result.push_back(str.substr(from,str.find(match,from)-from));
    from = str.find(match,from) + strlen(match);
  }
  result.push_back(str.substr(from));
  return result;
}

string join(vector<string> &v,string sep){
  string result=v[0];
  for (int i=1;i<v.size();i++){
    result.append(sep).append(v[i]);
  }
  return result;
}
