#include <sstream>
#include <fstream>
#include <iostream>
#include <string>
#include <vector>
#include <time.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#if __GNUC__>2
#include <ext/hash_map>
using __gnu_cxx::hash_map;
#else
#include <hash_map>
#endif
 
#define MODUL 1000000                                                                                  
using namespace std;

/**************************************************************************/

#ifndef HASHPAIRUINT
#define HASHPAIRUINT
class hashfpairuint : public unary_function<pair<unsigned int,unsigned int>, size_t>{
 public:
  size_t operator()(pair<unsigned int,unsigned int> const &p) const{
    unsigned int res=0;
    res += (size_t) p.first;
    res += (size_t) pow((double)p.second,double(2));
    return res%MODUL;
  }
};
class hasheqpairuint{
 public:
  bool operator()(pair<unsigned int,unsigned int> const &p1, pair<unsigned int,unsigned int> const &p2) const{
    return p1==p2;
  }
};
#endif

/**************************************************************************/

#ifndef HASHSTRING
#define HASHSTRING
class hashfstring : public unary_function<string, size_t>{
 public:
  size_t operator()(string const &s) const{
    unsigned int res=0;
    for (int i=0;i<s.size();i++){
      res += (size_t) (unsigned int) pow((double)s.at(i), double((i%6)+1));
    }
    return res%MODUL;
  }
};
class hasheqstring{
 public:
  bool operator()(string const &s1, string const &s2) const{
    return s1.compare(s2)==0;
  }
};
#endif

/**************************************************************************/
/**** GLOBAL DECLARATIONS *************************************************/
/**************************************************************************/
void usage();
void read_params(int, char **);
void compute_lex(string,unsigned int, ofstream &);
void introduce_lex_st(string, unsigned int);
void introduce_lex_ts(string, unsigned int);

string table;
string lex_st;
string lex_ts;
string models;
const float logTen=log(10*1.0);
const float maxLogCost=-40*logTen;
unsigned int srcNullId, trgNullId;
bool verbose,lc;

vector<string> dict_ws;
hash_map<string,unsigned int,hashfstring,hasheqstring> jash_ws;
hash_map<string,unsigned int,hashfstring,hasheqstring>::iterator it_jash_ws;

vector<string> dict_wt;
hash_map<string,unsigned int,hashfstring,hasheqstring> jash_wt;
hash_map<string,unsigned int,hashfstring,hasheqstring>::iterator it_jash_wt;

hash_map<pair<unsigned int,unsigned int>,float,hashfpairuint,hasheqpairuint> jash_lexst_prob;
hash_map<pair<unsigned int,unsigned int>,float,hashfpairuint,hasheqpairuint>::iterator it_jash_lexst_prob;
hash_map<pair<unsigned int,unsigned int>,float,hashfpairuint,hasheqpairuint> jash_lexts_prob;
hash_map<pair<unsigned int,unsigned int>,float,hashfpairuint,hasheqpairuint>::iterator it_jash_lexts_prob;

/**************************************************************************/
/********** MAIN **********************************************************/
/**************************************************************************/
int main(int argc, char **argv){
  cerr << "   =========================================================" << endl;
  cerr << "     feat_lex: computing lexicon probabilities over tuples  " << endl;
  cerr << "   =========================================================" << endl;
  verbose=false;
  lc=false;
  read_params(argc,argv);
  /***********************************************/
  cerr << "[feat_lex] reading Lex models" << endl;
  /***********************************************/

  ifstream flex_st(lex_st.data());
  if( !flex_st.is_open() ) {
    cerr << "ERROR: Opening lex_st file:" << lex_st << endl;
    exit(0);
  }
  unsigned int lineNb=0;
  string s;
  getline( flex_st, s );
  while( !flex_st.eof()  ){
    lineNb++;
    introduce_lex_st(s,lineNb);
    if (lineNb%10000==0) {if (lineNb%100000==0) cerr << lineNb; else cerr << ".";}
    getline( flex_st, s );
  }
  flex_st.close();
  cerr << " read s2t" << endl;

  ifstream flex_ts(lex_ts.data());
  if( !flex_ts.is_open() ) {
    cerr << "ERROR: Opening lex_ts file:" << lex_ts << endl;
    exit(0);
  }
  lineNb=0;
  getline( flex_ts, s );
  while( !flex_ts.eof()  ){
    lineNb++;
    introduce_lex_ts(s,lineNb);
    if (lineNb%10000==0) {if (lineNb%100000==0) cerr << lineNb; else cerr << ".";}
    getline( flex_ts, s );
  }
  flex_ts.close();
  cerr << " read t2s" << endl;

  /***********************************************/
  cerr << "[feat_lex] reading table file, computing models" << endl;
  /***********************************************/
  ofstream modelsFile(models.c_str());
  ifstream fin(table.data());
  if( !fin.is_open() ) {
    cerr << "ERROR: Opening table file:" << table << endl;
    exit(0);
  }
  lineNb=0;
  getline( fin, s );
  while( !fin.eof()  ){
    lineNb++;
    compute_lex(s,lineNb,modelsFile);
    if (lineNb%10000==0) {if (lineNb%100000==0) cerr << lineNb; else cerr << ".";}
    getline( fin, s );
  }
  fin.close();
  cerr << " read" << endl;
  cerr << "[feat_lex] all done." << endl;  
}

/**************************************************************************/
/**************************************************************************/
/**************************************************************************/

void introduce_lex_st(string line, unsigned int lineNb){
    string swstr,twstr;
    float prob;
    unsigned int swid,twid;
    istringstream iss(line);
    iss >> swstr >> twstr >> prob;

    it_jash_ws=jash_ws.find(swstr);
    if ( it_jash_ws == jash_ws.end() ) {
      //add the word into the hash
      jash_ws[swstr]=dict_ws.size();
      swid=dict_ws.size();
      dict_ws.push_back(swstr);
    }
    else swid = it_jash_ws->second;

    it_jash_wt=jash_wt.find(twstr);
    if ( it_jash_wt == jash_wt.end() ) {
      //add the word into the hash
      jash_wt[twstr]=dict_wt.size();
      twid=dict_wt.size();
      dict_wt.push_back(twstr);
    }
    else twid = it_jash_wt->second;

    jash_lexst_prob[make_pair(swid,twid)]=prob;

    //    if (verbose) cout << "added lex_st(" << swstr << "," << twstr << ")=" << prob << endl;
    return;
}

void introduce_lex_ts(string line, unsigned int lineNb){
    string swstr,twstr;
    float prob;
    unsigned int swid,twid;
    istringstream iss(line);
    iss >> twstr >> swstr >> prob;

    it_jash_ws=jash_ws.find(swstr);
    if ( it_jash_ws == jash_ws.end() ) {
      //add the word into the hash
      jash_ws[swstr]=dict_ws.size();
      swid=dict_ws.size();
      dict_ws.push_back(swstr);
    }
    else swid = it_jash_ws->second;

    it_jash_wt=jash_wt.find(twstr);
    if ( it_jash_wt == jash_wt.end() ) {
      //add the word into the hash
      jash_wt[twstr]=dict_wt.size();
      twid=dict_wt.size();
      dict_wt.push_back(twstr);
    }
    else twid = it_jash_wt->second;

    jash_lexts_prob[make_pair(twid,swid)]=prob;

    //    if (verbose) cout << "added lex_ts(" << twstr << "," << swstr << ")=" << prob << endl;
    return;
}

void compute_lex(string s,unsigned int lineNb, ofstream &file){
  //input is like:
  //src1 src1 ... srcJ ||| trg1 trg2 ... trgI ||| pos1 pos2 ... posI

  string::size_type pos1=s.find(" ||| ",0);
  if (pos1==string::npos){cerr << "ERROR: parsing line num=" << lineNb << endl; exit(0);}
  string::size_type pos2=s.find(" ||| ",pos1+1);
  string src_side,trg_side;
  src_side=s.substr(0,pos1);
  if (pos2==string::npos) trg_side=s.substr(pos1+5);
  else trg_side=s.substr(pos1+5,pos2-pos1-5);
  //lowercase if needed

  if (verbose) cout << "compute_lex(" << src_side << "|||" << trg_side << ")" << endl;
  vector<unsigned int> vecj;
  istringstream inStrSrc(src_side);
  while (inStrSrc){
    string word;
    inStrSrc >> word;
    if (!inStrSrc) continue;
    it_jash_ws=jash_ws.find(word);
    if ( it_jash_ws == jash_ws.end() ) {
      cerr << "ERROR: source word:" << word << " not found in vocab! cost set to " << maxLogCost << " line num=" << lineNb << ": " << s << endl; 
      //      exit(0);
      file << maxLogCost << " " << maxLogCost << endl; //positive costs    
      return;
    }
    vecj.push_back(it_jash_ws->second);
  }
  vector<unsigned int>::size_type J=vecj.size();
  float J_f = 1.0*J;
  
  vector<unsigned int> veci;
  istringstream inStrTrg(trg_side);
  while (inStrTrg){
    string word;
    inStrTrg >> word;
    if (!inStrTrg) continue;
    it_jash_wt=jash_wt.find(word);
    if ( it_jash_wt == jash_wt.end() ) {
      cerr << "ERROR: target word:" << word << " not found in vocab! cost set to " << maxLogCost << " line num=" << lineNb << ": " << s << endl; 
      //exit(0);
      file << maxLogCost << " " << maxLogCost << endl; //positive costs
      return;
    }
    veci.push_back(it_jash_wt->second);
  }
  vector<unsigned int>::size_type I=veci.size();
  float I_f = 1.0*I;

  if (verbose) cout << " J=" << J << " I=" << I << endl;
  
  //************************************
  // calculate LEX probability src->trg
  //************************************
  bool includeTrgNullWd;
  if (J>1){includeTrgNullWd=true;}
  else{includeTrgNullWd=false;}
  float sum_j_log=0;
  int numj=0; //useless 
  for (vector<unsigned int>::const_iterator j=vecj.begin();j<vecj.end();++j){
    numj++;
    double sum_i=0;
    if (includeTrgNullWd){
      sum_i=jash_lexst_prob[make_pair(*j,trgNullId)];
      if (verbose) cout << "   sum+=prob(" << dict_ws[*j] << "|||NULL) sum:" << sum_i << endl;
    }
    int numi=0;
    for (vector<unsigned int>::const_iterator i=veci.begin();i<veci.end();++i){
      numi++;
      sum_i += jash_lexst_prob[make_pair(*j,*i)];
      if (verbose) cout << "   j:" << numj << ",i:" << numi << " sum+=prob(" << dict_ws[*j] << "|||" << dict_wt[*i] << ") sum:" << sum_i << endl;
    }
    if (sum_i>0){sum_j_log += log(sum_i);}
    else {sum_j_log += maxLogCost;}
    if (verbose) cout << "  sum_j_log=" << sum_j_log << endl;
  }
  float logPower;
  if (includeTrgNullWd){logPower=J_f*log(I_f+1);}
  else {
    logPower = J_f*log(I_f);
  }
  if (verbose) cout << " logPower=" << logPower << endl;
  float lex_cost_st;
  lex_cost_st = (sum_j_log - logPower)/logTen;
  if (verbose) cout << " cost=" << lex_cost_st << endl;
  
  //************************************
  // calculate LEX probability trg->src
  //************************************
  if (I>1){includeTrgNullWd=true;}
  else{includeTrgNullWd=false;}
  float sum_i_log=0;
  for (vector<unsigned int>::const_iterator i=veci.begin();i<veci.end();++i){
    double sum_j=0;
    if (includeTrgNullWd){sum_j=jash_lexts_prob[make_pair(*i,srcNullId)];}
    else {sum_j=0;}
    for (vector<unsigned int>::const_iterator j=vecj.begin();j<vecj.end();++j){
      it_jash_lexts_prob=jash_lexts_prob.find(make_pair(*i,*j));
      if (it_jash_lexts_prob==jash_lexts_prob.end()){sum_j += jash_lexst_prob[make_pair(*j,*i)];}
      else{sum_j += it_jash_lexts_prob->second;}
    }
    if (sum_j>0){sum_i_log += log(sum_j);}
    else {sum_i_log += maxLogCost;}
  }
  if (includeTrgNullWd){logPower = I_f*log(J_f+1);}
  else {
    logPower = I_f*log(J_f);
  }
  float lex_cost_ts;
  lex_cost_ts = (sum_i_log - logPower)/logTen;
  
  file << -lex_cost_st << " " << -lex_cost_ts << endl; //positive costs
  return;
}

void read_params(int argc, char **argv){
  table="";
  lex_st="";
  lex_ts="";
  models="";
  for (int i=1;i<argc;i+=2){
    if (strcmp(argv[i],"-table")==0)    {table = argv[i+1];}
    else if (strcmp(argv[i],"-i_st")==0){lex_st = argv[i+1];}
    else if (strcmp(argv[i],"-i_ts")==0){lex_ts = argv[i+1];}
    else if (strcmp(argv[i],"-o")==0)   {models = argv[i+1];}
    else if (strcmp(argv[i],"-verbose")==0)   {verbose = true;}
    else if (strcmp(argv[i],"-lc")==0)   {lc = true;}
    else{cerr << "ERROR: invalid parameter " << argv[i] << endl;usage();}
  }
  if (table.size()==0 || lex_st.size()==0 || lex_ts.size()==0 || models.size()==0){
    cerr << "ERROR: Additional parameters needed!" << endl; usage();
  }
}

void usage(){
  cerr<<"USAGE: feat_lex -table unitsTable -i_st LexModel_src2trg -i_ts LexModel_trg2src -o outModelFile [-verbose]\n";
  cerr<<"  -table f: file with translation units (to read unigrams)\n";
  cerr<<"  -i_st  f: trg2src.actual.ti.final file with IBM1 probabilities (src trg prob)\n";
  cerr<<"  -i_ts  f: src2trg.actual.ti.final file with IBM1 probabilities (trg src prob)\n";
  cerr<<"  -o     f: output models file (format: costLex1 costLex2)\n";
  cerr<<"  -verbose: verbose output\n";
  cerr<<"\n";
  exit(0);
}
