#include <string.h>
#include "Auxiliar.h"

#include <fst/fstlib.h>
#include <fst/fst.h>
using namespace fst;

/**********************************************************************************/
/**********************************************************************************/
/**********************************************************************************/

void usage();
int buildNbest(vector<string> &,int);

int bestn=300;
bool raw1best=false;
bool moses=false;

int main(int argc, char** argv){
  cerr << "   ============================================================" << endl;
  cerr << "     gph2nbst: converts bincoder output graph into Nbest list" << endl;
  cerr << "     (ZMERT format | Moses format)" << endl;
  cerr << "     jmcrego[at]limsi[dot]fr (June 2011)" << endl;
  cerr << "   ============================================================" << endl;
  string list("");
  for (int i=1;i<argc;i++) list.append(argv[i]).append(" ");
  istringstream inStr(list);
  while (inStr){
    string word;
    inStr >> word;
    if (!inStr) continue;
    if (word.compare("-nbest")==0){
      inStr >> word; 
      bestn=atoi(word.c_str()); 
      continue;
    }
    if (word.compare("-raw1best")==0){
      raw1best=true; 
      continue;
    }
    if (word.compare("-moses")==0){
      moses=true; 
      continue;
    }
    cerr << "error: unparsed parameter " << word << endl;
    usage();
  }
  if (raw1best) bestn=1;

  cerr << "nbest    : " << bestn << endl;
  cerr << "moses    : " << moses << endl;
  cerr << "raw1best : " << raw1best << endl;

  int nsent=0;
  string str;
  vector<string> vgraph;
  while(cin) {
    getline(cin, str);
    if (str.compare(str_endofgraph)==0){
      nsent++;
      //      if (nsent%10==0) cerr << nsent; else cerr << ".";
      cerr << ".";
      int noutput=buildNbest(vgraph,nsent);
      if (noutput<bestn) cerr << "(" << nsent << ":" << noutput << ")";
      vgraph.clear();
    }
    else {vgraph.push_back(str);}
  }
  return 0;
}

int buildNbest(vector<string> &vgraph, int nsent){
  hash_map<int, hash_map<int, pair<string,float> > > node2node2wordcost;
  hash_set<int> Finals;
  int Initial=-1;
  int lastNode=-1;
  for (int k=0;k<vgraph.size();k++){
    stringstream ssline(vgraph[k]);
    vector<string> tokens;
    string str;
    while (ssline >> str) tokens.push_back(str);
    if (tokens.size()==4){
      int nodefrom=atoi(tokens[0].c_str());
      if (lastNode<nodefrom) lastNode=nodefrom;
      if (Initial==-1) Initial=nodefrom;
      int nodeto=atoi(tokens[1].c_str());
      if (lastNode<nodeto) lastNode=nodeto;
      string word=tokens[2];
      float cost=atof(tokens[3].c_str());
      if (node2node2wordcost.find(nodefrom)!=node2node2wordcost.end() && node2node2wordcost[nodefrom].find(nodeto)!=node2node2wordcost[nodefrom].end()){
	cerr << "error: repeated graph arc (" << nodefrom << "," << nodeto << ")" << endl;
	exit(1);
      }
      node2node2wordcost[nodefrom][nodeto]=make_pair(word,cost);
    }
    else if (tokens.size()==1){
      Finals.insert(atoi(tokens[0].c_str()));
    }
    else {
      cerr << "error: unparsed output graph line: " << vgraph[k] << endl;
      exit(1);
    }
  }
  //------------------
  // build fst
  //------------------
  fst::SymbolTable* isyms=new fst::SymbolTable("inputsym");
  isyms->AddSymbol(str_eps, 0);
  int next_isym=1;
  fst::SymbolTable* osyms=new fst::SymbolTable("outputsym");
  osyms->AddSymbol(str_eps, 0);
  int next_osym=1;
  StdVectorFst fst;
  for (int k=0;k<=lastNode;k++) fst.AddState();
  hash_map<int, hash_map<int, pair<string,float> > >::iterator it_nodefrom=node2node2wordcost.begin();
  while (it_nodefrom!=node2node2wordcost.end()){
    int nodefrom=(int64)it_nodefrom->first;
    hash_map<int, pair<string,float> >::iterator it_nodeto=it_nodefrom->second.begin();
    while (it_nodeto!=it_nodefrom->second.end()){
      //--------------------
      //extract graph arc info
      //--------------------
      int nodeto=(int64)it_nodeto->first;
      string arc=it_nodeto->second.first;
      float cost=it_nodeto->second.second;
      vector<string> tokens=split(arc,str_tplseparator.c_str());
      if (tokens.size()!=3) {cerr << "error: unparsed arc in graph: " << arc << endl; exit(1);}
      vector<string> TARGET=split(tokens[1],str_wrdseparator.c_str());
      if (TARGET.size()==1 && TARGET[0].compare(str_null)==0) TARGET[0]=str_eps;
      //--------------------
      //update fst i/o symbols
      //--------------------
      if (isyms->Find(arc) == SymbolTable::kNoSymbol) isyms->AddSymbol(arc,next_isym++);
      for (int k=0;k<TARGET.size();k++) if (osyms->Find(TARGET[k]) == SymbolTable::kNoSymbol) osyms->AddSymbol(TARGET[k],next_osym++);
      if (TARGET.size()==1) fst.AddArc(nodefrom, StdArc(isyms->Find(arc),osyms->Find(TARGET[0]),cost,nodeto));
      else{
	int64 nodeprev;
	int64 nodenext=fst.AddState();
	fst.AddArc(nodefrom, StdArc(isyms->Find(arc),osyms->Find(TARGET[0]),cost,nodenext));
	for (int k=1;k<TARGET.size()-1;k++){
	  nodeprev=nodenext;
	  nodenext=fst.AddState();
	  fst.AddArc(nodeprev, StdArc(isyms->Find(str_eps),osyms->Find(TARGET[k]),StdArc::Weight::One(),nodenext));
	}
	fst.AddArc(nodenext, StdArc(isyms->Find(str_eps),osyms->Find(TARGET.back()),StdArc::Weight::One(),nodeto));
      }
      if (nodefrom==Initial) fst.SetStart(nodefrom);  
      if (Finals.find(nodeto)!=Finals.end()) fst.SetFinal(nodeto,StdArc::Weight::One());  
      //----------	
      it_nodeto++;
    }
    it_nodefrom++;
  }
  Connect(&fst);
  //-------------------
  //output nbest
  //-------------------
  RmEpsilon(&fst);
  StdVectorFst fstnbest;
  ShortestPath(fst, &fstnbest, bestn*5);//, false, false, StdArc::Weight::Zero(), kNoStateId);
  Project(&fstnbest,PROJECT_INPUT);
  RmEpsilon(&fstnbest);
  fstToStrings nbest(&fstnbest,isyms);
  int noutput=0;
  hash_set<string> Output;
  for (int i=0; i<nbest.size()&&noutput<bestn; i++){
    vector<string> words;
    vector<string> names;
    vector<float> costs;
    vector<string> arcstotalscore=split(nbest.n(i),str_separator.c_str());
    vector<string> tokens=split(arcstotalscore[0]," ");
    float totalscore=-atof(arcstotalscore[1].c_str());
    for (int k=0;k<tokens.size();k++){
      vector<string> srctrgcosts=split(tokens[k],str_tplseparator.c_str());
      if (srctrgcosts.size()!=3){
	cerr << "error: unparsed token in output graph!" << endl;
	exit(1);
      }
      // target words
      vector<string> WORDS=split(srctrgcosts[1],str_wrdseparator.c_str());
      for (int j=0;j<WORDS.size();j++) 
	if (WORDS[j].compare(str_null)!=0 && WORDS[j].compare(str_begsent)!=0 && WORDS[j].compare(str_endsent)!=0) words.push_back(WORDS[j]);
      // scores
      vector<string> namecost=split(srctrgcosts[2],",");
      if (costs.size()==0){
	for (int j=0;j<namecost.size();j++){
	  vector<string> NAMECOST=split(namecost[j],"=");
	  names.push_back(NAMECOST[0]);
	  costs.push_back(-atof(NAMECOST[1].c_str()));
	}
      }
      else{
	for (int j=0;j<namecost.size();j++){
	  vector<string> NAMECOST=split(namecost[j],"=");
	  string name=NAMECOST[0];
	  if (name.compare(names[j])!=0){
	    cerr << "error: diff num of scores!" << endl;
	    exit(1);
	  }
	  costs[j]+=-atof(NAMECOST[1].c_str());
	}
      }
    }
    //check if repeated hypothesis
    if (Output.find(join(words," "))!=Output.end()) continue;
    else Output.insert(join(words," "));

    if (costs.size()!=names.size()){
      cerr << "error: diff num of scores/names!" << endl;
      exit(1);
    }
    if (raw1best) cout << join(words," ") << endl;
    else{
      if (moses){
	vector<string> Costs;
	for (int k=0;k<costs.size();k++){
	  ostringstream data;
	  data.precision(floatprecision);
	  data << names[k] <<  ": " << costs[k];
	  Costs.push_back(data.str());
	}
	cout << nsent-1 << str_separator << join(words," ") << str_separator << join(Costs," ") << str_separator << totalscore << endl;
      }
      else{// used for ZMERT 
	cout << nsent-1 << str_separator << join(words," ") << str_separator << join(costs," ") << str_separator << totalscore << endl;
      }
    }
    noutput++;
  }
  return noutput;
}

void usage(){
  cerr << "usage: gph2nbst [-nbest N] [-raw1best] < outgraph > outnbest" << endl;
  cerr << "         -nbest   i : size of the i-best list                   [default 300]" << endl;
  cerr << "         -raw1best  : output raw 1-best hypotheses              [default false]" << endl;
  cerr << "         -moses     : output Moses format                       [default false]" << endl;
  exit(1);
}
