#!/usr/bin/perl

$MAXCOST=20;
$MINPROB=exp(-$MAXCOST);

$verbose=0;
$usage="lexreord.perl -voc f [-verbose] < units
   -voc    f : input tuples vocabulary file
   -verbose  : verbose output
 units file are produced by unf+null.perl
";

while (@ARGV>0){
    $tok=shift @ARGV;
    if ($tok=~/^\-voc$/ && @ARGV>0)  { $fvoc=shift @ARGV; next; }
    if ($tok=~/^\-verbose$/)         { $verbose=1; next; }
    die "error: unparsed parameter: $tok\n$usage";
}

die "ERROR: missing parameter -voc\n$usage" unless (defined $fvoc);

print STDERR "   =================================================================
     lexreord.perl : computes lexicalized reordering probabilities
                     MSDCFB events
   =================================================================
";

print STDERR "Reading $fvoc\n";
open (FILE,"<$fvoc") or die "ERROR: unavailable voc file: $fvoc\n";
while (<FILE>){
    chomp;
    @tok = split / \|\|\| /;
    $unit=$tok[0]." ||| ".$tok[1];
    $vocunits{$unit}=1;
    $units[$#units+1]=$tok[0]." ||| ".$tok[1];
}
close FILE;
print STDERR "Done (Read ".($#units+1)." units)\n";

print STDERR "Reading unfold units\n";
$nsent=0;
$nline=0;
while (<>){
    $nline++;
    if (/^EOS$/) {
	$nsent++;
	if ($nsent%10000==0){if ($nsent%100000==0){print STDERR $nsent;} else {print STDERR ".";}}
	&update_stats();
	$#INDEXS=-1;
	$#UNITS=-1;
	next;
    }
    chomp;
    @tokens=split / \|\|\| /;
    if ($#tokens!=2) {die "ERROR: unparsed entry at line $nline: $_\n";}
    $UNITS[$#UNITS+1] = $tokens[0]." ||| ".$tokens[1];
    $INDEXS[$#INDEXS+1] = $tokens[2];
}
print STDERR "Done!\n";

print STDERR "Building stats\n";
#foreach $unit (sort {$vocunits{$a} <=> $vocunits{$b}} keys %vocunits){
for ($u=0;$u<=$#units;$u++){
    $unit=$units[$u];
    die "error: unit '$unit' did not appeared in train!" unless (exists $Orientation{$unit});
    $total=0;
    foreach $ori (keys %{$Orientation{$unit}}){$total+=$Orientation{$unit}{$ori};}
    die "error: unit seen 0!\t unit: $unit\n" if ($total==0);
    $nM=$Orientation{$unit}{"M"};
    $nS=$Orientation{$unit}{"S"};
    $nF=$Orientation{$unit}{"F"};
    $nB=$Orientation{$unit}{"B"};
    ### I consider that I saw a few more times each unit
    $extraseen=1/$total;
#    $extraseen=1/@units;
    print "         M:$nM S:$nS F:$nF B:$nB total:$total extraseen:$extraseen unit:$unit\n" if ($verbose);
    $total=$total+$extraseen;
    ### each of the events (M,S,F,B) is given the same portion of such few more times
    $nM=$nM+($extraseen/4);
    $nS=$nS+($extraseen/4);
    $nF=$nF+($extraseen/4);
    $nB=$nB+($extraseen/4);
    print "smoothed M:$nM S:$nS F:$nF B:$nB total:$total extraseen:$extraseen unit:$unit\n" if ($verbose);
    ### orientations collapsing several events are updated accordingly
    $nD=$nF+$nB;
    $nC=$nM+$nS;

    ### output statistics
    if ($nM/$total < $MINPROB) {$M=$MAXCOST;}
    else {$M=-log sprintf("%.10f",$nM/$total);}

    if ($nS/$total < $MINPROB) {$S=$MAXCOST;}
    else {$S=-log sprintf("%.10f",$nS/$total);}

    if ($nD/$total < $MINPROB) {$D=$MAXCOST;}
    else{$D=-log sprintf("%.10f",$nD/$total);}

    if ($nC/$total < $MINPROB) {$C=$MAXCOST;}
    else {$C=-log sprintf("%.10f",$nC/$total);}

    if ($nF/$total < $MINPROB) {$F=$MAXCOST;}
    else {$F=-log sprintf("%.10f",$nF/$total);}

    if ($nB/$total < $MINPROB) {$B=$MAXCOST;}
    else {$B=-log sprintf("%.10f",$nB/$total);}

    print "$M $S $D $C $F $B\n";
}
close FILE;
print STDERR "Done (".(keys %vocunits)." units)\n";

#####################################################################################
#####################################################################################
#####################################################################################

sub update_stats(){
#    print "Sent$nsent\n" if ($verbose);
    for ($u=0;$u<=$#UNITS;$u++){
	next unless (exists $vocunits{$UNITS[$u]});
	$curr=$UNITS[$u];
	$prev="<s>";
	$prev=$UNITS[$u-1] if ($u>0);
	@curr_IND=sort {$a <=> $b} split / /,$INDEXS[$u];
	@prev_IND=("-1");
	@prev_IND=sort {$a <=> $b} split / /,$INDEXS[$u-1] if ($u>0);
	$ori="";
	if    ($prev_IND[$#prev_IND]+1==$curr_IND[0]) {$ori="M";}
	elsif ($curr_IND[$#curr_IND]+1==$prev_IND[0]) {$ori="S";}
	elsif ($prev_IND[$#prev_IND]+1<$curr_IND[0])  {$ori="F";}
	elsif ($curr_IND[$#curr_IND]+1<$prev_IND[0])  {$ori="B";}
	else                                          {$ori="S";} ### one of the units goes within the other one ex: curr_IND: 1 8 9 prev_IND: 5
#	print "Unit $u prev: @prev_IND curr: @curr_IND orientations: $ori curr: $curr prev: $prev\n" if ($verbose);
	$Orientation{$curr}{$ori}++;
    }
}


