#include <sstream>
#include <iostream>
#include <fstream>
#include <string.h>
#include <math.h>
#include <ext/hash_map>
#include <vector>

using namespace std;
using namespace __gnu_cxx;

const string str_tplseparator = "|||";
const string str_ruleseparator = "///";

namespace __gnu_cxx
{
  template<> struct hash< std::string >
  {
    size_t operator()( const std::string& x ) const
    {
      return hash< const char* >()( x.c_str() );
    }
  };
}

void find_rules(vector<string> &, vector<int> &, int, bool);
void outrule(vector<string> &,vector<int> &, int, int, int);
void recursive_rule(vector<string> &,vector<int> &, int, int);
string vecstr2str(vector<string> &, int, int, string);
string vecint2str_rescaledtozero(vector<int> &, int, int, int, string);
vector<string> split(string,const char *);
void usage();
hash_map <string, hash_map<string, int> > Lside2Rside;
hash_map <string, int> Lside;
string tplseparator;
string ruleseparator;

int main(int argc, char **argv){
  cerr << "   ==========================================" << endl;
  cerr << "     posrules: extract rewrite rules" << endl;
  cerr << "     jmcrego[at]limsi[dot]fr, February 2011" << endl;
  cerr << "   ==========================================" << endl;
  if (argc<2) usage();

  string ftrn;
  string funf;
  int max=10;
  float smooth=1;
  bool nocostmon=false;
  bool nomonrule=false;
  bool recursive=false;
  string list("");
  for (int i=1;i<argc;i++) list.append(argv[i]).append(" ");
  istringstream inStr(list);
  while (inStr){
    string word;
    inStr >> word;
    if (!inStr) continue;
    if (word.compare("-trn")==0){
      inStr >> word; ftrn=word; 
      continue;
    }
    if (word.compare("-unf")==0){
      inStr >> word; funf=word; 
      continue;
    }
    if (word.compare("-max")==0){
      inStr >> word; max=atoi(word.c_str()); 
      continue;
    }
    if (word.compare("-smooth")==0){
      inStr >> word; smooth=atof(word.c_str()); 
      continue;
    }
    if (word.compare("-nocostmon")==0){
      nocostmon=true;
      continue;
    }
    if (word.compare("-nomonrule")==0){
      nomonrule=true;
      continue;
    }
    if (word.compare("-recursive")==0){
      recursive=true;
      continue;
    }
    cerr << "error: unparsed parameter " << word << endl;
    usage();
  }

  if (ftrn.size()==0){
    cerr << "error: parameter -trn not found!" << endl;
    usage();
  }

  if (funf.size()==0){
    cerr << "error: parameter -unf not found!" << endl;
    usage();
  }

  if (max<=1){
    cerr << "error: parameter -max must be >= 2!" << endl;
    usage();
  }

  if (smooth<0){
    cerr << "error: parameter -smooth must be >= 0!" << endl;
    usage();
  }

  tplseparator.append(" ").append(str_tplseparator).append(" ");
  ruleseparator.append(" ").append(str_ruleseparator).append(" ");

  /**************************************************************/
  /*** reading ftrn funf ****************************************/
  /**************************************************************/

  ifstream itrn;
  itrn.open(ftrn.data());
  if(!itrn) {
    cerr << "error: impossible to open trn file=" << ftrn << endl;
    exit(1);
  }

  ifstream iunf;
  iunf.open(funf.data());
  if(!iunf) {
    cerr << "error: impossible to open unf file=" << funf << endl;
    exit(1);
  }

  string line_trn;
  getline(itrn, line_trn);
  int nlines=0;
  cerr << "Find rules";
  while (!itrn.eof()){
    nlines++;
    if (nlines % 10000 == 0){if (nlines % 100000 == 0){cerr << nlines;} else {cerr << ".";}}
    istringstream inStr(line_trn);
    vector<string> words;
    while (inStr){
      string word;
      inStr >> word;
      if (!inStr) continue;
      words.push_back(word);
    }
    getline(itrn,line_trn);
    //unfold ex: le str_tplseparator NULL str_separator DET_ART str_separator 0
    string line_unf;
    vector<int> indexs;
    getline(iunf, line_unf);
    while (!iunf.eof()){
      if (line_unf.compare("EOS")==0) break;
      vector<string> tokens=split(line_unf, tplseparator.c_str());
      if (tokens[0].compare("NULL")!=0){
	vector<string> tok=split (tokens[3]," ");
	for (int i=0;i<tok.size();i++){
	  indexs.push_back(atoi(tok[i].c_str()));
	}
      }
      getline(iunf,line_unf);
    }
    if (indexs.size()!=words.size()){
      cerr << "warning: diff num of tokens " << indexs.size() << "<>" << words.size() << " in sentence " << nlines << endl;
      continue;
    }
    if (words.size()==0){
      cerr << "warning: no words in sentence " << nlines << endl;
      continue;
    }
    find_rules(words,indexs,max,recursive);
  }
  itrn.close();
  iunf.close();
  cerr << "Done [" << nlines << " sentences]" <<  endl;

  /**************************************************************/
  /*** counter examples using ftrn ******************************/
  /**************************************************************/

  itrn.open(ftrn.data());
  if(!itrn) {
    cerr << "error: impossible to open trn file=" << ftrn << endl;
    exit(1);
  }

  getline(itrn, line_trn);
  nlines=0;
  cerr << "Counter examples";
  while (!itrn.eof()){
    nlines++;
    if (nlines % 10000 == 0){if (nlines % 100000 == 0){cerr << nlines;} else {cerr << ".";}}
    istringstream inStr(line_trn);
    vector<string> words;
    while (inStr){
      string word;
      inStr >> word;
      if (!inStr) continue;
      words.push_back(word);
    }
    for (int i=0;i<words.size();i++){
      string lside=words[i];
      for (int j=i+1;j<words.size()&&j-i<max;j++){
	lside.append(" ").append(words[j]);
	hash_map <string, hash_map <string, int> >::iterator it_lside2rside=Lside2Rside.find(lside);
	if (it_lside2rside!=Lside2Rside.end()){
	  hash_map <string, int>::iterator it_lside=Lside.find(lside);
	  if (it_lside!=Lside.end()) Lside[lside]++;
	  else Lside[lside]=1;
	}
      }
    }
    getline(itrn,line_trn);
  }
  itrn.close();
  cerr << "Done [" << nlines << " sentences]" <<  endl;

  /**************************************************************/
  /*** output rules *********************************************/
  /**************************************************************/

  vector<string> linear;
  for (int i=0; i<max; i++){
    string aux="0";
    for (int j=1; j<=i; j++){
      char num[10]; sprintf(num,"%d",j); 
      aux.append(" ").append(num);
    }
    linear.push_back(aux);
  }

  int totalrules=0;
  hash_map <string, hash_map <string, int> >::iterator it_lside2rside=Lside2Rside.begin();
  cerr << "Output rules file ";
  while (it_lside2rside!=Lside2Rside.end()){
    string lside=it_lside2rside->first;
    vector<string> LSIDE=split(string(lside)," ");
    hash_map <string, int>::iterator it_lside=Lside.find(lside);
    if (it_lside==Lside.end()){
      cerr << "waning: reordered sequence '" << lside << "' not found in trn!" << endl; 
      it_lside2rside++; 
      continue;
    }
    hash_map <string, int> Rside=it_lside2rside->second;
    hash_map <string, int>::iterator it_rside=Rside.begin();
    //mass used to smooth is either fixed (smooth>0) or proportional to the number of instances of the rule source side
    float smooth_mass=smooth;
    if (smooth==0) smooth_mass=(float)1/it_lside->second;
    float smooth_inst=(float)smooth_mass/(Rside.size()+1);//the additional mass distributed for each rule target side
    float total=it_lside->second + smooth_mass;
    //reordered rules
    float total_reo=0;
    while (it_rside!=Rside.end()){
      string rside=it_rside->first;
      float total_inst=it_rside->second+smooth_inst;
      total_reo+=total_inst;
      cout << lside << ruleseparator << rside << ruleseparator << -log((float)total_inst/total) << endl;
      totalrules++;
      it_rside++;
    }
    //monotonic rule
    if (!nomonrule){
      float cost=-log((float)(total-total_reo)/total);
      if (nocostmon) cost=0;
      cout << lside << ruleseparator << linear[LSIDE.size()-1] << ruleseparator << cost << endl;
      totalrules++;
    }
    it_lside2rside++;
  }
  cerr << "Done [" << totalrules << " rules]" <<  endl;
  return 0;
}

void find_rules(vector<string> &words, vector<int> &indexs, int max, bool recursive){
  int i=0;
  int max_seen=indexs[0];
  while (i < indexs.size()){
    int start_gap=i;
    if (max_seen<indexs[i]) max_seen=indexs[i];
    while (max_seen>i){
      i++;
      if (max_seen<indexs[i]) max_seen=indexs[i];
    }
    int end_gap=i;
    if (end_gap>start_gap && max>end_gap-start_gap){
      outrule(words,indexs,start_gap,end_gap,start_gap);
      if (recursive){
	for (int from=start_gap;from<end_gap;from++){
	  for (int to=from+1;to<=end_gap;to++){
	    if (from!=start_gap || to!=end_gap) recursive_rule(words,indexs,from,to);
	  }
	}
      }
    }
    i++;
  }
}

void outrule(vector<string> &words,vector<int> &indexs, int start_gap, int end_gap, int min){
  string lside=vecstr2str(words,start_gap,end_gap," ");
  string rside=vecint2str_rescaledtozero(indexs,start_gap,end_gap,min," ");
  //  cout << lside << ruleseparator << rside << endl;
  hash_map <string, hash_map <string, int> >::iterator it_lside2rside=Lside2Rside.find(lside);
  if (it_lside2rside!=Lside2Rside.end()) it_lside2rside->second[rside]++;
  else{
    hash_map <string, int> Rside;
    Rside[rside]=1;
    Lside2Rside[lside]=Rside;
  }
}

void recursive_rule(vector<string> &words, vector<int> &indexs, int start_gap, int end_gap){
  int gap=end_gap-start_gap;
  int min=indexs[start_gap];
  int max=indexs[start_gap];
  bool is_reordered=false;
  for (int i=start_gap+1;i<=end_gap;i++){
    if (indexs[i]!=indexs[i-1]+1) is_reordered=true;
    if (indexs[i]<min) min = indexs[i];
    if (indexs[i]>max) max = indexs[i];
    if (max-min>gap) return; 
  }
  if (!is_reordered) return;
  outrule(words,indexs,start_gap,end_gap,min);
}

string vecstr2str(vector<string> &words, int start_gap, int end_gap, string sep){
  string result=words[start_gap];
  for (int i=start_gap+1;i<=end_gap;i++) result.append(sep).append(words[i]);
  return result;
}

string vecint2str_rescaledtozero(vector<int> &indexs, int start_gap, int end_gap, int min, string sep){
  char num[10]; 
  sprintf(num,"%d",indexs[start_gap]-min); 
  string result=num;
  for (int i=start_gap+1;i<=end_gap;i++){ 
    sprintf(num,"%d",indexs[i]-min); 
    result.append(sep).append(num);
  }
  return result;
}

vector<string> split(string str,const char *match){
  vector<string> result;
  size_t from=0;
  while (str.find(match,from) != string::npos){
    result.push_back(str.substr(from,str.find(match,from)-from));
    from = str.find(match,from) + strlen(match);
  }
  result.push_back(str.substr(from));
  return result;
}

void usage(){
  cerr << "usage: posrules -trn f -unf f [-max i] [-smooth f] [-nocostmon] [-nomonrule] [-recursive]" << endl;
  cerr << "  -trn    f  : training source file (POS)"  << endl;
  cerr << "  -unf    f  : unfolded training (src2trg.unfoldNULL)"  << endl;
  cerr << "  -max    i  : max rule size [default 10]"  << endl;
  cerr << "  -nocostmon : no cost is used for monotonic rules [default false]"  << endl;
  cerr << "  -nomonrule : do not output monotonic rules [default false]"  << endl;
  cerr << "  -recursive : consider recursive reordering rules [default false]"  << endl;
  cerr << "  -smooth f  : smooth mass equivalently distributed acrosss all rules target side [default 0]"  << endl;
  cerr << "       case smooth:"  << endl;
  cerr << "         0 mass=1/total_instances_rule_source_side; (higher the mass when lower the apparition of the pos sequence)"  << endl;
  cerr << "        >0 mass=smooth;"  << endl;
  cerr << endl;
  exit(1);
}

//g++ filterarpa.cpp -o filterarpa -Wno-deprecated
