#!/usr/bin/perl

print STDERR "   ================================================
     pruneVocTuples: prune voc of bilingual units 
     jmcrego[at]limsi[dot]fr, February 2010
   ================================================
";

$usage="USAGE: pruneVocTuples -voc f -scores f -ovoc f -oscores f -tnb n
   -voc      f : input units vocabulary
   -scores   f : unit scores (parallel to voc file)
   -ovoc     f : ouput units vocabulary
   -oscores  f : ouput unit scores (parallel to ovoc file)
   -tnb      n : n-best units are kept
ATTENTION: scores are positive costs (-logp)
";

$verbose=0;

while ($#ARGV>=0){
    $tok = shift @ARGV;
    if ($tok =~ /^\-voc$/  && $#ARGV>=0)    {$voc = shift @ARGV;next;}
    if ($tok =~ /^\-ovoc$/ && $#ARGV>=0)    {$ovoc = shift @ARGV; next;}
    if ($tok =~ /^\-scores$/  && $#ARGV>=0)  {$score = shift @ARGV;next;}
    if ($tok =~ /^\-oscores$/ && $#ARGV>=0)  {$oscore = shift @ARGV; next;}
    if ($tok =~ /^\-tnb$/ && $#ARGV>=0)     {$tnb = shift @ARGV; next;}
    if ($tok =~ /^\-verbose$/)              {$verbose=1;next;}
    die "ERROR: parsing parameter $tok!\n$usage";
}

die "ERROR: no vocabulary file passed!\n".$usage unless ($voc !~ /^$/);
die "ERROR: no output vocabulary file passed!\n".$usage unless ($ovoc !~ /^$/);
die "ERROR: no scores file passed!\n".$usage unless ($score !~ /^$/);
die "ERROR: no output scores file passed!\n".$usage unless ($oscore !~ /^$/);
die "ERROR: no tnb passed!\n".$usage unless ($tnb !~ /^$/);

###########################################################################
###########################################################################
###########################################################################

$maxcost=99;
print STDERR "[BEGIN reading voc file]\n";
$nunits=0;
open (FVOC,"<$voc") or die "ERROR: unavailable voc file: $voc\n";
open (FSCORE,"<$score") or die "ERROR: unavailable scores file: $score\n";
while ($unit=<FVOC>,$weights=<FSCORE>){
    $nunits++;
    if ($nunits % 100000 == 0) {if ($nunits % 1000000 == 0) {print STDERR $nunits;} else {print STDERR ".";}}

    ### unit ###
    chomp $unit;
    $unit =~ s/^ +//;
    $unit =~ s/ +$//;
    $unit =~ s/  +/ /g;
    print "READ1: $unit\n" if ($verbose);
    ($u,$v,$rest) = split / \|\|\| /,$unit;

    ### scores ###
    chomp $weights;
    $weights =~ s/^ +//;
    $weights =~ s/ +$//;
    $weights =~ s/  +/ /g;
    @WEIGHTS = split / /,$weights;
    $total=0;
    for ($w=0;$w<=$#WEIGHTS;$w++) {$total+=$WEIGHTS[$w];}
    
    ### keep record ###
    $Voc{$u}{$v}=$total;
    $Scores{$u}{$v}=$weights;
}
print STDERR "(nunits:$nunits)\n";

###########################################################################
###########################################################################
###########################################################################

print STDERR "\n[OUTPUT voc of units and weights]\n";
open (FVOC,">$ovoc") or die "ERROR: cannot create file: $ovoc\n";
open (FSCORE,">$oscore") or die "ERROR: cannot create file: $oscore\n";
$nunits=0;
foreach $u (keys %Voc){
    $n=1;
    foreach $v (sort {$Voc{$u}{$a} <=> $Voc{$u}{$b}} keys %{$Voc{$u}}){
	next if ($tnb>0 && $n>$tnb);
	print FVOC "$u ||| $v\n";
	print FSCORE "$Scores{$u}{$v}\n";
	$n++;
	$nunits++;
	if ($nunits % 100000 == 0) {if ($nunits % 1000000 == 0) {print STDERR $nunits;} else {print STDERR ".";}}
    }
}
print STDERR "(nunits:$nunits)\n";
