#include <sstream>
#include <fstream>
#include <iostream>
#include <string>
#include <list>
#include <math.h>
#include "math.h"

#if __GNUC__>2
#include <ext/hash_map>
using __gnu_cxx::hash_map;
#else
#include <hash_map>
#endif

#define MODUL 100000
#define MAXCOST 1000
using namespace std;

/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/

#ifndef HASHSTRING
#define HASHSTRING
class hashfstring : public unary_function<string, size_t>{  
 public:
  size_t operator()(string const &s) const{
    unsigned int res=0;
    for (int i=0;i<s.size();i++){
      res += (size_t) (unsigned int) pow((double)s.at(i), double((i%6)+1));
    }
    return res%MODUL;
  }
};
class hasheqstring{
 public:
  bool operator()(string const &s1, string const &s2) const{
    return s1.compare(s2)==0;
  }
};
#endif

/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/

void read_ibm1(string, hash_map<string,double,hashfstring,hasheqstring> &,bool);
void solveNULLs(list<string>&);
string toLowerCase(string& s);
string add_seguent(list<string>::iterator, list<string> &);
string add_anterior(list<string>::iterator, list<string> &);
double calcula_ibm1_tupla(string);
void read_params(int, char **);
void usage();

string fibm11,fibm12,ftuples;
bool lctables,verbose;
hash_map<string,double,hashfstring,hasheqstring> jash_ibm11;
hash_map<string,double,hashfstring,hasheqstring> jash_ibm12;

/**************************************************************************/
/********** MAIN **********************************************************/
/**************************************************************************/

int main(int argc, char **argv){
  cerr << "   ====================================================" << endl;
  cerr << "     solve_srcNULLed: solves source-side NULLed units  " << endl;
  cerr << "                      using lexicon probabilities      " << endl;
  cerr << "     jmcrego[at]limsi[dot]fr, February 2006            " << endl;
  cerr << "   ====================================================" << endl;

  read_params(argc,argv);
  read_ibm1(fibm11,jash_ibm11,true);
  read_ibm1(fibm12,jash_ibm12,false);

  ifstream fin( ftuples.data() );
  if( !fin.is_open() ) {
    cerr << "Error opening file=" << ftuples << endl;
    exit(0);
  }
  list<string> listtuples;
  unsigned int nsentences=0;
  while( !fin.eof()  ){
    string s;
    getline( fin, s );
    if (s.compare("EOS")==0){
      solveNULLs(listtuples);
      listtuples.clear();
      cout << "EOS" << endl;
      nsentences++;
      if (nsentences%10000==0){
	if (nsentences%100000==0) cerr << nsentences;
	else cerr << ".";
      }
      continue;
    }
    if (verbose>1) cerr << "read " << s << endl;

    //hem quedo amb la informacio dels dos primers camps (si hi ha mes info, la llenco)
    //reitero ||| i would like once again ||| 19
    //a ||| to ||| 20
    //NULL ||| wish ||| -1
    int pos1=s.find(" ||| ",0);
    int pos2=s.find(" ||| ",pos1+1);
    string t;
    if (pos2!=string::npos){ t=s.substr(0,pos2); }//I only use the tpl_src and tpl_trg
    else{ t=s; }
    listtuples.push_back(t); ///the unit is like: ^tpl_src ||| tpl_trg$
  }
  fin.close();
}

/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/

void read_ibm1(string file, hash_map<string,double,hashfstring,hasheqstring> &jash, bool dir){
  ifstream fin( file.data() );
  if( !fin.is_open() ) {
    cerr << "Error opening file=" << file << endl;
    exit(0);
    return;
  }
  cerr << "Reading IBM1 file=" << file << endl;
  unsigned int n=0;
  while( !fin.eof()  ){
    string s;
    getline( fin, s );
    string word1,word2;
    double prob;
    istringstream inStr(s);
    inStr >> word1;
    inStr >> word2;
    inStr >> prob; 
    string key;
    if (dir) {key=word1;key.append(" ").append(word2);}
    else {key=word2;key.append(" ").append(word1);}
    jash[key]=-log(prob); //I use positive costs
    if (verbose>1) cerr << "read " << key << " " << jash[key] << endl;
    n++;
  }
  fin.close();
  cerr << "Found " << n << " probability entries" << endl;
}

/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/

void solveNULLs(list<string> &listtuples){
  list<string>::iterator it=listtuples.begin();
  list<string>::iterator itnext,itprev;
  while (it != listtuples.end()){
    if (it->find("NULL ||| ")==0){
      if (verbose>0) cout << "-------------->> found NULL in tuple=(((" << *it << "))) " << endl;      
      /*************************************************************************/
      /************* NULL que cal desfer ***************************************/
      /*************************************************************************/
      itnext=it;itnext++;
      itprev=it;itprev--;
      if (it==listtuples.begin()){
	if (verbose>0) cout << "Last tuple !!" << endl;
	//es la primera tupla, segur que la afegirem a la seguent!!!
	string seguent=add_seguent(it,listtuples); //retorna listtuples[i+1] amb i al devant		
	itnext++;
	it=listtuples.erase(it,itnext);
	it=listtuples.insert(it,seguent);
	continue;
      }
      if (itnext==listtuples.end()){
	if (verbose>0) cout << "First tuple !!" << endl;
	//es la ultima tupla, segur que la afegirem a la anterior!!!
	string anterior=add_anterior(it,listtuples); //retorna listtuples[i-1] amb i al darrere
	it++;
	it=listtuples.erase(itprev,it);
	it=listtuples.insert(it,anterior);
	continue;
      }
      // es una tupla del mig, cal veure que passa...
      string seguent=add_seguent(it,listtuples);
      string anterior=add_anterior(it,listtuples);

      double cost_anterior=calcula_ibm1_tupla(anterior)+calcula_ibm1_tupla(*itnext);
      if (verbose>0) cout << "anterior=((((" << anterior << ")))) +++ ((((" << *itnext << ")))) = " << cost_anterior << endl;

      double cost_seguent=calcula_ibm1_tupla(seguent)+calcula_ibm1_tupla(*itprev);
      if (verbose>0) cout << "seguent=((((" << *itprev << ")))) +++ ((((" << seguent << ")))) = " << cost_seguent << endl;
 
      if (cost_seguent<=cost_anterior){
	itnext++;
	it=listtuples.erase(it,itnext);
	it=listtuples.insert(it,seguent);
	if (verbose>0) cout << "-------------------->SEGUENT" << endl;
	continue;
      }
      else{
	it++;
	it=listtuples.erase(itprev,it);
	it=listtuples.insert(it,anterior);
	if (verbose>0) cout << "-------------------->ANTERIOR" << endl;
	continue;
      }
      /*************************************************************************/
      /************* NULLs desfets *********************************************/
      /*************************************************************************/      
    }
    it++;
  }

  it=listtuples.begin();
  while (it != listtuples.end()){
    //ara evito que surti al target un null amb alguna altra paraula: src # NULL trg # 0
    string tpl=*it;
    int posini=tpl.find(" ||| ");
    string tpl_src=tpl.substr(0,posini);
    string tpl_trg=tpl.substr(posini+5);

    bool replaced=false;
    int pos;
    pos=tpl_trg.find("NULL "); 
    if (pos==string::npos) pos=tpl_trg.find(" NULL");
    while (pos!=string::npos){
      tpl_trg.replace(pos,5,"");
      pos=tpl_trg.find("NULL "); 
      if (pos==string::npos) pos=tpl_trg.find(" NULL");
      replaced=true;
    }

    if (replaced) cout << tpl_src << " ||| " << tpl_trg << endl;
    else cout << tpl << endl;
    it++;
  }

}

string add_seguent(list<string>::iterator it, list<string> &listtuples){
  string source=*it;
  it++;
  string target=*it;
  int pos;

  pos=source.find(" ||| ");
  string source_trg=source.substr(pos+5);

  pos=target.find(" ||| ");
  string target_src=target.substr(0,pos);
  string target_trg=target.substr(pos+5);

  string str=target_src;
  str.append(" ||| ");
  str.append(source_trg);
  str.append(" ");
  str.append(target_trg);

  return str;
}

string add_anterior(list<string>::iterator it, list<string> &listtuples){
  string source=*it;
  it--;
  string target=*it;
  int posini;

  posini=source.find(" ||| ");
  string source_trg=source.substr(posini+5);

  posini=target.find(" ||| ");
  string target_src=target.substr(0,posini);
  string target_trg=target.substr(posini+5);

  string str=target_src;
  str.append(" ||| ");
  str.append(target_trg);
  str.append(" ");
  str.append(source_trg);

  return str;
}

double calcula_ibm1_tupla(string tupla){
  int posini=tupla.find(" ||| ");
  string tupla_src=tupla.substr(0,posini);
  string tupla_trg=tupla.substr(posini+5);

  list<string> lsrc;
  istringstream srcStr(tupla_src);
  while (srcStr){ // per cada word de la linia actual
    string word;
    srcStr >> word;
    if (!srcStr) continue;
    if (lctables && word.compare("NULL")!=0) word=toLowerCase(word);
    lsrc.push_back(word);
  }

  list<string> ltrg;
  istringstream trgStr(tupla_trg);
  while (trgStr){ // per cada word de la linia actual
    string word;
    trgStr >> word;
    if (!trgStr) continue;
    if (lctables && word.compare("NULL")!=0) word=toLowerCase(word);
    ltrg.push_back(word);
  }

  double cost=0;
  hash_map<string,double,hashfstring,hasheqstring>::iterator it_jash;
  list<string>::iterator i;
  list<string>::iterator j;

  int Nlinks=lsrc.size()+ltrg.size();

  for (i=lsrc.begin();i!=lsrc.end();i++){
    for (j=ltrg.begin();j!=ltrg.end();j++){
      string key(*i);
      key.append(" ");
      key.append(*j);

      it_jash=jash_ibm11.find(key);
      if (it_jash==jash_ibm11.end()){
	if (verbose>1) cout << "ibm11(" << key << ")=" << MAXCOST << endl;
	cost=cost+MAXCOST;
      }
      else{
	double cst=it_jash->second/Nlinks;
	if (verbose>1) cout << "ibm11(" << key << ")=" << cst << endl;
	cost=cost+cst;
      }

      it_jash=jash_ibm12.find(key);
      if (it_jash==jash_ibm12.end()){
	if (verbose>1) cout << "ibm12(" << key << ")=" << MAXCOST << endl;
	cost=cost+MAXCOST;
      }
      else{
	double cst=it_jash->second/Nlinks;
	if (verbose>1) cout << "ibm12(" << key << ")=" << cst << endl;
	cost=cost+cst;
      }
    }
  }      
  return cost;
}

/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/

void read_params(int argc, char **argv){
  fibm11="";
  fibm12="";
  ftuples="";
  lctables=false;
  verbose=0;
  
  int i=1;
  while (i<argc){
    if (string(argv[i]).compare("-fibm11")==0)   {i++;fibm11=argv[i];i++; continue;}
    if (string(argv[i]).compare("-fibm12")==0)   {i++;fibm12=argv[i];i++; continue;}
    if (string(argv[i]).compare("-ftuples")==0)  {i++;ftuples=argv[i];i++;continue;}
    if (string(argv[i]).compare("-verbose")==0)  {i++;verbose=atoi(argv[i]);i++;continue;}
    if (string(argv[i]).compare("-lowercase")==0){i++;lctables=true;      continue;}
    cerr << "ERROR: Incorrect parameter " << argv[i] << " !!" << endl;
    usage();    
  }

  if (fibm11.size()==0){
    cerr << "ERROR: undefined fibm11 !!" << endl;
    usage();
  }
  if (fibm12.size()==0){
    cerr << "ERROR: undefined fibm12 !!" << endl;
    usage();
  }
  if (ftuples.size()==0){
    cerr << "ERROR: undefined ftupls !!" << endl;
    usage();
  }
}

// Make a lowercase copy of s:
string toLowerCase(string& s) {
  char* buf = new char[s.length()];
  s.copy(buf, s.length());
  for(int i = 0; i < s.length(); i++)
    buf[i] = tolower(buf[i]);
  string r(buf, s.length());
  delete buf;
  return r;
}

void usage(){
  cerr << "USAGE: solve_srcNULLed -fibm11 f -fibm12 f -ftuples f [Optional]" << endl;
  cerr << " -fibm11  f : file with ibm11 probabilities (trg src prob)" << endl;
  cerr << " -fibm12  f : file with ibm12 probabilities (src trg prob)" << endl;
  cerr << " -ftuples f : file with tuples sequence to resolve" << endl;
  cerr << " -verbose N : level of verbosity {0,1,2} [default 0]" << endl;
  cerr << "Optional" << endl;
  cerr << " -lowercase : assume IBM tables are in lowercase" << endl;
  exit(0);
}


