#include "Auxiliar.h"

vector<vector<string> > sequence;
hash_map<string,hash_map<string,int> > wrdunit2posunit2nseen;
hash_map<string,string> wrdunit2posunit;
hash_map<string,int> posunit2id;
hash_set<int> Id;
string voc;
string stag;
string ttag;
string name;
bool relate;

void usage();
void readUnits();
void parse_sentence(int,string,string,vector<vector<string> > &,vector<vector<string> > &,vector<string> &);
void computeVoc();
void outputFactors();
void outputSeq();

/**************************************************************************/
/********** MAIN **********************************************************/
/**************************************************************************/
int main(int argc, char **argv){
  cerr << "   ==========================================" << endl;
  cerr << "     unf2fact: factor extractor for bincoder "<< endl;
  cerr << "     jmcrego[at]limsi[dot]fr  -  (May 2011)  " << endl;
  cerr << "   ==========================================" << endl;
  
  /***************************************************************/
  /********** READING PARAMETERS *********************************/
  /***************************************************************/
  relate=false;
  if (argc<=1) usage();
  int narg=1;
  while (narg<argc){
    string word=argv[narg++];
    if (word.compare("-voc")==0  && narg<argc) {voc=argv[narg++]; continue;}
    if (word.compare("-stag")==0 && narg<argc) {stag=argv[narg++]; continue;}
    if (word.compare("-ttag")==0 && narg<argc) {ttag=argv[narg++]; continue;}
    if (word.compare("-name")==0 && narg<argc) {name=argv[narg++]; continue;}
    if (word.compare("-relate")==0)            {relate=true; continue;}
    cerr << "error: unparsed parameter: " << word << endl;
    usage(); 
  }

  readUnits();
  computeVoc();
  outputFactors();
  outputSeq();
  
  cerr << "Done" << endl;
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

void readUnits(){
  cerr << "Reading units,stag,ttag" << endl;
  std::ifstream fstag (stag.c_str(), std::ios_base::in);
  if (!fstag.is_open()) {cerr << "error: opening stag " << stag << endl; exit(1);}
  std::ifstream fttag (ttag.c_str(), std::ios_base::in);
  if (!fttag.is_open()) {cerr << "error: opening ttag " << ttag << endl; exit(1);}
  vector<vector<string> > unitSrc;
  vector<vector<string> > unitTrg;
  vector<string> unitInd;
  int nSent=0;
  string stag_line,ttag_line,line;
  while( getline(cin,line) ) {
    if (line.compare("EOS")==0){
      nSent++;
      if (nSent%1000==0) {if (nSent%10000==0) {cerr << nSent;} else {cerr << ".";}}
      getline(fstag, stag_line, '\n');
      getline(fttag, ttag_line, '\n');
      parse_sentence(nSent,stag_line,ttag_line,unitSrc,unitTrg,unitInd);
      unitSrc.clear();
      unitTrg.clear();
      unitInd.clear();
    }
    else{
      vector<string> tokens=split(line,str_separator.c_str());
      unitSrc.push_back(split(tokens[0]," "));
      unitTrg.push_back(split(tokens[1]," "));
      unitInd.push_back(tokens[2]);
    }
  } 
  fstag.close();
  fttag.close();
  cerr << endl << "Done (" << nSent << " sents)" << endl;
}

void parse_sentence(int nSent,string stag_line,string ttag_line,vector<vector<string> > &unitSrc,vector<vector<string> > &unitTrg,vector<string> &unitInd){
  vector<string> STAG=split(stag_line," ");
  vector<string> TTAG=split(ttag_line," ");
  
  int ntrg=0;
  vector<string> unitsequence;
  for (int u=0;u<unitSrc.size();u++){
    string wrdSrc=join(unitSrc[u]," ");
    string wrdTrg=join(unitTrg[u]," ");
    istringstream sInd(unitInd[u]);
    vector<int> indSrc;
    int pos;
    while (sInd >> pos) indSrc.push_back(pos);
    string posSrc;
    for (int p=0;p<indSrc.size();p++){
      if (posSrc.size()>0) posSrc.append(" ");
      if (indSrc[p]>=STAG.size()) {
	cerr << "error: wrong number of source tokens in sentence " << nSent << endl; 
	cerr << "STAG: " << stag_line << endl;
	for (int k=0;k<unitSrc.size();k++) cerr << "unitSrc[" << k << "]: " << join(unitSrc[k]," ") << endl;
	exit(1);
      }
      posSrc.append(STAG[indSrc[p]]);
    }
    string posTrg;
    for (int p=0;p<unitTrg[u].size();p++){
      if (posTrg.size()>0) posTrg.append(" ");
      if (unitTrg[u][p].compare("NULL")==0) posTrg.append("NULL");
      else{
	if (ntrg>=TTAG.size()) {
	  cerr << "error: wrong number of target tokens in sentence " << nSent << endl; 
	  cerr << "TTAG: " << ttag_line << endl;
	  for (int k=0;k<unitTrg.size();k++) cerr << "unitTrg[" << k << "]: " << join(unitTrg[k]," ") << endl;
	  exit(1);
	}
	posTrg.append(TTAG[ntrg++]);
      }
    }
    string wrdUnit,posUnit;
    wrdUnit.append(wrdSrc).append(" ||| ").append(wrdTrg);
    posUnit.append(posSrc).append(" ||| ").append(posTrg);
    if (wrdunit2posunit2nseen.find(wrdUnit)==wrdunit2posunit2nseen.end()){
      hash_map<string,int> h;
      h[posUnit]=1;
      wrdunit2posunit2nseen[wrdUnit]=h;
    }
    else wrdunit2posunit2nseen[wrdUnit][posUnit]++;
    unitsequence.push_back(wrdUnit);
  }
  sequence.push_back(unitsequence);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

void computeVoc(){
  cerr << "compute wrd:tag associations" << endl;
  hash_map<int,int> Total;
  hash_map<string, hash_map<string,int> >::iterator it=wrdunit2posunit2nseen.begin();
  while (it!=wrdunit2posunit2nseen.end()){
    string wrdunit=it->first;
    string posunit;
    int nseen=0;
    hash_map<string,int>::iterator it2=(it->second).begin();
    int total=0;
    while (it2!=(it->second).end()){
      total++;
      if (it2->second > nseen){
	posunit=it2->first;
	nseen=it2->second;
      }
      it2++;
    }
    Total[total]++;
    wrdunit2posunit[wrdunit]=posunit;
    int num=posunit2id.size();
    if (posunit2id.find(posunit)==posunit2id.end()) posunit2id[posunit]=num;
    it++;
  }
  hash_map<int,int>::iterator it2=Total.begin();
  while (it2!=Total.end()) {cerr << it2->first << ": " << it2->second << endl;it2++;}
  cerr << "Done!" << endl;
}

void outputFactors(){
  std::ofstream fobilrelate;
  if (relate){
    string obilrelate=voc;
    obilrelate.append(".").append(name).append(".bil.relate");
    fobilrelate.open(obilrelate.c_str(), std::ios_base::out);
    if (!fobilrelate.is_open()) {cerr << "error: opening obilrelate " << obilrelate << endl; exit(1);}
    cerr << "building " << obilrelate << endl;
  }
  string obilfactor=voc;
  obilfactor.append(".").append(name).append(".bil.factor");
  std::ofstream fobilfactor;
  fobilfactor.open(obilfactor.c_str(), std::ios_base::out);
  if (!fobilfactor.is_open()) {cerr << "error: opening obilfactor " << obilfactor << endl; exit(1);}
  cerr << "building " << obilfactor << endl;

  string osrcfactor=voc;
  osrcfactor.append(".").append(name).append(".src.factor");
  std::ofstream fosrcfactor;
  fosrcfactor.open(osrcfactor.c_str(), std::ios_base::out);
  if (!fosrcfactor.is_open()) {cerr << "error: opening osrcfactor " << osrcfactor << endl; exit(1);}
  cerr << "building " << osrcfactor << endl;

  string otrgfactor=voc;
  otrgfactor.append(".").append(name).append(".trg.factor");
  std::ofstream fotrgfactor;
  fotrgfactor.open(otrgfactor.c_str(), std::ios_base::out);
  if (!fotrgfactor.is_open()) {cerr << "error: opening otrgfactor " << otrgfactor << endl; exit(1);}
  cerr << "building " << otrgfactor << endl;

  string line;
  std::ifstream fvoc(voc.c_str(), std::ios_base::in);
  if (!fvoc.is_open()) {cerr << "error: opening voc " << voc << endl; exit(1);}
  while( getline(fvoc,line) ) {
    string wrdunit=line;
    hash_map<string,string>::iterator it_wrdunit=wrdunit2posunit.find(wrdunit);
    if (it_wrdunit==wrdunit2posunit.end()) {cerr << "error: unseen wrdunit: " << wrdunit << endl; exit(1);}
    string posunit=it_wrdunit->second;
    hash_map<string,int>::iterator it_posunit=posunit2id.find(posunit);
    if (it_posunit==posunit2id.end()) {cerr << "error: unseen posunit: " << posunit << endl; exit(1);}
    int id=it_posunit->second;
    vector<string> tokens=split(posunit,str_separator.c_str());
    if (relate) fobilrelate << posunit << endl;
    fobilfactor << id << endl;
    fosrcfactor << tokens[0] << endl;
    fotrgfactor << tokens[1] << endl;
    Id.insert(id);
  }  
  fvoc.close();
  if (relate) fobilrelate.close();
  fobilfactor.close();
  fosrcfactor.close();
  fotrgfactor.close();
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

void outputSeq(){
  string obilseqvoc=voc;
  obilseqvoc.append(".").append(name).append(".bil.seq.voc");
  std::ofstream fobilseqvoc;
  fobilseqvoc.open(obilseqvoc.c_str(), std::ios_base::out);
  if (!fobilseqvoc.is_open()) {cerr << "error: opening obilseqvoc " << obilseqvoc << endl; exit(1);}
  cerr << "building " << obilseqvoc << endl;
  hash_set<int>::iterator it=Id.begin();
  while (it!=Id.end()){
    fobilseqvoc << *it << endl;
    it++;
  }
  
  string obilseq=voc;
  obilseq.append(".").append(name).append(".bil.seq");
  std::ofstream fobilseq;
  fobilseq.open(obilseq.c_str(), std::ios_base::out);
  if (!fobilseq.is_open()) {cerr << "error: opening obilseq " << obilseq << endl; exit(1);}
  cerr << "building " << obilseq << endl;

  for (int i=0;i<sequence.size();i++){
    for (int j=0;j<sequence[i].size();j++){
      ostringstream num;
      string wrdunit=sequence[i][j];
      hash_map<string,string>::iterator it_wrdunit=wrdunit2posunit.find(wrdunit);
      if (it_wrdunit==wrdunit2posunit.end()) {cerr << "error: unseen wrdunit: " << wrdunit << endl; exit(1);}
      string posunit=it_wrdunit->second;
      hash_map<string,int>::iterator it_posunit=posunit2id.find(posunit);
      if (it_posunit==posunit2id.end()) {cerr << "error: unseen posunit: " << posunit << endl; exit(1);}
      int id=it_posunit->second;
      num << id;
      sequence[i][j]=num.str();
    }
    fobilseq << join(sequence[i]," ") << endl;
  }
  fobilseqvoc.close();
  fobilseq.close();
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

void usage(){
  cerr << "usage: unf2fact -stag f -ttag f -voc f -name s [-relate] < units > voc.name.{bil.{seq{,.voc},fator,vocab},src.factor,trg.factor}" << endl;
  cerr << "\t-voc    f : input tuples vocabulary file" << endl;
  cerr << "\t-stag   f : training source tagged file" << endl;
  cerr << "\t-ttag   f : training target tagged file" << endl;
  cerr << "\t-name   f : extension name" << endl;
  cerr << "\t-relate   : output voc.name.bil.relate file" << endl;
  cerr << "\tunits file are produced by unfold+null.perl" << endl;
  exit(1);
}
